#ifndef alg_base_hpp
#define alg_base_hpp

/******************************************************
*
 ** alg_base by Sinisa Segvic:
 *
 ** the entry function for an image processing algorithm
 ** INPUT: source images and user events
 ** OUTPUT: destination images and their annotations
 *
 ** in order to support factory method,
 ** each concrete class should:
 **  - declare public constructor with a single 
 **    argument of the type CtorArg 
 **  - define static data member of the type 
 **    util::RegistryCreator<alg_base,alg_concrete>
 **    which interfaces to the registry object
 **  - define static member s_name() returning a 
 **    distinguishing name used for identification;
 **    s_name() may be used for implementing name()
 *
 *******************************************************/

#include "../img/img_data.hpp"

class img_vectorAbstract;
class cli_wrapAbstract;
class win_ann_abstract;
class win_event_vectorAbstract;
class ui_reportAbstract;


class alg_base{
protected:
  alg_base();
public:
  virtual ~alg_base();

//identification
public:
  virtual char const* name() =0;

//mandatory operations
public:
  virtual void process( 
    const img_vectorAbstract& src, 
    const win_event_vectorAbstract& events,
    int msDayUtc, int frame);
  virtual void process( 
    const img_vectorAbstract& src, 
    const win_event_vectorAbstract& events,
    int msDayUtc); 
//discretionary operations
public:
  virtual void config(cli_wrapAbstract& cli);
  virtual void profile(ui_reportAbstract& );

//output
protected:
  void checkId(int id, char const* context);
public:
  virtual int countDst() =0; // >=1!
  virtual char const* captionDst(int id);
  virtual const img_fmt& fmtDst(int id);
  virtual const img_data& imgDst(int id) =0;
  virtual win_ann_abstract const* pAnnDst(int id); 

// alg_base construction
public:
  // concrete classes ctor argument
  class CtorArg{
  public:
    img_fmt fmt;//image format
    char const* cfg;//configuration string
    CtorArg(const img_fmt& f, char const* c):fmt(f),cfg(c){}
    CtorArg(const img_fmt& f):fmt(f),cfg(""){}
  };
  
  //factory for the concrete classes 
  static alg_base* create(
    const char* name, 
    const CtorArg& arg);
};

#endif
