#include "../util/util_registry.hpp"
#include "../util/util_string.hpp"
#include "../alg/alg_base.hpp"

#include <algorithm>
#include <iostream>
#include <sstream>
#include <string>

/******************************************************
*
** alg_dllAdapter by Sinisa Segvic
*
** C interface for creating and destroying 
** concrete alg_base-s defined within
** an explicitly linked library
*
** exceptions from real object should not propagate
** to main app since the dll would be unloaded 
** before the exception is catched (?)
*
** supposed to be used in conjuction with alg_dllStub
*
*******************************************************/

#if defined(_WIN32) || defined (__WIN32__)
 #if defined(_USRDLL) || defined(_BUILD_DLL)
  #define EXPORT __declspec(dllimport)
 #else
  #define EXPORT __declspec(dllexport)
 #endif
#else
 #define EXPORT 
#endif


namespace {
  int handleError(
    char const*msg)
  {
    std::ostringstream oss;
    oss <<"alg_dllAdapter: caught exception in library object:\n" 
        <<msg <<std::endl <<std::endl;
    std::cerr <<oss.str().c_str();
    return 1;
  }
  
  std::string getAlg(char const* p){
    std::string rv;
    //search for explicit algorithm
    size_t len=strlen(p);
    char const* pos=std::find(p, p+len, '^');
    if (pos<p+len){
      rv=std::string(pos+1);
    } else{
      //remove path and extension from the file name    
      util_string::retractPath(p, rv);
      size_t pos=rv.find_first_of(".");
      if (pos != std::string::npos){
        rv.erase(rv.begin()+pos, rv.end());
      }
    }
    return rv;
  }

}




// exported functions
extern "C" EXPORT void* alg_dllAdapter_create(void const* arg);
extern "C" EXPORT void alg_dllAdapter_destroy(void* obj);


/////////////////////////////////////
// construction

extern "C" EXPORT void* 
alg_dllAdapter_create(void const* pArg){
  const alg_base::CtorArg& arg=*
    static_cast<alg_base::CtorArg const*>(pArg);
  util::Registry& reg=util::Registry::instance();

  std::string name=getAlg(arg.cfg);

  //search for the matching creator
  util::Registry::Iterator it=reg.find<alg_base>(name.c_str());
  if (it == reg.end<alg_base>()){
    //search for the first creator
    it=reg.begin<alg_base>();
  }

  alg_base* pAlg=0;
  if (it != reg.end<alg_base>()){
    std::cerr <<"alg_dllAdapter: trying algorithm " 
      <<'"' <<it->name_derived() <<'"' <<std::endl;

    try {
      pAlg=reg.create<alg_base>(it,arg);
    } catch(std::exception& e){
      handleError(e.what());
    }
  } else{
    std::cerr <<"alg_dllAdapter: no algorithms found in " 
      <<arg.cfg <<std::endl;
  }
  
  return pAlg;
}
  

extern "C" EXPORT void 
alg_dllAdapter_destroy(void* p){
  alg_base *pAlg=static_cast<alg_base*>(p);
  delete pAlg;
}



