#include "alg_base.hpp"

#include "../util/util_registry.hpp"
#include "../util/util_dll.hpp"
#include "../util/util_string.hpp"

#include <string.h>
#include <algorithm>

/******************************************************
*
** alg_dllStub by Sinisa Segvic
*
** a cvsh proxy algorithm marshalling the procedure calls 
** to another algorithm residing within the dynamic library
*
** exceptions from real object should not propagate
** to main app since the dll would be unloaded 
** before the exception is catched (?)
*
** Proxy:Proxy 
*
** references:
** - http://aegisknight.org/cppinterface.html
*******************************************************/

///////////////////////////////////
// declaration of function types

//ptr to ctor type
typedef void* (* PtrCtor)(void const* ctorarg);
//ptr to dtor type
typedef void (* PtrDtor)(void* obj);


///////////////////////////////////
// declaration of alg_dllStub

class alg_dllStub:
  public alg_base
{
public:
  alg_dllStub(const CtorArg&);
  virtual ~alg_dllStub();
public: 
  virtual char const* name();
  static char const* s_name(){return "dllStub";}
public:
  virtual void process( 
    const img_vectorAbstract& src, 
    const win_event_vectorAbstract& events, 
    int msDayUtc); 
  virtual void config(cli_wrapAbstract& cli);
  virtual void profile(ui_reportAbstract& );

//output
public:
  virtual int countDst();//>=1!
  virtual char const* captionDst(int id);
  virtual const img_fmt& fmtDst(int id);
  virtual const img_data& imgDst(int id);
  virtual win_ann_abstract const* pAnnDst(int id);
private:
  void handleError(char const*);
private:
  std::string dllName_;
  util_dll dll_;
  int dllId_;

  PtrCtor pCtor_;
  PtrDtor pDtor_;

  alg_base* pAlgDll_;
};

namespace{
  util::RegistryCreator<alg_base,alg_dllStub> creator;

  std::string getDll(char const *p){
    char const* pos=std::find(p, p+strlen(p), '^');
    return std::string(p, pos);
  }
}


///////////////////////////////////
// alg_dllStub construction

alg_dllStub::alg_dllStub(const CtorArg& a):
  dllName_(getDll(a.cfg)),
  dll_(dllName_.c_str())
{
  pCtor_=(PtrCtor)
    dll_.getFun("alg_dllAdapter_create");
  pDtor_=(PtrDtor)
    dll_.getFun("alg_dllAdapter_destroy");

  pAlgDll_=static_cast<alg_base*>((*pCtor_)(&a));
  
  if (!pAlgDll_){
    std::string str("alg_dllStub::alg_dllStub - error creating library object\n");
    str+="  (";
    str+=a.cfg;
    str+=")";
    throw std::runtime_error(str);
  }
}

alg_dllStub::~alg_dllStub(){
  (*pDtor_)(pAlgDll_);
}


char const* alg_dllStub::name(){
  return pAlgDll_->name();
}



///////////////////////////////////
// alg_dllStub operations

void alg_dllStub::process(
  const img_vectorAbstract& src, 
  const win_event_vectorAbstract& events,
  int msDayUtc)
{
  pAlgDll_->process(src,events,msDayUtc);
}
void alg_dllStub::config(cli_wrapAbstract& cli){
  pAlgDll_->config(cli);
}
void alg_dllStub::profile(ui_reportAbstract& pd){
  pAlgDll_->profile(pd);
}



///////////////////////////////////
// alg_dllStub output

int alg_dllStub::countDst(){
  return pAlgDll_->countDst();
}
char const* alg_dllStub::captionDst(int id){
  return pAlgDll_->captionDst(id);
}
const img_fmt& alg_dllStub::fmtDst(int id){
  return pAlgDll_->fmtDst(id);
}
const img_data& alg_dllStub::imgDst(int id){
  return pAlgDll_->imgDst(id);
}
win_ann_abstract const* alg_dllStub::pAnnDst(int id){
  return pAlgDll_->pAnnDst(id);
}



