#include "alg_base.hpp"

#include "../util/util_registry.hpp"
#include "../img/img_vector.hpp"
#include "../win/ann/win_ann.hpp"
#include "../cli/cli_wrapAbstract.hpp"

#include <boost/smart_ptr.hpp>

#include <sstream>

////////////////////////////////////
// alg_multiplex declaration

class alg_multiplex: 
  public alg_base
{
public:
  alg_multiplex(const CtorArg& arg);
  virtual ~alg_multiplex();
public:
  virtual char const* name(){return s_name();}
  static char const* s_name(){return "multiplex";}
private:
  virtual void process( 
    const img_vectorAbstract& src, 
    const win_event_vectorAbstract& events,
    int msDayUtc); 
  virtual void config(cli_wrapAbstract& cli);
  virtual void profile(ui_reportAbstract& );
private:
  virtual int countDst(); // >=1!
  virtual char const* captionDst(int id);
  virtual const img_data& imgDst(int id);
  virtual win_ann_abstract const* pAnnDst(int id); 
private:
  std::vector<img_data> imgs_;
  std::vector<win_ann> anns_;
  boost::scoped_ptr<alg_base> spAlg_;
};

////////////////////////////////////
// static registry creator object

namespace{
  util::RegistryCreator<alg_base,alg_multiplex> creator;
} // unnamed namespace end

  
////////////////////////////////////
// alg_multiplex implementation

alg_multiplex::alg_multiplex(const CtorArg& arg):
  imgs_(1),
  anns_(1),
  spAlg_(alg_base::create("nop",arg))
{
  imgs_[0].reformat(arg.fmt);
}

alg_multiplex::~alg_multiplex(){}

void alg_multiplex::process(
  const img_vectorAbstract& src,
  const win_event_vectorAbstract& events, int time)
{
  imgs_.resize(src.size());
  anns_.resize(src.size());

  img_vector tmp(1);
  for (int i=0; i<src.size(); ++i){
    tmp[0]=src[i];
    spAlg_->process(tmp, events,time);
    imgs_[i]=spAlg_->imgDst(0);
    const win_ann_abstract* pAnn=spAlg_->pAnnDst(0);
    if (pAnn)
      anns_[i]=*pAnn;
    else
      anns_[i].clear();
  }
}

void alg_multiplex::profile(
  ui_reportAbstract& pd)
{
  //todo
  spAlg_->profile(pd);
}
void alg_multiplex::config(
  cli_wrapAbstract& cli)
{
  std::ostringstream oss;
  oss <<"Current " <<name() <<" algorithm: " <<spAlg_->name();  
  oss <<"Please supply the new slave algorithm:\n";
  cli.put(oss.str().c_str());

  char const* buf=cli.get();
  if (strlen(buf)>0){
    if (strcmp(buf, spAlg_->name())){
      spAlg_.reset(alg_base::create(buf,
        alg_base::CtorArg(imgs_[0].fmt())));
    }
  }

  oss.str("The new slave algorithm is: ");
  oss <<spAlg_->name() <<"\n";
  cli.put(oss.str().c_str());

  spAlg_->config(cli);
}

//////////////////////////////////
// output

int alg_multiplex::countDst(){
  return static_cast<int>(imgs_.size());
}
char const* alg_multiplex::captionDst(int id){
  checkId(id, "alg_multiplex");
  return spAlg_->captionDst(id);
}
const img_data& alg_multiplex::imgDst(int id){
  checkId(id, "alg_multiplex::imgDst");
  return imgs_[id];
}
win_ann_abstract const* alg_multiplex::pAnnDst(int id){
  checkId(id, "alg_multiplex");
  return &anns_[id];
}
