#include "cli_parser.hpp"
#include "cli_wrap.hpp"

#include "../util/util_string.hpp"


#include <iostream>
#include <stdexcept>
#include <sstream>
#include <string>
#include <vector>
#include <algorithm>

#include <assert.h>



// Examples of usage
// # vs2vd -b -sf=path.ext#from_frame-to_frame -dn=port
//    (from path to IP port)

// # vs2vd -b -sn=host:port -dw 
//    (from IP port@host to window)

// # vs2vd -b -sg=win32MCI@0.1.0:320x200x1 -df=path.ext
//     (from grabber@board.channel.mode:wxhxszPix to file)

// # vs2vd -sg=win32MCI@0:320x200x1 -i="p n -1" -c="1 1 loop 0"
//     (from grabber@board, interactive, with initial command 
//      and configuration )

// # vs2vd -i -sg=win32MCI
//     (from grabber, interactive)


////////////////////////////////
// private local functions 

namespace{
  static void badParam(
    int i, char const* par,
    char const* msg)
  {
    std::ostringstream oss;
    oss <<"cli_parser::processParam: Error in parameter " 
        <<i <<" \"" <<par <<"\":\n  " <<msg;
    std::string str=oss.str();
    //std::cerr <<str <<"\n";
    //assert(0);
    throw std::runtime_error(str);
  }
  inline void assertParam(
    bool cond,
    char const* msg)
  {
    if (!cond)
      throw std::runtime_error(std::string(msg));
  }
}


////////////////////////////////
// local functions (parsing)

void cli_parser::parseRange(
  std::string& str, 
  cli_dev& dev)
{
  int from=0, to=-1;
  size_t pos=str.find('#');
  if (pos!=std::string::npos) {
    size_t pos2=str.find_last_of('-');
    if (pos2==std::string::npos){
      pos2=str.size();
    } else{
      std::string strTo(str.begin()+pos2+1, str.end());
      std::istringstream(strTo) >>to;
    }
    std::string strFrom(str.begin()+pos+1, str.begin()+pos2);
    std::istringstream(strFrom) >>from;

    dev.setRange(from,to);

    str.erase(pos);
  }
}

void cli_parser::parseFile(
  const std::string& str, 
  cli_dev& dev)
{
  assertParam(str[0]=='=', 
    "syntax error, '=' expected");
  dev.setType(cli_dev::File);

  std::string tmp=str;
#ifdef _WIN32
  std::replace(tmp.begin(), tmp.end(), '/', '\\');
#endif
  dev.setName(tmp.c_str()+1);
}


void cli_parser::parseNet(
  const std::string& str, 
  cli_dev& dev)
{
  assertParam(str[0]=='=', 
    "syntax error, '=' expected");
  dev.setType(cli_dev::Net);
  size_t pos=str.find(':');

  int port;
  if (pos!=std::string::npos){
    dev.setName(std::string(str.begin()+1,str.begin()+pos).c_str());
    std::string tmp(str.begin()+pos+1, str.end());
    std::istringstream(tmp) >>port;
  } else{
    std::string tmp(str.begin()+1, str.end());
    std::istringstream(tmp) >>port;
  }
  dev.pushPar(port);
}

void cli_parser::parseGrab(
  const std::string& str, 
  cli_dev& dev)
{
  //std::cerr <<"str=" <<str <<"\n";
  assertParam(str[0]=='=', 
    "syntax error, '=' expected");
  dev.setType(cli_dev::Grab);

  // all six params (board, channel, mode, width , height, depth)
  // are set to default value (0)
  std::vector<int> par(6,0);

  // search for the grabber name
  size_t posNameEnd=str.find_first_of(":@");
  if (posNameEnd==std::string::npos){
    posNameEnd=str.size();
  } 
  dev.setName(str.substr(1, posNameEnd-1).c_str());
  
  // do we have a board-channel specification?
  size_t posBoardBeg=str.find('@');
  if (posBoardBeg!=std::string::npos){
    size_t pos0=posBoardBeg+1;
    for (int i=3; i<6; i++){
      bool done=false;
      size_t pos1=str.find_first_not_of("0123456789", pos0);
      if (pos1==std::string::npos){
        done=true;
        pos1=str.size();
      } else {
        done=(str[pos1]!='.');
      }
        
      par[i]=util_string::strToInt(str.substr(pos0,pos1-pos0));
        
      if (done)
        break;
      pos0=pos1+1;
    }
  }
    
  // do we have a dimensions specification?
  size_t posDimBeg=str.find(':');
  if (posDimBeg!=std::string::npos){
    size_t pos0=posDimBeg+1;
    for (int i=0; i<3; i++){
      bool done=false;
      size_t pos1=str.find_first_not_of("0123456789", pos0);
      if (pos1==std::string::npos){
        done=true;
        pos1=str.size();
      } else {
        done=(str[pos1]!='x');
      }
        
      par[i]=util_string::strToInt(str.substr(pos0,pos1-pos0));
        
      if (done)
        break;
      pos0=pos1+1;
    }
  }
    
  dev.setPar(par);
}

void cli_parser::parseWnd(
  const std::string& str, 
  cli_dev& dev)
{
  if (!str.empty())
    throw std::runtime_error("unknown modifier for window.");

  dev.setType(cli_dev::Win);
}

void cli_parser::parseSrc(
  char const* param, 
  cli_dev& src)
{
  std::string str(param+1);
  parseRange(str, src);

  //cli_dev type?
  switch (param[0]){
  case 'f':
    parseFile(str, src);
    break;
  case 'n':
    parseNet(str, src);
    break;
  case 'g':
    parseGrab(str, src);
    break;
  default:
    throw std::runtime_error("unknown video source");
  }
}

void cli_parser::parseDst(
  char const* param, 
  cli_dev& dev)
{
  std::string str(param+1);

  //cli_dev type?
  switch (param[0]){
  case 'f':
    parseFile(str, dev);
    break;
  case 'n':
    parseNet(str, dev);
    break;
  case 'w':
    parseWnd(str, dev);
    break;
  default:
    throw std::runtime_error("unknown video destination");
  }
}

void cli_parser::parseCam(
  const std::string& str, 
  cli_dev& dev)
{
  std::string rest(str);

  //both params (port, hfov)
  //are set to default value (0)
  std::vector<int> vecPar(2,0);

  size_t pos=rest.find(':');
  if (pos!=std::string::npos){
    dev.setId(std::string(rest.begin()+pos+1, rest.end()).c_str());
    rest=std::string(rest.begin(),rest.begin()+pos);
  } 

  pos=rest.find('@');
  if (pos!=std::string::npos){
    std::string params(rest.begin()+pos+1, rest.end());
    rest=(std::string(rest.begin(),rest.begin()+pos));

    for (size_t i=0; i<vecPar.size(); ++i){
      size_t pos=params.find('.');
      if (pos!=std::string::npos){
        std::string par(params.begin(), params.begin()+pos);
        std::istringstream(par) >>vecPar[i];
        params=std::string(params.begin()+pos+1,params.end());
      } else{
        std::istringstream(params) >>vecPar[i];
        break;
      }
    } 
  } 
  dev.setName(rest.c_str());
  dev.setPar(vecPar);
}





////////////////////////////////
// local functions (construction)

void cli_parser::processParam(
  char const* param)
{
  cli_dev tmp;

  if (param[0]=='-'){
    switch (param[1]){
    case 's':
      //first source?
      //if (src_.type()!=cli_dev::Null) 
      //  throw std::runtime_error("multiple sources");
      parseSrc(param+2, tmp);
      vec_src_.push_back(tmp);
      break;
    case 'd':
      parseDst(param+2, tmp);
      vec_dst_.push_back(tmp);
      break;
    case 'a':
      //first algorithm?
      if (!alg_.empty()) 
        throw std::runtime_error("multiple algorithms");
      assertParam(param[2]=='=', 
        "syntax error, '=' expected");
      alg_=std::string(param+3);
      break;
    case 'c':
      //first configuration string?
      if (!config_.empty()) 
        throw std::runtime_error("multiple configuration strings");
      assertParam(param[2]=='=', 
        "syntax error, '=' expected");
      config_=std::string(param+3);
      std::replace(config_.begin(), config_.end(), '_', ' ');
      break;
    case 'i':
      //first initial command?
      interactive_=true;
      if (param[2]){
        if (!command_.empty()) 
          throw std::runtime_error("multiple initial commands");
        assertParam(param[2]=='=', 
          "syntax error, '=' expected");
        command_=std::string(param+3);
        std::replace(command_.begin(), command_.end(), '_', ' ');
      }
      break;
    case 'm':
      {
        //maxfps constraint?
        assertParam(param[2]=='=', 
          "syntax error, '=' expected");
        int value=util_string::strToInt(
          std::string(param+3));
        assertParam(value>0, 
          "syntax error, positive value expected");
        minDelayMs_=1000/value;
      }
      break;
    case '-':
      {
        std::string option(param+2);
        size_t pos=option.find('=');
        assertParam(pos!=std::string::npos, 
          "syntax error, '=' expected");
        std::string optionpar(option.begin()+pos+1, option.end());
        assertParam(optionpar.size()>0, 
          "syntax error, expected option parameters");
        option.erase(option.begin()+pos, option.end());
      
        if (option=="ptu"){
          parseCam(optionpar, ptu_);
          ptu_.setType(cli_dev::PTU);
        } else if (option=="cam"){
          parseCam(optionpar, cam_);
          cam_.setType(cli_dev::Camera);
        } else{
          std::string msg="Unknown option - "+option;
          assertParam(false, msg.c_str());
        }
      }
      break;

    default:
      throw std::runtime_error("unrecognized option");
    }
  } else {
    throw std::runtime_error("unrecognized argument");
  }
}


/////////////////////////////////
// construction 
// (all the work is performed here)

cli_parser::cli_parser(
  int argc, 
  char* argv[])
{
  if (argc==1)
    throw std::runtime_error("cli_parse: got no parameters");

  interactive_=false;
  minDelayMs_=0;
  alg_.erase();
  config_.erase();
  command_.erase();

  for (int i=1; i<argc; i++){
    try{
      processParam(argv[i]);
    } catch (char const* msg){
      badParam(i, argv[i], msg);
    } catch (const std::exception& e){
      badParam(i, argv[i], e.what());
    }
  }

  //if no sources were given, set default
  if (vec_src_.empty()){
    vec_src_.push_back(cli_dev("test", cli_dev::Grab, ""));
    vec_src_[0].setPar(std::vector<int>(6,0));
    vec_src_[0].setRange(0,1);
  }

  if (alg_.empty()){
    alg_="nop";
  }

  if (interactive_ || vec_dst_.empty()){
    //interactive sessions require exactly one
    //vd_win which may be omitted on the command line
    bool havewin=false;
    for (size_t j=0; j<vec_dst_.size(); j++){
      if (vec_dst_[j].type()==cli_dev::Win){
        havewin=true;
      }
    }
    if (!havewin){
      cli_dev dev;
      dev.setType(cli_dev::Win);
      vec_dst_.push_back(dev);
    }
  }

  //complain if no destinations specified
  if (vec_dst_.empty()){
    throw "cli_parser::cli_parser - "
      "at least one destination should be specified "
      "in the batch mode";
  }

  //check for multiple vdWin-s
  int nWin=0;
  {for (size_t i=0; i<vec_dst_.size(); i++){
    if (vec_dst_[i].type()==cli_dev::Win){
      nWin++;
    }
  }}
  if (nWin>1){
    throw "cli_parser::cli_parser - "
      "multiple destination windows are"
      "not permitted";
  }

  //assign names to unnamed destinations
  for (size_t j=0; j<vec_dst_.size(); j++){
    if (vec_dst_[j].type()==cli_dev::Win)
      vec_dst_[j].setName(vec_src_[0].name());
  }
}


/////////////////////////////////
// static member functions

void cli_parser::usageSrc(cli_wrap& cli){
  std::ostringstream oss;
  oss <<"  -s<Source>[<Range>] " <<std::endl;
  oss <<"    <Source>:" <<std::endl
      <<"      - f=filename" <<std::endl
      <<"      - n=host:port" <<std::endl
      <<"      - g=name[@board[.channel[.mode]][:wxh[xszpix]]]" <<std::endl
      <<"        unixV4L:" <<std::endl
      <<"          mode=0 => PAL (default)" <<std::endl
      <<"          mode=2 => NTSC" <<std::endl
      <<"          mode=3 => AUTO" <<std::endl
      <<"        w32PXC:" <<std::endl
      <<"          mode=0 => ConsiderSingleField (default)" <<std::endl
      <<"          mode=1 => AllwaysBothFields" <<std::endl;
  cli.msg(oss.str().c_str());
  usageRange(cli);
}
  
void cli_parser::usageDst(cli_wrap& cli){
  std::ostringstream oss;
  oss <<"  -d<Dest> [-d<Dest>]" <<std::endl;
  oss <<"    <Dest>:" <<std::endl
      <<"      - f=filename" <<std::endl
      <<"      - n=port" <<std::endl
      <<"      - w" <<std::endl;
  cli.msg(oss.str().c_str());
  usageRange(cli);
}

void cli_parser::usageAlg(cli_wrap& cli){
  std::ostringstream oss;
  oss <<"  -a=<Algorithm> " <<std::endl;
  oss <<"    <Algorithm>:" <<std::endl
      <<"      - a precompiled module identifier (e.g. nop)\n"
      <<"      - path to the plug-in module (dynamic library)\n";
  oss <<"  -c=<ConfigString> " <<std::endl;
  oss <<"    <ConfigString>:" <<std::endl
      <<"      (algorithm dependent)" <<std::endl;
  cli.msg(oss.str().c_str());
}
void cli_parser::usageOpt(cli_wrap& cli){
  std::ostringstream oss;
  oss <<"  -m=maxfps             (processing speed limit [fps])" <<std::endl
      <<"  -i[=<CommandString>]  (interactive processing)" <<std::endl;
  oss <<"    <CommandString>:" <<std::endl
      <<"      (invoke with -i and type help to obtain" <<std::endl
      <<"       the list of available commands)" <<std::endl;
  cli.msg(oss.str().c_str());
}
void cli_parser::usageRange(cli_wrap& cli){
  std::ostringstream oss;
  oss <<"    <Range>:" <<std::endl;
  oss <<"      #frameStart-frameEnd" <<std::endl;
  cli.msg(oss.str().c_str());
}
void cli_parser::usageCam(cli_wrap& cli){
  std::ostringstream oss;
  oss <<"  --ptu=<PtuModel>@port:id"     <<std::endl
      <<"  --cam=<CamModel>@port.hfov:id" <<std::endl;
  cli.msg(oss.str().c_str());
}




