#include "cvsh_createAlg.hpp"
#include "cvsh_createVS.hpp"
#include "cvsh_createVD.hpp"

#include "../alg/alg_base.hpp"

#include "../util/util_registry.hpp"

#include "../vd/vd_base.hpp"
#include "../vd/vd_win.hpp"
#include "../vs/vs_base.hpp"

#include "../cli/cli_dev.hpp"
#include "../cli/cli_parser.hpp"
#include "../cli/cli_wrap.hpp"

#include "../ui/ui.hpp"

#include "../util/util_string.hpp"
#include "../util/util_os.hpp"
#include "../util/util_fs.hpp"

#include <sys/types.h>
#include <sys/stat.h>
//#include <unistd.h>

#include <boost/smart_ptr.hpp>

#include <iostream>
#include <memory>
#include <sstream>
#include <exception>

#include <time.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>

/////////////////////////
// local functions begin
namespace{

void cvsh_signal_handler(int signum){
  std::ostringstream oss;
  util_os::printBacktrace();

  std::cerr <<"Caught signal: " <<signum <<"\n";
  std::string tmpdir(util_fs::getTmpDir());
  std::cerr <<"Check for core in: " <<tmpdir <<"\n";
//  chdir(tmpdir.c_str());
  
  signal(SIGABRT, SIG_DFL);
  abort();
}

typedef void (*handler_t)(int);

void setsignals(handler_t f){
  //signal(SIGINT, f);
  //signal(SIGSEGV, f);
  //signal(SIGTERM, f);
  //signal(SIGABRT, f);
  // signal(SIGPIPE, f);
  // signal(SIGWINCH, f);
  //signal(SIGFPE, f);
}


int handleError(
  cli_wrapAbstract& cli, 
  char const*msg)
{
  std::ostringstream oss;
  oss <<"\nCaught exception in main:\n" ;
  std::string indented;
  util_string::indent(msg, indented);
  oss <<indented <<std::endl;
  oss <<"  Invoke with no parameters for "
      <<"more information.\n" <<std::endl;
  cli.msg(oss.str().c_str());
  //cli_parser::usage();
  return 1;
}



void process(
  cli_wrap& cli,
  cli_parser& parser)
{
  std::ostringstream oss;
  oss <<"Creating data structures:\n";
  // create video sources
  // ranges and initial seek are also handled here
  std::vector<boost::shared_ptr<vs_base> > vecSpVs;
  cvsh_createVS(oss, parser.src(), vecSpVs);

  //get the first source
  assert(vecSpVs.size()>0);
  vs_base& vs0(*vecSpVs[0]);

  //create image processing procedure (worker object)
  boost::shared_ptr<alg_base> spalg
    (cvsh_createAlg(oss, parser.alg(), vs0.fmt()));

  //create video destinations
  //   !! strange things used to happen when 
  //   !! invalid format had been given to vd_win (?)
  std::vector<boost::shared_ptr<vd_base> > vecSpVd;
  cvsh_createVD(oss, parser.dst(), spalg->fmtDst(0), vecSpVd);

  // report the creation stage
  cli.msg(oss.str().c_str());

  // configuration
  ui x(cli, vecSpVs, spalg, vecSpVd, parser.minDelayMs());
  x.config(parser.config());

  // command loop
  if (parser.interactive()){
    x.loop(parser.command());
  } else{
    x.batch(vs0.size(), 1);
  }

  // video sources and destinations are deleted
  // automagically from smart ptr destructors
}

/////////////////////////
} // end of the unnamed namespace
// local functions end
  


// Examples of usage
// # cvsh -sf=path.ext#from_frame-to_frame -dn=port
//    (from path to IP port)
// # cvsh -sn=host:port -dw 
//    (from IP port@host to window)
// # cvsh -sg=linuxV4L320x200x8 -df=path.ext
//     (from grabber linuxV4L)
// # cvsh -sf=c:/temp/avi/z7.avi#25-56 -dw
// # cvsh -sf=c:/temp/avi/z7.avi#25-56 -df=c:/temp/avi/bla.avi -dw
// # cvsh -sf=c:/temp/avi/z7.avi#25-60 -df=c:/temp/avi/bla.srs -dw
// # cvsh -sf=c:/temp/avi/z7a.srs -dw
// # cvsh -sf=c:/temp/avi/z7a.srs -df=c:/temp/avi/bla.avi -dw

//////////////////////////////////////////////////////
// main
int main(int argc, char* argv[]){
  setsignals(cvsh_signal_handler);

  cli_wrap cli(std::cin, std::cout, std::cout);

  std::string str=util_os::env::report(argc,argv);
  cli.msg(str.c_str());

  if (argc==1){
    std::ostringstream oss;
    oss <<"Usage: cvsh {<Option>}" <<std::endl;
    oss <<"<Option>:" <<std::endl;
    cli.msg(oss.str().c_str());
    cli_parser::usageSrc(cli);
    cli_parser::usageDst(cli);
    cli_parser::usageAlg(cli);
    cli_parser::usageOpt(cli);

    oss.str("");
    oss <<"NB: -i should not be used with -d." <<std::endl;
    cli.msg(oss.str().c_str());
  } else{

    try{
      //parse command line
      cli_parser parser(argc, argv);

      //report results
      std::ostringstream oss;
      oss <<"Command line interpretation:"           <<std::endl;
      for (unsigned int i=0; i<parser.src().size(); i++){
        oss <<"  src[" <<i <<"]:      " <<parser.src()[i] <<std::endl;
      }
      for (unsigned int i=0; i<parser.dst().size(); i++){
        oss <<"  dst[" <<i <<"]:      " <<parser.dst()[i] <<std::endl;
      }
      oss <<"  algorithm:   " <<parser.alg()         <<std::endl;
      oss <<"  config:      " <<parser.config()      <<std::endl;
      oss <<"  interactive: " <<parser.interactive() <<std::endl;
      oss <<"  command:     " <<parser.command()     <<std::endl;
      oss <<"  minDelayMs:  " <<parser.minDelayMs()  <<std::endl;
      cli.msg(oss.str().c_str());

      //do processing
      process(cli, parser);
    } catch(char const* msg){
      return handleError(cli, msg);
    } catch(std::exception& e){
      return handleError(cli, e.what());
/*
    } catch (...){
       cli.msg("Caught unknown exception in main.\n");
       return 1;
*/
    }
  }

  // cli.msg("Exiting the program.\n");
  return 0;
}


/*****************************
 ** project.conf
 *****************************
    NAME=cvsh
    PACKAGES=
      cli/
      cvsh/
      img/
      grid/
      ipp/
      alg/
      pd/
      thr/
      ui/
      util/
      vd/
      vs/
      win/
    FILTER=
      alg_dllAdapter
      thr_IPserver
      cvsh_createCcam
    CONFIG= grab win_x
******************************/

