#include "grid_draw.hpp"

#include <assert.h>

#include <memory>
#include <stdexcept>


/********************************************
*
**  grid::draw 
*
**  description and encapsulation of the
**  mapping between logical and physical
**  coordinate spaces
*
*********************************************/

//////////////////////////////////////
// construction
grid::draw::draw(
  int x_f, int x_o,
  int y_f, int y_o)
{
  reformat(x_f,x_o, y_f,y_o);
}
grid::draw::~draw()
{}

void grid::draw::reformat(
  int x_f, int x_o,
  int y_f, int y_o)
{
  x_factor=x_f;
  x_offset=x_o;
  y_factor=y_f;
  y_offset=y_o;
}

//////////////////////////////////////
// translation

int grid::draw::transformx(int x){
  return x_factor*x+x_offset;
}
int grid::draw::transformy(int y){
  return y_factor*y+y_offset;
}



//////////////////////////////////////
// Pixel drawing functions

void grid::draw::drawPtDot(
  const Pixel& p)
{
  drawPtDotImp(
    transformx(p.x()),
    transformy(p.y()));
}

void grid::draw::drawPtDotBig(
  const Pixel& p)
{
  int x=transformx(p.x());
  int y=transformy(p.y());
  drawPtDotBigImp(x,y);
}

void grid::draw::drawPtDotBigImp(
  int x, int y)
{
//  drawPtDotImp(x-1,y-1);
//  drawPtDotImp(x-1,y  );
//  drawPtDotImp(x-1,y+1);
//  drawPtDotImp(x  ,y-1);
  drawPtDotImp(x  ,y  );
  drawPtDotImp(x  ,y+1);
//  drawPtDotImp(x+1,y-1);
  drawPtDotImp(x+1,y  );
  drawPtDotImp(x+1,y+1);
}

void grid::draw::drawPtBold(
  const Pixel& p, int size)
{
  drawPtBoldImp(
    transformx(p.x()),
    transformy(p.y()), 
    size);
}

void grid::draw::drawPtBoldSmall(
  const Pixel& p)
{
  drawPtBold(p, 1);
}

void grid::draw::drawPtBoldLarge(
  const Pixel& p)
{
  drawPtBold(p, 2);
}

void grid::draw::drawPtBoldHuge(
  const Pixel& p)
{
  drawPtBold(p, 4);
}

//////////////////////////////////////
// String drawing functions

void grid::draw::drawStr(
  const Pixel& p, char const* str, int sz)
{
  drawStrImp(
    transformx(p.x()),
    transformy(p.y()),
    str, sz);
}

//////////////////////////////////////
// Line drawing functions

void grid::draw::drawLine(
  const Pixel& p_s, const Pixel& p_e, 
  int width, int style)
{
  assert(style>=0 && style<2);
  if (style<0) style=-style;
  if (style>=2) style=style%2;
 
  switch (style){
  case 0: // normal
    drawLineImp(
      transformx(p_s.x()), transformy(p_s.y()),
      transformx(p_e.x()), transformy(p_e.y()),
      width);
    break;
  case 1: // dashed
    drawLineDashedImp(
      transformx(p_s.x()), transformy(p_s.y()),
      transformx(p_e.x()), transformy(p_e.y()),
      width);
    break;
  default: // error
    throw std::runtime_error("grid::draw::drawLine - unknown style");
  }
}

void grid::draw::drawLineHair(
  const Pixel& p_s, const Pixel& p_e)
{
  drawLine(p_s, p_e, 1, 0);
}


//////////////////////////////////////
// Circle drawing functions

void grid::draw::drawCircle(
  const Pixel& p, int radius, 
  int linestyle,  int fillstyle, 
  int width)
{
  assert(linestyle>=0 && linestyle<2);
  if (linestyle<0) linestyle=-linestyle;
  if (linestyle>=2) linestyle=linestyle%2;
  assert(fillstyle>=0 && fillstyle<2);
  if (fillstyle<0) fillstyle=-fillstyle;
  if (fillstyle>=2) fillstyle=fillstyle%2;
 
  switch (fillstyle){
  case 0: // no fill
    switch (linestyle){
    case 0: // normal
      drawCircleImp(
        transformx(p.x()), transformy(p.y()),
        radius, width);
      break;
    case 1: // dashed
      drawCircleDashedImp(
        transformx(p.x()), transformy(p.y()),
        radius, width);
      break;
    }
    break;
  case 1: // filled
    drawCircleFilledImp(
      transformx(p.x()), transformy(p.y()),
      radius);
    break;
  default: // error
    throw std::runtime_error("grid::draw::drawCircle - unknown style");
  }
}


//////////////////////////////////////
// Contour drawing functions

void grid::draw::drawContour(
  const Contour& c, int style)
{
  std::vector<std::pair<int,int> > 
    ps(c.size());
  for (unsigned i=0; i<c.size(); i++){
    const grid::Pixel& p=c[i];
    ps[i]=std::pair<int,int>(
      transformx(p.x()),
      transformy(p.y())
    );
  }

  assert(style>=0);
  if (style<0) style=0;
  if (style>1) style=1;

  switch (style){
  case 0: // dot
    drawContourDotImp(ps);
    break;
  case 1: // bold
    drawContourBoldImp(ps);
    break;
  default: // error
    throw std::runtime_error("grid::draw::drawContour - unknown style");
  }
}

void grid::draw::drawContourDotImp(
  const std::vector<std::pair<int,int> >& points)
{
  for (unsigned i=0; i<points.size(); i++){
    drawPtDotImp(
      points[i].first, points[i].second
    );
  }
}

void grid::draw::drawContourBoldImp(
  const std::vector<std::pair<int,int> >& points)
{
  for (unsigned i=0; i<points.size(); i++){
    drawPtDotBigImp(
      points[i].first, points[i].second
    );
  }
}

//////////////////////////////////////
// Area drawing functions

//draws the convex hull around the
//ordered set of points in c

void grid::draw::drawArea(
  const Contour& c)
{
  std::vector<std::pair<int,int> > 
    ps(c.size());
  for (unsigned i=0; i<c.size(); i++){
    const grid::Pixel& p=c[i];
    ps[i]=std::pair<int,int>(
      transformx(p.x()),
      transformy(p.y())
    );
  }

  drawAreaImp(ps);
}

void grid::draw::drawAreaImp(
  const std::vector<std::pair<int,int> >& )
{
  //by default not supported, derived classes
  //may provide another implementation
  //(translation issues must be taken care of)

  throw std::runtime_error("grid::draw::drawArea - operation not supported");
}


//////////////////////////////////////
// Colour drawing functions

// all pure virtual
