#ifndef grid_draw_hpp
#define grid_draw_hpp


/********************************************
 *
 *  component:
 **    grid_draw
 *
 * provides a common interface to 
 * implementation dependent drawing 
 * of geometric primitives
 * (points, lines and text)
 *
 * the origin of the standard client
 * coordinate system is in the lower 
 * left corner of the image
 *
 ** author: Sinisa Segvic
 ** version 2.1 (2001-08-24)
 *
 ********************************************/

#include "../root/grid_primitives.hpp"


namespace grid{

class draw{

//construction
protected:
  // concrete classes needing coordinate systems
  // which are different from client coordinate
  // system with origin at the lower left angle,
  // may provide the matching transform parameters 
  // x_factor, x_offset, y_factor, y_offset
  
  draw(int x_factor, int x_offset,
       int y_factor, int y_offset);
  virtual ~draw();

  void reformat(
    int x_factor, int x_offset,
    int y_factor, int y_offset);

//translation
private:
  int transformx(int x);
  int transformy(int y);

//operations
public:
  virtual void flush() =0;

//string drawing
private:
  virtual void drawStrImp(
    int x, int y, char const*, int sz) =0;
public:
  //supported sizes are normal (0), small (<0)
  //and large (>0)
  void drawStr(
    const Pixel& p, char const* s, 
    int sz);

//point drawing
private:
  virtual void drawPtDotImp(
    int x, int y) =0;
  virtual void drawPtBoldImp(
    int x, int y, int sz) =0;
protected:
  virtual void drawPtDotBigImp(
    int x, int y);
public:
  void drawPtDot(
    const Pixel& p);
  void drawPtDotBig(
    const Pixel& p);
  void drawPtBold(
    const Pixel& p, int size);
  void drawPtBoldSmall(
    const Pixel& p);
  void drawPtBoldLarge(
    const Pixel& p);
  void drawPtBoldHuge(
    const Pixel& p);

//line drawing
private:
  virtual void drawLineImp(
    int xs, int ys, int xe, int ye,  
    int width) =0;
  virtual void drawLineDashedImp(
    int xs, int ys, int xe, int ye,  
    int width) =0;
public:
  //supported styles are normal (0) and dashed (1)
  void drawLine(
    const Pixel& p0, const Pixel& p1, 
    int width, int style);
  void drawLineHair(
    const Pixel& p0, const Pixel& p1);

//circle drawing 
public:
  // supported styles are normal (0),
  // dashed (1) and filled (2)
  void drawCircle(
    const Pixel& p0, int radius, 
    int linestyle, int fillstyle, 
    int width);
private:
  virtual void drawCircleImp(
    int x, int y, int radius, 
    int width)=0;
  virtual void drawCircleDashedImp(
    int x, int y, int radius, 
    int width)=0;
  virtual void drawCircleFilledImp(
    int x, int y, int radius)=0;


//contour drawing 
public:
  //supported point styles are dot (0) and bold (1)
  void drawContour(
    const Contour& c, int style);
protected:
  virtual void drawContourDotImp(
    const std::vector<std::pair<int,int> >& points);
  virtual void drawContourBoldImp(
    const std::vector<std::pair<int,int> >& points);

//area drawing ...
public:
  void drawArea(
    const Contour& c);
protected:
  virtual void drawAreaImp(
    const std::vector<std::pair<int,int> >& points);

//colours 
public:
  //for i<0 returns grey(-i-1)
  virtual void colour(int i) =0;
  //grey(0) := black
  //grey(2,4,6,...) := progression black -> grey 50% 
  //grey(oo) := grey 50%
  //grey(1) := white
  //grey(1,3,5,...) := progression white -> grey 50%
  virtual void grey(int i) =0;
  virtual void bw(bool white) =0;

//music (optional)
public:
  virtual void playSound(int,int){}

//data
private:
  int x_factor, x_offset;
  int y_factor, y_offset;

//forbidden
private:
  const draw& operator=(const draw& other);
};

} //namespace grid

#endif
