#include "grid_primitives.hpp"

#include "../../util/util_math.hpp"

#include <stdexcept>

#include <assert.h>
#include <iostream>

using namespace grid;

///////////////////////
// private helpers

namespace{
  int myround(double f){
    return static_cast<int>(f<0?f-.5:f+.5);
  }

  void mytestx(double rho, double phi, 
    double y, double x0,double x1,
    std::vector<grid::Pixel>& vec)
  {
    double x=(-rho+::cos(phi)*y)/sin(phi);
    if (x>=x0 && x<x1){
      vec.push_back(grid::Pixel(myround(x),myround(y)));
    }
  }

  void mytesty(
    double rho, double phi, 
    double x, double y0,double y1,
    std::vector<grid::Pixel>& vec)
  {
    double y=(rho+::sin(phi)*x)/cos(phi);
    if (y>=y0 && y<y1){
      vec.push_back(grid::Pixel(myround(x),myround(y)));
    }
  }
}

/////////////////////////////////
// Pixel
namespace {

  inline int fixByOne(int t){
    if (t<0){
      t++;
    } else if (t>0){
      t--;
    } 
    return t;
  }

  int ss_round (double f){
    if (f>0) f+=.5;
    if (f<0) f-=.5;
    return static_cast<int>(f);
  }
}


int Pixel::l2gapSquared(
  const Pixel& o) const
{
  int dx=fixByOne(o.x()-x());
  int dy=fixByOne(o.y()-y());
  return dx*dx+dy*dy;
}

int Pixel::l2squared(
  const Pixel& o) const
{
  int dx=o.x()-x();
  int dy=o.y()-y();
  return dx*dx+dy*dy;
}


bool Pixel::operator==(
  const Pixel& o) const
{
  return x()==o.x() && y()==o.y();
}

const Pixel& Pixel::operator*=(double m){
  x_=ss_round(x_*m);
  y_=ss_round(y_*m);
  return *this;
}
const Pixel& Pixel::operator+=(const Pixel& o){
  x_+=o.x();
  y_+=o.y();
  return *this;
}
const Pixel& Pixel::operator-=(const Pixel& o){
  x_-=o.x();
  y_-=o.y();
  return *this;
}


Pixel grid::operator+(
  const Pixel& lhs, const Pixel& rhs)
{
  return grid::Pixel(lhs)+=rhs;
}
Pixel grid::operator-(
  const Pixel& lhs, const Pixel& rhs)
{
  return grid::Pixel(lhs)-=rhs;
}
Pixel grid::operator*(
  double m, const Pixel& rhs)
{
  return grid::Pixel(rhs)*=m;
}


std::ostream& grid::operator<<(
  std::ostream& os, 
  const grid::Pixel& p)
{
  os <<p.x() <<" " <<p.y();
  return os;
}

std::istream& grid::operator>>(
  std::istream& is, 
  grid::Pixel& p)
{
  is >>p.x_ >>p.y_;
  return is;
}

/////////////////////////////////
// Line

Line::Line(
  double rho, double phi, 
  int x0,int y0, int x1,int y1)
{
  std::vector<grid::Pixel> ptln;
  mytesty(rho, phi, x0, y0,y1, ptln);
  mytesty(rho, phi, x1, y0,y1, ptln);
  mytestx(rho, phi, y0, x0,x1, ptln);
  mytestx(rho, phi, y1, x0,x1, ptln);
  assert(ptln.size()>=2);
  p1_=ptln[0];
  p2_=ptln[1];
}

std::ostream& grid::operator<<(
  std::ostream& os, 
  const grid::Line& l)
{
  os <<l.p1_ <<" " <<l.p2_;
  return os;
}

std::istream& grid::operator>>(
  std::istream& is, 
  grid::Line& l)
{
  is >>l.p1_ >>l.p1_;
  return is;
}


/////////////////////////////////
// Weighted

bool grid::operator<(
  const grid::Weighted& lhs, 
  const grid::Weighted& rhs)
{
  return lhs.wt() < rhs.wt();
}


////////////////////////////////////////
// WtContourWt

void WtContourWt::clearAll(){
  clear();
  wtSet(0);
}


/////////////////////////////////
// freestanding functions

int grid::l2squared(
  const grid::Pixel& pt,
  const grid::ContourWt& c,
  int& pos)
{
  return l2squared(pt,c,0,static_cast<int>(c.size()),pos);
}

int grid::l2squared(
  const grid::Pixel& pt,
  const grid::ContourWt& c, 
  int begin, int end, 
  int& pos)
{
  pos=-1;
  int minSqrDist=INT_MAX;
  for (int i=begin; i<end; i++){
    const grid::Pixel& pt_i=c[i];

    int sqrDist=pt.l2squared(pt_i);
    if (sqrDist<minSqrDist){
      minSqrDist=sqrDist;
      pos=i;
    }
  }
  assert(pos>=0);
  return minSqrDist;
}

grid::Line grid::findBox(const std::vector<grid::Pixel>& pts){
  size_t sz=pts.size();
  if (sz==0)
    throw std::runtime_error("grid::findBox - zero points on input");
  int minx=pts[0].x();
  int miny=pts[0].y();
  int maxx=minx;
  int maxy=miny;
  for (unsigned i=1; i<sz; ++i){
    maxx=std::max(maxx, pts[i].x());
    maxy=std::max(maxy, pts[i].y());
    minx=std::min(minx, pts[i].x());
    miny=std::min(miny, pts[i].y());
  }
  ++maxx; ++maxy;
  return grid::Line(grid::Pixel(minx,miny), grid::Pixel(maxx,maxy));
}
