#ifndef grid_primitives_hpp
#define grid_primitives_hpp

#include <utility>
#include <assert.h>

#include "../../util/util_cxx.hpp"

namespace grid{

  //////////////////////////////////////////////////
  // Pixel
  class Pixel{
  public:
    Pixel(){}
    Pixel(int x, int y):x_(x),y_(y){}
    Pixel(const Pixel& o):x_(o.x_),y_(o.y_){}
  public:
    int x() const {return x_;}
    int y() const {return y_;}
    int norm() const {return x_*x_+y_*y_;}
  public:
    int l2squared(const Pixel& o) const;
    int l2gapSquared(const Pixel& o) const;
  public:
    bool operator==(const Pixel& o) const;
  public:
    const Pixel& operator*=(double m);
    const Pixel& operator+=(const Pixel& o);
    const Pixel& operator-=(const Pixel& o);
  private:
    int x_;
    int y_;

    friend std::ostream& operator<<(std::ostream& os, const Pixel&);
    friend std::istream& operator>>(std::istream&, Pixel&);
  };

  Pixel operator+(const Pixel& lhs, const Pixel& rhs);
  Pixel operator-(const Pixel& lhs, const Pixel& rhs);
  Pixel operator*(double m, const Pixel& rhs);
  
  inline bool less(
    grid::Pixel lhs, grid::Pixel rhs)
  {
    return lhs.y()<rhs.y() ||
      lhs.y()==rhs.y() && lhs.x()<rhs.x();
  }

  std::ostream& operator<<(std::ostream& os, const Pixel&);
  std::istream& operator>>(std::istream& is, Pixel&);

  //////////////////////////////////////////////////
  // Line
  class Line{
  public:
    Line(){}
    Line(const Pixel& p1, const Pixel& p2):p1_(p1),p2_(p2){}
    Line(double rho, double phi, int x0,int y0, int x1,int y1);
    Line(const Line& o):p1_(o.p1_),p2_(o.p2_){}
    const Pixel& first() const {return p1_;}
    const Pixel& second() const {return p2_;}
  public:
    int l2squared() const {return p1_.l2squared(p2_);}
  private:
    Pixel p1_;
    Pixel p2_;

    friend std::ostream& operator<<(std::ostream& os, const Line&);
    friend std::istream& operator>>(std::istream&, Line&);
  };

  std::ostream& operator<<(std::ostream& os, const Line&);
  std::istream& operator>>(std::istream& is, Line&);

  //////////////////////////////////////////////////
  // collections of basic types
  class Contour:
    public util_cxx::cvector<Pixel>
  {};

  //////////////////////////////////////////////
  // mixin for Weighted objects of all types
  class Weighted{
  public:
    Weighted():wt_(0){}
    Weighted(int wt):wt_(wt){}
    Weighted(const Weighted& o):wt_(o.wt_){}

    int wt() const {return wt_;}
    void wtAdd(int w) {wt_+=w;}
    void wtSet(int w) {wt_=w;}
  private:
    int wt_;
  };

  bool operator<(
    const Weighted& lhs, 
    const Weighted& rhs);


  //////////////////////////////////////////////////
  // Weighted classes
  class WtPixel:
    public Weighted,
    public Pixel
  {
  public:
    WtPixel(){}
    WtPixel(const Pixel& p, int w):
      Weighted(w), Pixel(p){}
  };

  class WtContour:
    public Weighted,
    public Contour
  {};

  //////////////////////////////////////////////////
  // collections of weighted types
  class ContourWt: 
    public util_cxx::cvector<WtPixel>
  {};

  //////////////////////////////////////////////////
  // more concrete Weighted classes
  class WtContourWt: 
    public Weighted,
    public ContourWt
  {
    public:
      void clearAll();
  };


  ///////////////////////////////////////////////////
  // operations
  int l2squared(
    const grid::Pixel& pt,
    const grid::ContourWt& c, int& pos);
  int l2squared(
    const grid::Pixel& pt,
    const grid::ContourWt& c, 
    int begin, int end, int& pos);
  grid::Line findBox(const std::vector<grid::Pixel>&);
  

}


#endif

