#ifndef img_access_guard_hpp
#define img_access_guard_hpp

/******************************************************
*
*  component: img_access
*  class: img_access_guard
*
** provides guarded access to images
** of different pixel types using
** static polymorphism
*
** must be parameterized with a class
** implementing img_accessor concept
*
** uses img_data (for construction only)
*
** author: Sinisa Segvic
** version 2.0 (2001-02-26)
*
*******************************************************/
#include <assert.h>
#include "../img_data.hpp"

/////////////////////////////////////////////
// img_accessor concept

//concept img_accessor

// typedef value_type
// typedef pointer
// typedef reference

// conversion to ptr
// op +=(int)



/////////////////////////////////////////////
//Accessor traits

//general img_accessor traits
template<class Accessor>
struct img_accessor_traits{
  typedef typename Accessor::value_type value_type;
  typedef typename Accessor::pointer    pointer;
  typedef typename Accessor::reference  reference;
};


//the following code would allow using an 
//ordinary pointer as img_accessor

//However, MSVC does not support 
//partial template specialization 
//so that accessor wrapper around 
//pointer has to be used :-(
//see img_access_wrap.h for details


/*
  //img_access<pointer> traits
  template<class T>
  struct img_accessor_traits<T*>{
    typedef T& reference;
    typedef T* pointer;
    typedef T  value_type;
  };

  //img_access<const pointer> traits
  template<class T>
  struct img_accessor_traits<T const*>{
    typedef const T& reference;
    typedef const T* pointer;
    typedef T        value_type;
  };
*/


///////////////////////////////////////////////////
// declaration of the img_access_guard template
template<class Accessor>
class img_access_guard{
public:
  typedef typename img_accessor_traits<Accessor>::value_type Pixel;
  typedef typename img_accessor_traits<Accessor>::pointer    Pointer;
  typedef typename img_accessor_traits<Accessor>::reference  Reference;

//construction
public:
  img_access_guard(img_data& img);
  img_access_guard(const img_data& img);
public:
  template<class Accessor2>
    img_access_guard(const img_access_guard<Accessor2>& other);
  template<class Accessor2> const img_access_guard<Accessor>& 
    operator=(const img_access_guard<Accessor2>& other);
/*
  //member template functions defined in-class
  //because of the buggy MS compiler
  template<class Accessor2>
    img_access_guard(const img_access_guard<Accessor2>& other)
  :
    origin_(other.origin()), ptr_(other.ptr()),
    szLine_(other.szLine()), height_(other.height())
  {
    assert(ptr_);
  }

public:
  template<class Accessor2> const img_access_guard<Accessor>& 
    operator=(const img_access_guard<Accessor2>& other)
  {
    ptr_=other.ptr_; 
    assert(isNear()); 
    assert(origin_==other.origin());
    assert(szLine_==other.szLine()); 
    assert(height_==other.height());
    return *this;
  }
*/

//access
public:
  Reference operator*() const;
  Pointer operator->() const;
  Pointer ptr() const;

//vertical relative displacement
  void incLine();
  void incLine(int howMany);
  void decLine();
  void decLine(int howMany);
//horizontal relative displacement
  void operator++();
  void operator--();
  void operator+=(int val);
  void operator-=(int val);
//other relative displacements
  void beginOfLine(); //set(current_row)
  void newLine();     //set(current_row+1)

//absolute displacement
  void seek(int row);
  void seek(int row, int col);
  void tell(int& row, int& col) const;
  int row() const; 
  int col() const; 

//diagnostics
  bool isWithin() const;
  bool isNear() const;
  Pointer origin() const {return origin_;};
  int szLine() const {return szLine_;};
  int height() const {return height_;};
  int width() const {return szLine_/sizeof(Pixel);};

//checked pointer arithmetic
private:
  void advance(int howMany);
  int offset() const;

private:
  Pointer origin_;  //pointer to the image origin
  Pointer ptr_;    //pointer to the image pixel
  int szLine_;    //width of scanline (in pixels!!!)
  int height_;    //height of the image
};



///////////////////////////////////////////////////
// inline definitions of the img_access_guard template

//construction
template<class A> inline 
img_access_guard<A>::img_access_guard(img_data& img):
  origin_(static_cast<Pointer>(static_cast<void*>(img.bits()))), 
  ptr_(origin_), 
  szLine_(img.fmt().szLine()/img.fmt().szPixel()), 
  height_(img.fmt().height())
{
  assert(ptr_!=0);
  assert(sizeof(Pixel)==img.fmt().szPixel());
}

template<class A> inline 
img_access_guard<A>::img_access_guard(const img_data& img):
  origin_(static_cast<Pointer>(static_cast<const void*>(img.bits()))), 
  ptr_(origin_), 
  szLine_(img.fmt().szLine()/img.fmt().szPixel()), 
  height_(img.fmt().height())
{
  assert(ptr_!=0);
  assert(sizeof(Pixel)==img.fmt().szPixel());
}

//member template functions defined in class
//because of the buggy MS compiler
/*
template<class A> template<class A2> inline 
img_access_guard<A>::img_access_guard(
  const img_access_guard<A2>& other)
:
  origin_(other.origin_), ptr_(other.ptr_),
  szLine_(other.szLine_), height_(other.height_)
{
  assert(ptr_);
}

template<class A> template<class A2> inline 
const img_access_guard<A>& img_access_guard<A>::operator=(
  const img_access_guard<A2>& other)
{
  ptr_=other.ptr_; 
  assert(isNear());
  assert(origin_==other.origin_);
  assert(szLine_==other.szLine_);
  assert(height_==other.height_);
  return *this;
}
*/

//access
template<class A> inline 
typename img_access_guard<A>::Reference 
img_access_guard<A>::operator*() const{
  return *ptr();
}

template<class A> inline 
typename img_access_guard<A>::Pointer 
img_access_guard<A>::operator->() const{
  return ptr();
}
template<class A> inline 
typename img_access_guard<A>::Pointer 
img_access_guard<A>::ptr() const{
  assert(isWithin());
  return ptr_;
}

//relative displacement (horizontal)
template<class A> inline 
void img_access_guard<A>::operator++(){advance(1);}

template<class A> inline 
void img_access_guard<A>::operator+=(int howMany){
  advance(howMany);
}

template<class A> inline 
void img_access_guard<A>::operator--(){advance(-1);}

template<class A> inline 
void img_access_guard<A>::operator-=(int howMany){
  advance(-howMany);
}

//relative displacement (vertical)
template<class A> inline 
void img_access_guard<A>::incLine(){advance(szLine_);}

template<class A> inline 
void img_access_guard<A>::incLine(int howMany){
  advance(szLine_*howMany);
}

template<class A> inline 
void img_access_guard<A>::decLine(){advance(-szLine_);}

template<class A> inline 
void img_access_guard<A>::decLine(int howMany){
  advance(-szLine_*howMany);
}

//relative displacement (other)
template<class A> inline 
void img_access_guard<A>::beginOfLine(){
  advance(-col());
}
template<class A> inline 
void img_access_guard<A>::newLine(){
  beginOfLine(); incLine();
}



//absolute displacement
template<class A> inline 
void img_access_guard<A>::seek(int row){
  ptr_=origin_;
  advance(row*szLine_);
}
template<class A> inline 
void img_access_guard<A>::seek(int row, int col){
  seek(row);
  advance(col);
}
template<class A> inline 
int img_access_guard<A>::row() const{
  return offset()/szLine_;
}
template<class A> inline 
int img_access_guard<A>::col() const{
  return offset()%szLine_;
}
template<class A> inline 
void img_access_guard<A>::tell(int& r, int& c) const{
  r=row();
  c=col();
}

//diagnostics
template<class A> inline 
bool img_access_guard<A>::isWithin() const{
  if (ptr_<origin_) 
    return false;
  if (ptr_>origin()+szLine_*height_) 
    return false;
  return true;
}
template<class A> inline 
bool img_access_guard<A>::isNear() const{
  if (ptr_<origin_) 
    return false;
  if (ptr_>origin()+szLine_*height_+szLine_) 
    return false;
  return true;
}

//checked pointer arithmetic
template<class A> inline 
void img_access_guard<A>::advance(int howMany){
  ptr_+=howMany;
  assert(isNear());
}
template<class A> inline 
int img_access_guard<A>::offset() const {
  assert(isNear());
  return static_cast<int>(ptr_-origin_);
}




///////////////////////////////////////////////////
// inline definitions of the img_access_guard friends

template<class A1, class A2> inline 
bool operator==(
  const img_access_guard<A1>& lhs, 
  const img_access_guard<A2>& rhs)
{
  assert(lhs.origin()==rhs.origin());
  return lhs.ptr() == rhs.ptr();
}
/*
template<class A1, class A2> inline 
bool operator!=(
  const img_access_guard<A1>& lhs, 
  const img_access_guard<A2>& rhs)
{
  assert(lhs.origin()==rhs.origin());
  return lhs.ptr() != rhs.ptr();
}
*/
template<class A1, class A2> inline 
bool operator<(
  const img_access_guard<A1>& lhs, 
  const img_access_guard<A2>& rhs)
{
  assert(lhs.origin()==rhs.origin());
  return lhs.ptr() < rhs.ptr();
}
/*
template<class A1, class A2> inline 
bool operator>=(
  const img_access_guard<A1>& lhs, 
  const img_access_guard<A2>& rhs)
{
  assert(lhs.origin()==rhs.origin());
  return lhs.ptr() >= rhs.ptr();
}
*/
template<class A1, class A2> inline 
int operator-(
  const img_access_guard<A1>& lhs, 
  const img_access_guard<A2>& rhs)
{
  assert(lhs.origin()==rhs.origin());
  return static_cast<int>(lhs.ptr() - rhs.ptr());
}
template<class A1, class A2> inline 
int rowdiff(
  const img_access_guard<A1>& lhs, 
  const img_access_guard<A2>& rhs)
{
  assert(lhs.origin()==rhs.origin());
  return (lhs.row() - rhs.row());
}

template<class A1> inline 
img_access_guard<A1> operator+(
  const img_access_guard<A1>& op, int val)
{
  img_access_guard<A1> acc(op);
  acc+=val;
  return acc;
}
template<class A1> inline 
img_access_guard<A1> operator-(
  const img_access_guard<A1>& op, int val)
{
  return op+(-val);
}


#endif
