#ifndef img_pix_hpp
#define img_pix_hpp

#include <float.h>
#include <assert.h>

/******************************************************
*
** img_pix by Sinisa Segvic
*
** generic representations for pixels of
** 1, 2, 3, 4, 6 and 8 bytes
*
*******************************************************/

//concept Pixel
// must have:

// typedef PixelComb Combination 
// typedef Data      Data      //the type of encapsulated data
// typedef Scalar    Scalar    //the portable (exact) scalar type 
// typedef EffScalar EffScalar //the unportable (most efficient) scalar type
// MaxScalar;
// fpDivisor;

// sizeof() must match!

// Pixel()
// operator=(Pixel&)
// operator==(Pixel&)

// setScalar(int)
// setScalar(double)
// setSint(int)
// setUint(int)
// setRgb(r,g,b)

// Scalar scalar() const;
// int sint() const;
// int uint() const;
// int red()  const;
// int green() const;
// int blue() const;

// the most efficient access, 
// unportable across different pixels
// EffScalar effScalar() const; 
// void setEffScalar(EffScalar) const; 


class img_comb_scalar;
class img_comb_rgb;

namespace{
  inline int myround(double f){
    return static_cast<int>(f<0?f-.5:f+.5);
  }
  inline int myrounddiv(int val, int div){
    int val2=val<0?(val-div/2):(val+div/2);
    return val2/div;
  }
}



class img_pix_char{
  unsigned char value_;

public:
  typedef img_comb_scalar       Combination;
  typedef unsigned char Data;
  typedef int Scalar;
  typedef int EffScalar;
  enum {fpDivisor=1};
  enum {MaxScalar=255};

public:
  bool operator==(const img_pix_char& o) const{
    return value_==o.value_;
  }
  template <class Pixel>
  img_pix_char& operator=(const Pixel& other){
    setUint(other.uint());
    return *this;
  }

public:
  void setScalar(int i){setSint(i);}
  void setScalar(double f){setSint(myround(f));}
  void setEffScalar(EffScalar s) {setScalar(s);}
  void setSint(int i){if (i<0) i=-i; setUint(i);}
  void setUint(int i){
    value_=static_cast<unsigned char>(i);
    assert (value_==i);
  }
  void setRgb(int r, int g, int b){
    setScalar((3+2*(r+g+b))/6);}
  void setRgb(double r, double g, double b){
    setScalar((r+g+b)/3.0);}

public:
  Scalar scalar() const {return uint();}
  EffScalar effScalar() const {return scalar();}
  int uint() const {return value_;}
  int sint() const {return uint();}
  Scalar red()   const {return scalar();}
  Scalar green() const {return scalar();}
  Scalar blue()  const {return scalar();}
};



class img_pix_fixed16{
  short value_;
public:
  typedef img_comb_scalar Combination;
  typedef short  Data;
  typedef double Scalar;
  typedef int    EffScalar;
  enum {fpDivisor=64};
  enum {MaxScalar=511};
public:
  bool operator==(const img_pix_fixed16& o) const{
    return value_==o.value_;
  }
  template <class Pixel> 
  img_pix_fixed16& operator=(const Pixel& other){
    setScalar(other.scalar());
    return *this;
  }
  img_pix_fixed16& operator=(const img_pix_fixed16& other){
    value_=other.value_;
    return *this;
  }

public:
  void setScalar(double f){
    value_=static_cast<short>(myround(f*fpDivisor));
  }
  void setScalar(int i){setSint(i);}
  void setEffScalar(EffScalar s) {value_=s;}
  void setSint(int s){value_=static_cast<short>(s*fpDivisor);}
  void setUint(int s){setSint(s);}
  template <class T>
  void setRgb(T r, T g, T b){
    setScalar((r+g+b)/3.0);}

public:
  Scalar scalar() const {
    return static_cast<double>(value_)/fpDivisor;
  }
  EffScalar effScalar() const {return value_;}
  int uint() const {return ((unsigned int)value_+fpDivisor/2)/fpDivisor;}
  int sint() const {return myrounddiv(value_,fpDivisor);}
  Scalar red() const {return scalar();}
  Scalar green() const {return scalar();}
  Scalar blue() const {return scalar();}
};



class img_pix_3char{
  struct Rgb{
    unsigned char r_;
    unsigned char g_;
    unsigned char b_;
  };
  Rgb d;
public:
  typedef img_comb_rgb   Combination;
  typedef img_pix_3char::Rgb  Data;
  typedef int            Scalar;
  typedef int            EffScalar;
  enum {fpDivisor=1};
  enum {MaxScalar=255};

public:
  bool operator==(const img_pix_3char& o) const{
    return d.r_==o.d.r_&&d.g_==o.d.g_&&d.b_==o.d.b_;
  }
  template <class Pixel>
  img_pix_3char& operator=(const Pixel& other){
    setRgb(other.red(), 
           other.green(), 
           other.blue());
    return *this;
  }
  img_pix_3char& operator=(const img_pix_3char& other){
    d=other.d;
    return *this;
  }

public:
  void setScalar(int i){setSint(i);}
  void setScalar(double f){setSint(myround(f));}
  void setEffScalar(EffScalar s) {setRgb(0,s,0);}
  void setSint(int i){if (i<0) i=-i; setUint(i);}
  void setUint(int i){setRgb(i,i,i);}
  template <class T>
  void setRgb(T r, T g, T b){
    d.r_=static_cast<unsigned char>(r); 
    d.g_=static_cast<unsigned char>(g); 
    d.b_=static_cast<unsigned char>(b);
  }

public:
  Scalar scalar() const {return uint();}
  EffScalar effScalar() const {return green();}
  int uint() const {return (2*(red()+green()+blue())+3)/6;}
  int sint() const {return uint();}
  Scalar red()   const {return d.r_;}
  Scalar green() const {return d.g_;}
  Scalar blue()  const {return d.b_;}
};



class img_pix_int32{
  //static_assert(sizeof(int)==4)
  int value_;
public:
  typedef img_comb_scalar    Combination;
  typedef int Data;
  typedef int Scalar;
  typedef int EffScalar;
  enum {fpDivisor=1};
  enum {MaxScalar=255};

public:
  bool operator==(const img_pix_int32& o) const{
    return value_==o.value_;
  }
  template <class Pixel>
  img_pix_int32& operator=(const Pixel& other){
    setScalar(other.scalar());
    return *this;
  }
  img_pix_int32& operator=(const img_pix_int32& other){
    value_=other.value_;
    return *this;
  }

public:
  void setScalar(int i){value_=i;}
  void setScalar(double f){value_=static_cast<int>(f);}
  void setEffScalar(EffScalar s) {value_=s;}
  void setSint(int i){setScalar(i);}
  void setUint(int i){setScalar(i);}
  void setRgb(int r, int g, int b){
    setScalar((3+2*(r+g+b))/6);}
  void setRgb(double r, double g, double b){
    setScalar((r+g+b)/3.0);}
  
public:
  Scalar scalar() const {return value_;}
  EffScalar effScalar() const {return value_;}
  int uint() const {return sint();}
  int sint() const {return static_cast<int>(value_+.5);}
  Scalar red()   const {return scalar();}
  Scalar green() const {return scalar();}
  Scalar blue()  const {return scalar();}
};


class img_pix_float32{
  //static_assert(sizeof(float)==4)
  float value_;
public:
  typedef img_comb_scalar Combination;
  typedef float Data;
  typedef float Scalar;
  typedef float EffScalar;
  enum {fpDivisor=1};
  enum{MaxScalar=255};

public:
  bool operator==(const img_pix_float32& o) const{
    return value_==o.value_;
  }
  template <class Pixel>
  img_pix_float32& operator=(const Pixel& other){
    setScalar(other.scalar());
    return *this;
  }
  img_pix_float32& operator=(const img_pix_float32& other){
    value_=other.value_;
    return *this;
  }

public:
  void setScalar(int i){value_=static_cast<float>(i);}
  void setScalar(double f){value_=static_cast<float>(f);}
  void setEffScalar(EffScalar s) {value_=s;}
  void setSint(int i){setScalar(i);}
  void setUint(int i){setScalar(i);}
  void setRgb(int r, int g, int b){
    setScalar((r+g+b)/3.0);}
  void setRgb(double r, double g, double b){
    setScalar((r+g+b)/3.0);}
  
public:
  Scalar scalar() const {return value_;}
  EffScalar effScalar() const {return value_;}
  int uint() const {return sint();}
  int sint() const {return static_cast<int>(value_+.5);}
  Scalar red()   const {return scalar();}
  Scalar green() const {return scalar();}
  Scalar blue()  const {return scalar();}
};


class img_pix_3fixed16{
  struct Rgb{
    img_pix_fixed16 r_;
    img_pix_fixed16 g_;
    img_pix_fixed16 b_;
  };
  Rgb d;

public:
  typedef img_comb_rgb        Combination;
  typedef img_pix_3fixed16::Rgb       Data;
  typedef img_pix_fixed16::Scalar    Scalar;
  typedef img_pix_fixed16::EffScalar EffScalar;
  enum {fpDivisor=img_pix_fixed16::fpDivisor};
  enum {MaxScalar=img_pix_fixed16::MaxScalar};

public:
  bool operator==(const img_pix_3fixed16& o) const{
    return d.r_==o.d.r_&&d.g_==o.d.g_&&d.b_==o.d.b_;
  }
  template <class Pixel> 
  img_pix_3fixed16& operator=(const Pixel& other){
    setRgb(other.red(), 
           other.green(), 
           other.blue());
    return *this;
  }
  img_pix_3fixed16& operator=(const img_pix_3fixed16& other){
    d=other.d;
    return *this;
  }

public:
  void setScalar(int i){setSint(i);}
  void setScalar(double f){setRgb(f,f,f);}
  void setEffScalar(EffScalar s) {setRgb(s,s,s);}
  void setSint(int i){setRgb(i,i,i);}
  void setUint(int i){setSint(i);}
  template <class T>
  void setRgb(T r, T g, T b){
    d.r_.setScalar(r); 
    d.g_.setScalar(g); 
    d.b_.setScalar(b); 
  }

public:
  Scalar scalar() const {return (red()+green()+blue())/3.0;}
  EffScalar effScalar() const {return d.g_.effScalar();}
  int sint() const {return myround(scalar());}
  int uint() const {return sint();}
  Scalar red()   const {return d.r_.scalar();}
  Scalar green() const {return d.g_.scalar();}
  Scalar blue()  const {return d.b_.scalar();}
};




class img_pix_float64{
  double value_;
public:
  typedef img_comb_scalar    Combination;
  typedef double             Data;
  typedef double             Scalar;
  typedef double             EffScalar;
  enum {fpDivisor=1};
  enum {MaxScalar=1};

public:
  bool operator==(const img_pix_float64& o) const{
    return value_==o.value_;
  }
  template <class Pixel>
  img_pix_float64& operator=(const Pixel& other){
    setScalar(other.scalar());
    return *this;
  }
  img_pix_float64& operator=(const img_pix_float64& other){
    value_=other.value_;
    return *this;
  }

public:
  void setScalar(int i){value_=i;}
  void setScalar(double f){value_=f;}
  void setEffScalar(EffScalar s) {value_=s;}
  void setSint(int i){setScalar(i);}
  void setUint(int i){setScalar(i);}
  template <class T>
  void setRgb(T r, T g, T b){
    setScalar((r+g+b)/3.0);
  }
  
public:
  Scalar scalar() const {return value_;}
  EffScalar effScalar() const {return value_;}
  int uint() const {return sint();}
  int sint() const {return static_cast<int>(value_+.5);}
  Scalar red()   const {return scalar();}
  Scalar green() const {return scalar();}
  Scalar blue()  const {return scalar();}
};




/////////////////////////////////
// simple black/gray/white logic

namespace{

template <class Pixel>
struct img_pix_values{
  Pixel black;
  Pixel gray;
  Pixel white;
private:
  img_pix_values(){
    black.setScalar(0);
    gray.setScalar(Pixel::MaxScalar/2);
    white.setScalar(Pixel::MaxScalar);
  }
public:
  static img_pix_values<Pixel>& instance(){
    static img_pix_values<Pixel> v;
    return v;
  }
};

template <class Pixel>
bool img_pix_isBlack(const Pixel& p){
  return p==img_pix_values<Pixel>::instance().black;
}
template <class Pixel>
bool img_pix_isGray(const Pixel& p){
  return p==img_pix_values<Pixel>::instance().gray;
}
template <class Pixel>
bool img_pix_isWhite(const Pixel& p){
  return p==img_pix_values<Pixel>::instance().white;
}

template <class Pixel>
void img_pix_black(Pixel& p){
  p=img_pix_values<Pixel>::instance().black;
}
template <class Pixel>
void img_pix_gray(Pixel& p){
  p=img_pix_values<Pixel>::instance().gray;
}
template <class Pixel>
void img_pix_white(Pixel& p){
  p=img_pix_values<Pixel>::instance().white;
}

}

#endif
