#include "img_data.hpp"


#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <iostream>
#include <sstream>
#include <stdexcept>

namespace{
  void throwif(bool exp, char const* s, int op){
    if (exp){
      std::ostringstream oss;
      oss <<"img_data:" <<s <<" (" <<op <<")";
      throw std::runtime_error(oss.str());
    }
  }

}

///////////////////////////////////
// img_fmt

img_fmt::img_fmt():
  width_(32), height_(32),
  szPixel_(1), 
  szLine_(32),
  type_(TDefault)
{
  init();
}

img_fmt::img_fmt(int w, int h, int szpix, Type t):
  width_(w), height_(h),
  szPixel_(szpix), 
  szLine_(szPixel_*width_),
  type_(t)
{
  init();
}
img_fmt::img_fmt(const img_fmt& f, int szpix, Type t):
  width_(f.width()), height_(f.height()),
  szPixel_(szpix), 
  szLine_(szPixel_*width_),
  type_(t)
{
  init();
}
void img_fmt::init(){
  throwif(szPixel_<1 || szPixel_==5 || 
         szPixel_==7 || szPixel_>8, 
    "img_fmt: unsupported pixel size", szPixel_);

  if (type_==TDefault){
    switch (szPixel_){
    case 1: type_=TSigned; break;
    case 2: type_=TFixed; break;
    case 3: type_=Type(TTriple+TSigned); break;
    case 4: type_=TFloat; break;
    case 6: type_=Type(TTriple+TFixed); break;
    case 8: type_=TFloat; break;
    }
  }
}

img_fmt img_fmt::makeFmtFloat() const{
  Type ttry=Type(42);
  Type t=TDefault;
  int sz=0;
  switch (szPixel()){
  case 1: t=TFixed; sz=2;break;
  case 2: t=TFixed; sz=2;break;
  case 3: t=Type(TTriple+TFixed); sz=6;break;
  case 4: 
    if (integral()){
      t=TFloat; sz=8;
    } else{
      t=TFloat; sz=4;
    }
    break;
  case 6: t=Type(TTriple+TFixed); sz=6;break;
  case 8: t=TFloat; sz=8; break;
  }
  return img_fmt(*this, sz, t);  
}

bool operator==(
  const img_fmt& lhs, 
  const img_fmt& rhs)
{
  if(
    lhs.width()  ==rhs.width() &&
    lhs.height() ==rhs.height() &&
    lhs.szPixel()==rhs.szPixel() &&
    lhs.szLine() ==rhs.szLine())
    return true;
  else
    return false;
}





///////////////////////////////////
// img_data construction

img_data::img_data():
  fmt_(img_fmt(32,32,1)),
  pBits_(0),
  allocated_(false)
{
  //reformat(fmt_);
}
img_data::img_data(const img_fmt& df):
  fmt_(df),
  pBits_(0),
  allocated_(false)
{
  reformat(fmt_);
}
img_data::img_data(const img_data& other):
  fmt_(other.fmt_),
  pBits_(0),
  allocated_(false)
{
  if (other.allocated_){
    reformat(fmt_);
    setBits(other.bits());
  }
}
img_data::img_data(const img_fmt& df, void* buf, int szbuf):
  fmt_(df),
  pBits_(static_cast<char*>(buf)),
  allocated_(false)
{
  if (szbuf){
    throwif(szbuf != fmt_.szBits(),
      "img_data() - illegal buffer size", szbuf);
  }
}

img_data::~img_data(){
  unalloc();
}


///////////////////////////////////
// img_data restructuring

const img_data& img_data::operator=(
  const img_data& other)
{
  reformat(other.fmt_);
  setBits(other.bits());
  return *this;
}

void img_data::reformat(
  const img_fmt& df)
{
  if (fmt_!=df || !pBits_){
    unalloc();
    fmt_=df;
    pBits_=new char[fmt_.szBits()];
    allocated_=true;
  }
}


////////////////////////////////////////////////
// attributes and operations

char* img_data::bits(){
  if (!pBits_){   
    pBits_=new char[fmt_.szBits()];
    allocated_=true;
  }
  assert(pBits_);
  return pBits_;
}
char const* img_data::bits() const{
  assert(pBits_);
  return pBits_;
}

void img_data::setLine(int line, void const* p){
  assert(pBits_);assert(p);
  void *address=pBits_+line*fmt_.szLine();
  memcpy(address, p, fmt_.szLine());
}
void img_data::setBits(void const* p){
  assert(pBits_);assert(p);
  memcpy(pBits_, p, fmt_.szBits());
}


////////////////////////////////////////////////
// private operations

void img_data::unalloc(){
  if (allocated_){
    delete [] pBits_;
    pBits_=0;
    allocated_=false;
  }
}

