#include "img_io_load.hpp"
#include "../img_data.hpp"
#include "../img_util.hpp"

#include "../../util/util_registry.hpp"
#include "../../util/util_bmp.hpp"

#include <assert.h>

#include <stdexcept>
#include <sstream>
#include <string>

/******************************************************
*
** img_io_load_bmp.cpp by Sinisa Segvic
*
** reading images from bmp file format
** to a img_data structure
*
** important properties: 
** - 16bpp imapges are stored in little endian
** - unlike BMP format, we use rgbrgbrgb for 24bpp images 
** - bmp maps scanlines in a bottom-up manner
*
** depends on:
**   img_data, img_io_load, img_util, 
**   util_bmp 
*
*******************************************************/

class img_io_load_bmp:
  public img_io_load
{
public:
  img_io_load_bmp(const CtorArg& dummy);
  virtual ~img_io_load_bmp();

  static char const* s_name(){return "bmp";}
  char const* name(){return s_name();}
  
  virtual void operator()(char const* name, img_data&);
};




////////////////////////////////////
// static registry creator object

namespace{
  util::RegistryCreator<img_io_load,img_io_load_bmp> creator;
} // unnamed namespace end



////////////////////////////////////
// img_io_load_bmp

img_io_load_bmp::img_io_load_bmp(const CtorArg&){}
img_io_load_bmp::~img_io_load_bmp(){}


void img_io_load_bmp::operator()(
  char const* name, 
  img_data& img)
{
  ReadFile rf(name, "img_io_load_bmp");

  //read header
  util_bmp_BFH bfh;
  rf.read(&bfh, sizeof(bfh));
  bfh.check();
  int sz_bi=bfh.sz_bi();
  int sz_ct=(sz_bi-sizeof(util_bmp_BIH))/sizeof(util_bmp_RgbUnion);
  if (sz_bi>static_cast<int>(sizeof(util_bmp_BI)))
    throw std::runtime_error("img_io_load_bmp::init "
        "- unsupported file format");

  //read BI 
  util_bmp_BI bi;
  bi.setNColours(sz_ct);
  //read bitmapinfoheader
  rf.read(&bi, sizeof(util_bmp_BIH));
  //read palette and position the file pointer to the beginning of bits
  if (sz_ct){
    rf.read(bi.palette(), sz_ct*sizeof(util_bmp_RgbUnion));
  }
  bi.check();

  //read bits
  int szpix=bi.bpp()/8;
  int padding=4*((bi.w()*szpix+3)/4)-bi.w()*szpix;
  switch (szpix){
  case 1: 
  case 2:
  case 3:
    img.reformat(img_fmt(bi.w(), bi.h(), szpix));
    for (int i=0; i<bi.h(); ++i){
      rf.read(img.bits()+img.fmt().szLine()*i, img.fmt().szLine());
      rf.advance(padding);
    }
    break;
  case 4:
    img.reformat(img_fmt(bi.w(), bi.h(), 3));
    for (int i=0; i<bi.h(); ++i){
      for (int j=0; j<bi.w(); ++j){
        rf.read(img.bits()+img.fmt().szLine()*i+j*3, 3);
        rf.advance(1);
      }
      rf.advance(padding);
    }
    break;
  default:
    throw std::runtime_error("img_io_load_bmp::init "
          "- unsupported file format");
  }
  
  //depaletisation, endianness
  switch (szpix){
  case 1: 
    if (sz_ct){
      assert(sz_ct==256);
      unsigned char bp[256];
      bi.greyScalePalette(bp);
      img_util::depaletise(img, bp, sz_ct);
    }
    break;
  case 2:
    if (util_endian::endianBig())
      img_util::flipEndian2(img);
    break;
  case 3:
  case 4:
    img_util::flipEndian(img);
    break;
  }
}
