#include "img_io_load_ras.hpp"
#include "img_io_load.hpp"
#include "../img_data.hpp"
#include "../img_util.hpp"
#include "../../util/util_ras.hpp"
#include "../../util/util_registry.hpp"

#include <assert.h>

#include <stdexcept>
#include <sstream>
#include <string>

//////////////////////////////
// implementation data

class img_io_load_rasImp{
public:
  img_io_load_rasImp(char const* name);
 
  int count(){return static_cast<int>(szFile_/szFrame_);}  
  const img_fmt& fmt(){return fmt_;}

  void seek(int frame);
  void doit(img_data&);
private:
  void skipHdr();
private:
  enum {ctCols=256};
  int sz_pal;
  bool isGray;
  
  unsigned char pal_red[ctCols];
  unsigned char pal_blue[ctCols];
  unsigned char pal_green[ctCols];

  img_fmt fmt_;
  img_io_load::ReadFile rf;

  bool firstTime_;
  size_t szFile_;
  int szFrame_;
  int szData_;
};


img_io_load_rasImp::img_io_load_rasImp(char const* name):
  fmt_(10,10,1),
  rf(name, "img_io_load_ras"),
  firstTime_(true),
  szFile_(rf.size())
{ 
  //read header
  util_ras_hdr rfh;
  rf.read(&rfh, sizeof(rfh));
  rfh.check();

  //read palette and position the file pointer 
  //to the beginning of image data
  sz_pal=0;
  if (rfh.has_palette()){
    assert (rfh.maplen()==3*ctCols);
    unsigned char palette[3*ctCols];
    rf.read((char*)palette, 3*ctCols);
    isGray=true;

    //calc colour palette
    unsigned char *pr=palette;
    unsigned char *pg=palette+ctCols;
    unsigned char *pb=palette+ctCols*2;

    for (int i=0; i<ctCols; i++){
      pal_red[i]=*pr++;
      pal_green[i]=*pg++;
      pal_blue[i]=*pb++;

	    if ( !(
  		  pal_red[i]   == pal_green[i] &&
	  	  pal_green[i] == pal_blue[i]))
  	  {
	  	  isGray=false;
	    }
	  }
    sz_pal=3*ctCols;
  }

  //determine image format
  int szpix=rfh.bpp()/8;
  if (sz_pal && !isGray)
    szpix=3;
  fmt_=img_fmt(rfh.w(), rfh.h(), szpix);

  //calc szData_ and szFrame_
  szData_=rfh.size();
  szFrame_=sizeof(rfh)+rfh.size();
  if (rfh.has_palette())
    szFrame_+=rfh.maplen();
}

void img_io_load_rasImp::skipHdr(){
  unsigned char buf[3*ctCols+sizeof(util_ras_hdr)];
  int count=sz_pal?sizeof(buf):sizeof(util_ras_hdr);
  rf.read(buf, count);
}


void img_io_load_rasImp::seek(int frame){
  firstTime_=false;
  size_t fpos=frame*szFrame_;
  rf.seek(fpos);
}
void img_io_load_rasImp::doit(img_data& img)
{
  if (firstTime_){
    firstTime_=false;
  } else{
    skipHdr();
  }
  img.reformat(fmt_);

  //read bits
  rf.read(img.bits(), szData_);

  //depaletisation
  if (sz_pal){
    if (isGray){
      assert(szData_==img.fmt().szBits());
      img_util::depaletise(img, pal_green, sz_pal);
    } else{
      assert(3*szData_==img.fmt().szBits());
      boost::scoped_array<char> buf(
        new char[img.fmt().width()*img.fmt().height()]);
      memcpy(buf.get(), img.bits(), szData_);
      img_util::depaletise(buf.get(), 
        pal_red, pal_green, pal_blue, 
        sz_pal, img);
    }
  } else{
    assert(szData_==img.fmt().szBits());
  }

  //16bpp imapges are stored in little endian
  //we use rgbrgbrgb for 24bpp images
  //we use rgbxrgbxrgbx for 32bpp images
  switch (img.fmt().szPixel()){
    case 1: 
      break;
    case 2:
      if (util_endian::endianBig())
        img_util::flipEndian2(img);
      break;
    case 3:
    case 4:
      img_util::flipEndian(img); 
      break;
    default:
      throw "img_io_load_ras::operator() - "
            "illegal szPixel()";
  }
  //sun raster maps scanlines in a top-down manner
  img_util::flipVertical(img);

}


//////////////////////////////////
// img_io_load_rasMany


img_io_load_rasMany::img_io_load_rasMany(char const* name):
  pimp(new img_io_load_rasImp(name))
{
}
img_io_load_rasMany::~img_io_load_rasMany(){}


int img_io_load_rasMany::count(){
  return pimp->count();
}
const img_fmt& img_io_load_rasMany::fmt(){
  return pimp->fmt();
}

void img_io_load_rasMany::seek(int frame){
  pimp->seek(frame);
}


void img_io_load_rasMany::operator()(
  img_data& img)
{
  pimp->doit(img);
}


//////////////////////////////////
// img_io_load_ras

class img_io_load_ras:
  public img_io_load
{
public:
  img_io_load_ras(const CtorArg& dummy);
  virtual ~img_io_load_ras();

  static char const* s_name(){return "ras";}
  char const* name(){return s_name();}
  
  virtual void operator()(char const* name, img_data&);
};


namespace{
  util::RegistryCreator<img_io_load,img_io_load_ras> creator;
} // unnamed namespace end


img_io_load_ras::img_io_load_ras(const CtorArg&) {}
img_io_load_ras::~img_io_load_ras(){}


void img_io_load_ras::operator()(
  char const* name, 
  img_data& img)
{
  img_io_load_rasImp imp(name);
  imp.doit(img);
}

