#include "ipp_equalize.hpp"
#include "ipp_range.hpp"
#include "ipp_loop.hpp"

#include "../img/img_data.hpp"
#include "../img/img_util.hpp"
#include "../img/access/img_access_dispatch.hpp"
#include "../img/access/img_access_guard.hpp"
#include "../img/access/img_access_wrap.hpp"
#include "../img/access/img_access_pix.hpp"
#include "../img/access/img_access_comb.hpp"


#include "../util/util_cxx.hpp"

#include <vector>
#include <iostream>




//////////////////////////////////////////////
// the worker classes
// * provide function: process
// * supposed to be called from loop2

template <class PixelSrc, class PixelDst>
class EqualizeWorker{
  const EqualizeWorker& operator=(const EqualizeWorker&);
public:
  typedef img_access_guard<img_access_wrap_const<PixelSrc> > AccessorSrc;
  typedef img_access_guard<img_access_wrap<PixelDst> > AccessorDst;
private:
  PixelSrc offset_;
  double factor_;
  double minval_;
  double maxval_;
  int maxscalar_;
public:
  EqualizeWorker(double minv, double maxv, int maxs):
    minval_(minv), maxval_(maxv), maxscalar_(maxs)
  {
    if (maxval_<=minval_)
      throw std::runtime_error("EqualizeWorker - maxval_<=minval_");
    if (maxscalar_==-1)
      maxscalar_=PixelDst::MaxScalar;
    offset_.setScalar(minval_);
    factor_=maxscalar_;
    factor_/=(maxval_-minval_);
    //std::cout <<"equalize: (min,max,f)= (" 
    //  <<minval_ <<"," <<maxval_ <<"," <<factor_ <<")\n";
  }
  void process(const AccessorSrc& psrc, const AccessorDst& pdst){
    //std::cout <<psrc.row() <<" " <<psrc.col()  <<" ";
    //std::cout <<psrc->scalar() <<" ";
    if (psrc->scalar()<minval_){
      pdst->setScalar(0);
    } else if (psrc->scalar()>maxval_){
      pdst->setScalar(maxscalar_);
    } else{
      typename PixelSrc::Combination comb;
      comb.add(*psrc,factor_);
      //std::cout <<comb.value() <<" ";
      comb.add(offset_,-factor_);
      //std::cout <<comb.value() <<" ";
      comb.put(*pdst);
    }
    //std::cout <<pdst->scalar() <<" ";
    //std::cout <<"\n";
  }
};



///////////////////////////////////////////////
// classes for marshalling parameters
// * typedefs are used by dispatch template functions

class EqualizeCaller{
  double minval_;
  double maxval_;
  int maxscalar_;

  const EqualizeCaller& operator=(const EqualizeCaller&);
public:
  typedef const img_data& Param1;
  typedef img_data& Param2;

  EqualizeCaller(double minv, double maxv, int maxs):
    minval_(minv), maxval_(maxv), maxscalar_(maxs) {}

public:
  template <class PixelSrc, class PixelDst>
  inline int operator() (
    const img_data& src, 
    img_data& dst, 
    const PixelSrc& dummy1,
    const PixelDst& dummy2)
  {
    EqualizeWorker<PixelSrc,PixelDst> w(minval_,maxval_, maxscalar_);
    loop2(src, dst, 0, w, dummy1,dummy2);
    return 0;
  }
};


////////////////////////////////
// entry functions

// doesn't work for rgb images!!!!!!!!!

void ipp::equalize(
  const img_data& src, 
  double minval, 
  double maxval,
  int maxscalar,
  img_data& dst)
{
  if (minval==-1 || maxval==-1){
    range(src, minval,maxval);
  }
  if (minval==maxval){
    img_util::fillBits(dst,0);
  } else{
    EqualizeCaller c(minval, maxval, maxscalar);
    img_access_dispatch2Scalars(src,dst,c);
  }
}

void ipp::equalize(
  const img_data& src, 
  int maxscalar,
  img_data& dst)
{
  equalize(src, -1,-1,maxscalar, dst);
}


void ipp::equalize(
  const img_data& src, 
  img_data& dst)
{
  equalize(src, -1, dst);
}

