#include "ipp_histogram.hpp"
#include "ipp_loop.hpp"

#include "../img/img_data.hpp"
#include "../img/access/img_access_dispatch.hpp"
#include "../img/access/img_access_guard.hpp"
#include "../img/access/img_access_wrap.hpp"
#include "../img/access/img_access_pix.hpp"
#include "../img/access/img_access_comb.hpp"


#include "../util/util_cxx.hpp"

#include <vector>




//////////////////////////////////////////////
// the worker classes
// * provide function: process
// * supposed to be called from loop

template <class PixelSrc>
class HistogramWorker{
  const HistogramWorker& operator=(const HistogramWorker&);
public:
  typedef img_access_guard<img_access_wrap_const
    <PixelSrc> > AccessorSrc;
  typedef img_pix_char PixelMask;
  typedef img_access_guard<img_access_wrap_const
    <img_pix_char> > AccessorMask;
private:
  std::vector<int>& hist_;
public:
  HistogramWorker(std::vector<int>&v): hist_(v){
    hist_.assign(PixelSrc::MaxScalar+1,0);
  }
  void process(const AccessorSrc& psrc){
    int value=psrc->uint();
    assert(value>=0 && value<=PixelSrc::MaxScalar);
    hist_[value]++;
  }
  void process(const AccessorSrc& psrc, const AccessorMask& pmask){
    if (pmask->scalar()!=0) 
      process(psrc);
  }
};



///////////////////////////////////////////////
// classes for marshalling parameters
// * member template functions defined in-class
//   because of the buggy MS compiler
// * dummies needed because the buggy MSVC supports 
//   only implicit resolution of template functions
// * typedefs are used by dispatch template functions

class HistogramCaller{
  int margin_;
  std::vector<int>& hist_;

  const HistogramCaller& operator=(const HistogramCaller&);
public:
  typedef const img_data& Param1;

  HistogramCaller(int m, std::vector<int>&v):
    margin_(m), hist_(v) {}

public:
  template <class PixelSrc>
  inline int operator() (
    const img_data& src, 
    const PixelSrc& dummy1)
  {
    HistogramWorker<PixelSrc> w(hist_);
    loop1(src, margin_, w, dummy1);
    return 0;
  }
};

class HistogramMaskCaller{
  const img_data& mask_;
  int margin_;
  std::vector<int>& hist_;

  const HistogramMaskCaller& operator=(const HistogramMaskCaller&);
public:
  typedef const img_data& Param1;

  HistogramMaskCaller(const img_data& m, int d, std::vector<int>&v):
    mask_(m), margin_(d), hist_(v) {}

public:
  template <class PixelSrc>
  inline int operator() (
    const img_data& src, 
    const PixelSrc& dummy1)
  {
    HistogramWorker<PixelSrc> w(hist_);
    loop1m(src, mask_, margin_, w, dummy1);
    return 0;
  }
};

////////////////////////////////
// entry functions

// construction
void ipp::histogram(
  const img_data& src, 
  int margin,
  std::vector<int>& hist)
{
  HistogramCaller c(margin,hist);
  img_access_dispatch1(src,c);
}

void ipp::histogram(
  const img_data& src, 
  std::vector<int>& hist)
{
  HistogramCaller c(0,hist);
  img_access_dispatch1(src,c);
}

void ipp::histogram(
  const img_data& src, 
  const img_data& mask, 
  int margin,
  std::vector<int>& hist)
{
  HistogramMaskCaller c(mask,margin,hist);
  img_access_dispatch1(src,c);
}




// access
int ipp::getHistPos(
  const std::vector<int>& hist, 
  double p)
{
  //build cumulative histogram
  int sum=0;
  std::vector<int> acc(hist.size());
  {for (unsigned i=0; i<hist.size(); i++){
    sum+=hist[i];
    acc[i]=sum;
  }}
  //find out the limit index
  unsigned i;
  int limit=static_cast<int>(p*sum+.5);
  for (i=0; i<hist.size(); i++){
    if (acc[i]>=limit)
      break;
  }
  assert(i<hist.size());

  return i;
}


