#include "ipp_regionGrowing.hpp"
#include "ipp_loop.hpp"

#include "../img/img_data.hpp"
#include "../img/img_util.hpp"
#include "../img/access/img_access_pix.hpp"
#include "../img/access/img_access_wrap.hpp"
#include "../img/access/img_access_guard.hpp"
#include "../img/access/img_access_dispatch.hpp"

#include "../util/util_debug.hpp"

typedef grid::WtRegion Blob;

namespace{
  //log file for the whole component
  static util::NoLog log_("/tmp/ipp_regionGrowing.cpp");

  //void testPoint(){
  //  static int i=0;
  //  i++;
  //}
}

template <class PixelSrc>
class RegionGrowing{
  typedef img_access_guard<img_access_wrap_const
    <PixelSrc> > AccessorSrc;
  typedef PixelSrc PixelDst;
  typedef img_access_guard<img_access_wrap
    <PixelDst> > AccessorDst;

  const img_data& imgSrc_;
  int thPix_;
  img_data& imgMark_;
  Blob blob_;

public:
  RegionGrowing(const img_data&s, int tp, img_data& imgMark);
public:
//level1
  int growBlob(const grid::Pixel& pos, Blob& b);

private:
//level1
  AccessorSrc scanLeft(AccessorSrc pos, AccessorSrc posStart);
  AccessorSrc scanRight(AccessorSrc pos, AccessorSrc posStop);
//level2
  void growBlobDir(AccessorSrc posBeg, AccessorSrc posEnd, 
      int row, int dir);
  AccessorSrc seekObjectRight(AccessorSrc pos, AccessorSrc posStop);
//level3
  AccessorSrc scanLeftWithCheckDir(
      AccessorSrc pos, AccessorSrc posStart, int row, int dir);
  AccessorSrc scanRightWithCheckDir(
      AccessorSrc pos, AccessorSrc posStop, int row, int dir);
//updateBlob members
  void updateBlobPixel(int value);
  int updateBlobLine(int colbeg, int colend, int row);
  void updateBlobBottomTop(int row1, int row2);
  int updateBlobFinal();

  

//utilities
public:
  AccessorDst toAccessorDst(AccessorSrc pos);
  bool wasHere(int row, int col);
  bool wasHereDst(AccessorDst pos);
  bool wasHereSrc(AccessorSrc pos);
  bool registerPixel(AccessorSrc pos);
  bool isObjAt(AccessorSrc pos);

//forbidden
private:
  RegionGrowing(const RegionGrowing&);
  const RegionGrowing &operator=(const RegionGrowing &);
};




///////////////////////////////////////////////////////
// construction
template <class PixelSrc>
RegionGrowing<PixelSrc>::RegionGrowing(
  const img_data&s, int tp, 
  img_data& m)
:
  imgSrc_(s),
  thPix_(tp),
  imgMark_(m)
{
  imgMark_.reformat(imgSrc_.fmt());
  img_util::fillBits(imgMark_,0);
}

///////////////////////////////////////////////////////
// growBlob, level1
template <class PixelSrc>
int RegionGrowing<PixelSrc>::
growBlob(const grid::Pixel& from, Blob& b)
{
  blob_.reset();

  int row=from.y();
  int col=from.x();

  // log_ <<"Entering growBlob (" <<col <<"," <<row <<")\n";
  // log_.flush();

  AccessorSrc pos(imgSrc_);
  pos.seek(row, col);
  assert(isObjAt(pos));
  if (wasHereSrc(pos)) {
    b.reset();
    return 0;
  }

  // if (col==319){
  //   testPoint();
  // }

  //SimpleScanLine
  AccessorSrc posStart(pos); posStart.beginOfLine();
  AccessorSrc posStop(pos);  posStop.newLine();
  AccessorSrc posBeg=scanLeft(pos, posStart);
  AccessorSrc posEnd=scanRight(pos, posStop);

  //update blob...
  int colBeg=posBeg-posStart;
  int colEnd=posEnd-posStart;
  updateBlobLine(colBeg, colEnd, row);
  updateBlobBottomTop(row, row+1);

  // trying to go up and down...
  growBlobDir(posBeg, posEnd, row, 1);
  growBlobDir(posBeg, posEnd, row, -1);

  // log_ <<"Before updateBlobFinal \n";
  // log_ <<" Rect= " 
  //   <<blob_.left()  <<"," <<blob_.top()    <<"," 
  //   <<blob_.right() <<"," <<blob_.bottom() <<"\n";
  // log_ <<" Size= " << static_cast<int>(blob_.size()) <<"\n";

  int n=updateBlobFinal();
  b=blob_;
  return n;
}

//from pos, exclusive,
//till posStart, inclusive
//returns last valid pos
template <class PixelSrc>
typename RegionGrowing<PixelSrc>::AccessorSrc 
RegionGrowing<PixelSrc>::scanLeft(
  AccessorSrc pos, 
  AccessorSrc posStart)
{
  // we go one position left from the 
  // last known object pixel
  assert(isObjAt(pos));
  --pos;

  while (pos.ptr()>=posStart.ptr()){
    if (!isObjAt(pos)){
      ++pos;
      break;
    }
    if (!registerPixel(pos))
      break;
    --pos;
  } //pos
  return pos;
}




//from pos, inclusive,
//till posStop, Exclusive
//returns first non-valid pos
template <class PixelSrc>
typename RegionGrowing<PixelSrc>::AccessorSrc 
RegionGrowing<PixelSrc>::scanRight(
  AccessorSrc pos, 
  AccessorSrc posStop)
{
  while(pos<posStop){
    if (!isObjAt(pos))
      break;
    if(!registerPixel(pos))
      break;
    ++pos;
  }//pos
  return pos;
}

///////////////////////////////////////////////////////
// growBlob, level2

template <class PixelSrc>
void RegionGrowing<PixelSrc>::
growBlobDir(
  AccessorSrc posBeg, 
  AccessorSrc posEnd, 
  int firstrow, int dir)
{
  assert(dir==1 || dir==-1);

  int lastrow=firstrow;
  int npixels=0;
  firstrow+=dir;

  for (int row=firstrow; row>=0 && row<posBeg.height(); row+=dir){
    posBeg.incLine(dir); assert(posBeg.isWithin());
    posEnd.incLine(dir); assert(posEnd.isWithin());

    AccessorSrc posStart(posBeg); posStart.beginOfLine();
    AccessorSrc posStop(posStart); posStop.incLine();

    if (posStart<posBeg)
      --posBeg;
    if (posEnd<posStop)
      ++posEnd;

    AccessorSrc posLeft=seekObjectRight(posBeg, posEnd);
    if (posLeft==posEnd)
      break;//nema vise...
    if (wasHereSrc(posLeft))
      break;//vec skanirana linija
    AccessorSrc posRight=scanRight(posLeft, posEnd);
 
    if (posBeg<posLeft)
      posBeg=posLeft;
    else//posLeft==posBeg
      posBeg=scanLeftWithCheckDir(posLeft, posStart, row,-dir);
 
    AccessorSrc pos=posRight;
    if (pos==posEnd){
      pos=posEnd=scanRightWithCheckDir(posRight, posStop, row,-dir);
    } else{//does the blob fork at this line?
      AccessorSrc posMark=pos;
      while (1){
        pos=seekObjectRight(pos, posEnd);
        if (pos==posEnd || 
            wasHereSrc(pos))
          break;

        //yes, we have a fork at this point
        AccessorSrc posChkBeg=pos;
        pos=scanRight(posChkBeg, posEnd);
        if (pos==posEnd)
          pos=posEnd=scanRightWithCheckDir(pos, posStop, row,-dir);
        AccessorSrc posChkEnd=pos;

        int colBeg=posChkBeg-posStart;
        int colEnd=posChkEnd-posStart;
        npixels+=updateBlobLine(colBeg, colEnd, row);

        growBlobDir(posChkBeg, posChkEnd, row, dir);
      }
      posEnd=posMark;
    }

    int colBeg=posBeg-posStart;
    int colEnd=posEnd-posStart;
    npixels+=updateBlobLine(colBeg, colEnd, row);

    lastrow=row;
  }//row

  if (npixels>0){
    assert(
        dir==1  && lastrow>=firstrow ||
        dir==-1 && lastrow<=firstrow);
    updateBlobBottomTop(firstrow, lastrow);
  }
}

////////////////////////////////////////////////////
// region growing helpers... (level 2)

//from pos, inclusive,
//till posStop, Exclusive
//returns first object pos or NULL
template <class PixelSrc>
typename RegionGrowing<PixelSrc>::AccessorSrc 
RegionGrowing<PixelSrc>::seekObjectRight(
  AccessorSrc pos, 
  AccessorSrc posStop)
{
  AccessorSrc rv=posStop;

  while (pos<posStop){
    if (isObjAt(pos)){
      rv=pos;
      break;
    }
    ++pos;
  }//pos
  
  return rv;
}


///////////////////////////////////////////////////////
// growBlob, level3
//from pos, exclusive,
//till posStart, inclusive
//returns last valid pos
template <class PixelSrc>
typename RegionGrowing<PixelSrc>::AccessorSrc 
RegionGrowing<PixelSrc>::scanLeftWithCheckDir(
  AccessorSrc pos, 
  AccessorSrc posStart, 
  int row, int dir)
{
  assert(isObjAt(pos));
  assert(dir==1 || dir==-1);

  bool hadObject=false;

  AccessorSrc posChk=pos; posChk.incLine(dir);
  if (!posChk.isWithin())
    return scanLeft(pos, posStart);

  AccessorSrc posChkBeg(pos);
  AccessorSrc posChkEnd(pos);
  while (pos.ptr()>=posStart.ptr()){
    --pos;
    --posChk;

    if (isObjAt(posChk)){//nonconvex check...
      if (!hadObject){
        posChkEnd=pos;++posChkEnd;//exclusive
        hadObject=true;
      }
    }else{
      if (hadObject){
        posChkBeg=pos;++posChkBeg;
        growBlobDir(posChkBeg, posChkEnd, row, dir);
        hadObject=false;
        // has the growBlob invocation 
        // filled us as well?
        if (wasHereSrc(pos)){
          ++pos;
          break;
        }
      }
    }
    if (!isObjAt(pos)){
      ++pos;
      break;
    }
    registerPixel(pos);
  }//col
  if (hadObject){
    //up/dn difference
    growBlobDir(pos, posChkEnd, row, dir);
  }

  return pos;
}


//from pos, inclusive,
//till posStop, exclusive
//returns first non-valid pos
template <class PixelSrc>
typename RegionGrowing<PixelSrc>::AccessorSrc 
RegionGrowing<PixelSrc>::scanRightWithCheckDir(
  AccessorSrc pos, 
  AccessorSrc posStop,
  int row, int dir)
{
  if (!isObjAt(pos)) return pos;
  assert(dir==1 || dir==-1);

  bool hadObject=false;

  AccessorSrc posChk=pos; posChk.incLine(dir);
  if (!posChk.isWithin())
    return scanRight(pos, posStop);

  AccessorSrc posChkBeg(pos);
  AccessorSrc posChkEnd(pos);
  while (pos<posStop){
    if (isObjAt(posChk)){//nonconvex check...
      if (!hadObject){
        posChkBeg=pos;
        hadObject=true;
      }
    }else{
      if (hadObject){
        posChkEnd=pos;
        growBlobDir(posChkBeg, posChkEnd, row,dir);
        hadObject=false;
        // has the growBlob invocation 
        // filled us as well?
        if (wasHereSrc(pos)){
          --pos;
          break;
        }
      }
    }
    if (!isObjAt(pos))
      break;
    registerPixel(pos);

    ++pos;
    ++posChk;
  }//col
  if (hadObject){
    //up/dn difference
    growBlobDir(posChkBeg, pos, row, dir);
  }
  return pos;
}




///////////////////////////////////////////////////////
// utilities

template <class PixelSrc>
typename RegionGrowing<PixelSrc>::AccessorDst 
RegionGrowing<PixelSrc>::toAccessorDst(
  AccessorSrc pos)
{
  assert(imgMark_.fmt().szPixel()==
         imgSrc_.fmt().szPixel());
  AccessorDst pdst(imgMark_);
  pdst+=static_cast<int>(pos.ptr()-pos.origin());
  assert(pdst.isWithin());
  return pdst;
}

//template <class PixelSrc>
//inline bool RegionGrowing<PixelSrc>::
//wasHere(int row, int col){
//  AccessorDst pos(imgMark_);
//  pos.seek(row,col);
//  return wasHere(pos);
//}

template <class PixelSrc>
inline bool RegionGrowing<PixelSrc>::
wasHereDst(AccessorDst pos){
  return !img_pix_isBlack(*pos);
}
template <class PixelSrc>
inline bool RegionGrowing<PixelSrc>::
wasHereSrc(AccessorSrc pos){
  AccessorDst pdst=toAccessorDst(pos);
  return wasHereDst(pdst);
}

template <class PixelSrc>
inline bool RegionGrowing<PixelSrc>::
registerPixel(AccessorSrc pos){
  AccessorDst pdst=toAccessorDst(pos);
  if (wasHereDst(pdst))
    return false;

  updateBlobPixel(pos->sint());
  img_pix_white(*pdst);
  return true;
}
template <class PixelSrc>
inline bool RegionGrowing<PixelSrc>::
isObjAt(AccessorSrc pos){
  return pos->scalar()>=thPix_;
}
//////////////////////
//updateBlob members
template <class PixelSrc>
inline void RegionGrowing<PixelSrc>::
updateBlobPixel(int value){
  blob_.wtAdd(value);
}

template <class PixelSrc>
inline int RegionGrowing<PixelSrc>::
updateBlobLine(int colBeg, int colEnd, int row){
  for (int i=colBeg; i<colEnd; ++i){
    blob_.push_back(grid::Pixel(i,row));
  }

  blob_.updateLeft(colBeg);
  blob_.updateRight(colEnd);

  int n=colEnd-colBeg;
  int sumx=colBeg*n+n*(n-1)/2;
  int sumy=row*n;
  blob_.updatePos(sumx,sumy);
  return n;
}

template <class PixelSrc>
inline void RegionGrowing<PixelSrc>::
updateBlobBottomTop(int bottom, int top){
  blob_.updateBottom(bottom);
  blob_.updateTop(top);
}

template <class PixelSrc>
inline int RegionGrowing<PixelSrc>::
updateBlobFinal(){
  blob_.normalize();
  blob_.wtSet(static_cast<int>(blob_.size()));
  return static_cast<int>(blob_.size());
}

//////////////////////////////////////////////
// the worker class
// * provides function: process
// * supposed to be called from loop

template <class PixelSrc>
class GrowBlobsWorker{
private:
  int thBlob_;
  std::vector<Blob>& blobs_;
  RegionGrowing<PixelSrc> rg_;
public:
  typedef img_access_guard<img_access_wrap_const
    <PixelSrc> > AccessorSrc;
public:
  GrowBlobsWorker( const img_data& src, img_data& dst,
    int tg, int tb, std::vector<Blob>& b)
  :
    thBlob_(tb),blobs_(b), rg_(src, tg, dst)
  {
    blobs_.clear();
  }
  ~GrowBlobsWorker(){
  }
  void process(const AccessorSrc& psrc) {
    if (rg_.isObjAt(psrc) && !rg_.wasHereSrc(psrc)){
      grid::Pixel pt(psrc.col(),psrc.row());
      Blob blob;
      int sz=rg_.growBlob(pt,blob);
      if (sz>0 && sz>thBlob_){
        blobs_.push_back(blob);
      }
    }
  } 
};


///////////////////////////////////////////////
// classes for static dispatch and 
// marshalling parameters
// * member template functions defined in-class
//   because of the buggy MS compiler
// * dummies needed because the buggy MSVC supports 
//   only implicit resolution of template functions
// * typedefs are used by dispatch template functions

class GrowBlobCaller{
  int th_;
  const grid::Pixel& pos_;
  Blob& blob_;
  img_data imgDst_;
public:
  typedef const img_data& Param1;
  GrowBlobCaller(int t, const grid::Pixel& p, Blob& b):
    th_(t),pos_(p),blob_(b)
  {}
  template <class PixelSrc>
  inline int operator() (
    const img_data& imgSrc, 
    const PixelSrc& )
  {
    RegionGrowing<PixelSrc> w(imgSrc, th_,imgDst_);
    return w.growBlob(pos_,blob_);
  }
private:
  GrowBlobCaller(const GrowBlobCaller&);
  const GrowBlobCaller& operator=(const GrowBlobCaller&);
};

class GrowBlobsCaller{
  int margin_;
  img_data& imgDst_;
  int thGray_;
  int thBlob_;
  std::vector<Blob>& blobs_;
public:
  typedef const img_data& Param1;
  GrowBlobsCaller(
    int m, img_data& dst, int tg, int tb, std::vector<Blob>& b) :
      margin_(m), imgDst_(dst),thGray_(tg), thBlob_(tb), blobs_(b) {}
  template <class PixelSrc>
  inline int operator() (
    const img_data& imgSrc, 
    const PixelSrc& dummy1)
  {
    GrowBlobsWorker<PixelSrc> w(
        imgSrc, imgDst_, thGray_, thBlob_, blobs_);
    loop1(imgSrc, margin_, w, dummy1);
    return 0;
  }
private:
  GrowBlobsCaller(const GrowBlobsCaller&);
  const GrowBlobsCaller& operator=(const GrowBlobsCaller&);
};




/////////////////////////////
// entry functions
//
int ipp::growBlob(
  const img_data& imgGray, 
  int thGray, const grid::Pixel& pos,
  Blob& blob)
{
  GrowBlobCaller c(thGray,pos,blob);
  return img_access_dispatch1(imgGray, c);
}

int ipp::growBlobs(
  const img_data& imgGray, int margin, img_data& imgDst,
  int thGray, int thBlob, std::vector<Blob>& blobs)
{
  // log_ <<"Entering growBlobs\n";
  GrowBlobsCaller c(margin, imgDst,thGray,thBlob,blobs);
  img_access_dispatch1(imgGray, c);
  return static_cast<int>(blobs.size());
}

