#include "ipp_subsample.hpp"

#include "../img/img_data.hpp"
#include "../img/access/img_access_dispatch.hpp"
#include "../img/access/img_access_guard.hpp"
#include "../img/access/img_access_wrap.hpp"
#include "../img/access/img_access_pix.hpp"
#include "../img/access/img_access_comb.hpp"

#include <stdexcept>



///////////////////////////////////////////////
// local class SubsampleWorker
//   The parameterized class performing 
//   the subsampling operation
//   (parameter=kernel size)

template<class PixelSrc, class PixelDst, int SIZE>
class SubsampleWorker{
  const SubsampleWorker& operator=(const SubsampleWorker&);
public:  
  SubsampleWorker(){}
  void operator()(const img_data& src, img_data& img_dst);
};


template<class PixelSrc, class PixelDst, int SIZE>
void SubsampleWorker<PixelSrc, PixelDst, SIZE>::
operator()(const img_data& img_src, img_data& img_dst)
{
  //standard prolog
  typedef img_access_guard<img_access_wrap_const<PixelSrc> > AccessorSrc;
  typedef img_access_guard<img_access_wrap<PixelDst> > AccessorDst;

  int width =img_src.fmt().width()/SIZE;
  int height=img_src.fmt().height()/SIZE;
  img_dst.reformat(img_fmt(
    width, height, img_dst.fmt().szPixel()));

  //subsampling algorithm
  AccessorDst itDstRow(img_dst);

  for (int row=0; row<height; row++){
    AccessorDst itDst(itDstRow);
    for (int col=0; col<width; col++){
      typename PixelSrc::Combination comb;

      AccessorSrc itSrcRow(img_src);
      itSrcRow.seek(row*SIZE);
      for (int i=0; i<SIZE; i++){
        AccessorSrc itSrc(itSrcRow);
        itSrc+=(col*SIZE);
        for (int j=0; j<SIZE; j++){
          comb.add(*itSrc,1);
          ++itSrc;
        }
        itSrcRow.incLine();
      }
      comb.divide(SIZE*SIZE);
      comb.put(*itDst);
      ++itDst;
    }//cols dst
    itDstRow.incLine();
  }//rows dst
}



///////////////////////////////////////////////
// local class SubsampleWorkerGeneral
//   The suboptimal subsampling class 

template<class PixelSrc, class PixelDst>
class SubsampleWorkerGeneral{
  const SubsampleWorkerGeneral& operator=(const SubsampleWorkerGeneral&);
public:  
  SubsampleWorkerGeneral(){}
  void operator()(const img_data& src, img_data& img_dst, int size);
};


template<class PixelSrc, class PixelDst>
void SubsampleWorkerGeneral<PixelSrc, PixelDst>::
operator()(const img_data& img_src, img_data& img_dst, int size)
{
  //standard prolog
  typedef img_access_guard<img_access_wrap_const<PixelSrc> > AccessorSrc;
  typedef img_access_guard<img_access_wrap<PixelDst> > AccessorDst;

  int width =img_src.fmt().width()/size;
  int height=img_src.fmt().height()/size;
  if (
      width !=img_dst.fmt().width() ||
      height!=img_dst.fmt().height())
    img_dst.reformat(img_fmt(
      width, height, img_dst.fmt().szPixel()));

  //subsampling algorithm
  AccessorDst itDstRow(img_dst);

  for (int row=0; row<height; row++){
    AccessorDst itDst(itDstRow);
    for (int col=0; col<width; col++){
      typename PixelSrc::Combination comb;

      AccessorSrc itSrcRow(img_src);
      itSrcRow.seek(row*size);
      for (int i=0; i<size; i++){
        AccessorSrc itSrc(itSrcRow);
        itSrc+=(col*size);
        for (int j=0; j<size; j++){
          comb.add(*itSrc,1);
          ++itSrc;
        }
        itSrcRow.incLine();
      }
      comb.divide(size*size);
      comb.put(*itDst);
      ++itDst;
    }//cols dst
    itDstRow.incLine();
  }//rows dst
}


///////////////////////////////////////////////
// class SubsampleCaller

// member template functions defined in-class
// because of the buggy MS compiler

// typedefs are used by dispatch template functions

class SubsampleCaller{
  int dimension_;
public:
  typedef const img_data& Param1;
  typedef img_data& Param2;
public:
  SubsampleCaller(int d):
    dimension_(d) {}

public:
  template <class PixelSrc, class PixelDst>
  inline int operator() (
    const img_data& src, img_data& dst, 
    const PixelSrc& , const PixelDst& )
  {
    switch (dimension_){
      case 2:{
        SubsampleWorker<PixelSrc,PixelDst,2> ssw;
        ssw(src, dst);
        break;
      } case 3:{
        SubsampleWorker<PixelSrc,PixelDst,3> ssw;
        ssw(src, dst);
        break;
      } case 4:{
        SubsampleWorker<PixelSrc,PixelDst,4> ssw;
        ssw(src, dst);
        break;
      } default:{
        if (dimension_<2){
          throw std::runtime_error("SubsampleCaller::() "
              "- got unsuported scale");
        }
        SubsampleWorkerGeneral<PixelSrc,PixelDst> ssw;
        ssw(src, dst, dimension_);
      }
    }
    return 0;
  }
};





void ipp::subsample(
  const img_data& src, 
  int dimension,
  img_data& dst)
{
  SubsampleCaller ssc(dimension);
  img_access_dispatch2(src,dst, ssc);
}
  

