#ifndef thr_base_hpp
#define thr_base_hpp

/******************************************************
*
** thr_base.hpp by Sinisa Segvic:
*
** Threads
**   * ThreadSimple - interface for a simple thread,
**     whose termination is not explicitly requested
**   * ThreadSync - interface for a syncronous thread,
**     i.e. a thread which terminates on parent's 
**     request from object's destructor
**   * ThreadDeferred - an almost syncronous thread,
**     i.e. a thread which terminates soon after the 
**     parent's request from object's destructor
**   * both thread templates should be parameterized
**     with a class derived from ThreadWorker
*
** Lock, CondVar
**   exception safe synchronization objects
* 
** depends_on:
**   thr_primitives
*
*******************************************************/

#include "thr_primitives.hpp"

#include <boost/shared_ptr.hpp>

namespace thr{

  void* ThreadEntryFunction(void* pw);

  // joinable simple thread
  class ThreadSimpleWorker{
    friend void* ThreadEntryFunction(void* pw);
    bool done_;
  public:
    ThreadSimpleWorker();
    virtual ~ThreadSimpleWorker();
    virtual int operator()()=0;
    bool done(){return done_;};
  };

  class ThreadSimple{
  private:
    boost::shared_ptr<ThreadSimpleWorker> pw_;
  private:
    ThreadHandle* pth_;
  public:
    ThreadSimple(const boost::shared_ptr<ThreadSimpleWorker>& pw);
    ~ThreadSimple();

    ThreadSimpleWorker& worker(){return *pw_;}
    void join();
  };

  // synchronized threads
  class ThreadSyncWorker:
    public ThreadSimpleWorker{
  public:
    virtual ~ThreadSyncWorker();
    virtual int operator()()=0;
    virtual void cleanup()=0;
  };

  class ThreadSync:
    public ThreadSimple
  {
  public:
    ThreadSync(const boost::shared_ptr<ThreadSyncWorker>& pw);
    ~ThreadSync();
  };
  class ThreadDeferred:
    public ThreadSimple
  {
  public:
    ThreadDeferred(const boost::shared_ptr<ThreadSyncWorker>& pw);
    ~ThreadDeferred();
  };

  //mutex, lock and condition
  class Mutex{
    friend class Lock;
    friend class CondVar;
    MutexHandle* pimp;
    bool closed;
  public:
    Mutex();
    ~Mutex();
    void enter();
    void leave();
  };

  class Lock{
    Mutex& mtx_;
  public:
    Lock(Mutex& m);
    ~Lock();
  private:
    const Lock& operator=(const Lock& );
  };

  class CondVar{
    CondVarHandle* pcv;
  public:
    CondVar();
    ~CondVar();

    void wait(Mutex& m);
    bool timedWait(Mutex& m, int t_ms);

    void signal();
    void broadcast();
  };

} //end namespace

#endif

