#include "thr_primitives.hpp"

#include <unistd.h>

#include <sys/time.h>

#include <pthread.h>

#include <stdexcept>

#include <errno.h>
#include <assert.h>

/////////////////////////////////////////
// general

void thr::sleep(int ms){
  // usleep(1000*ms);
  static pthread_cond_t      cond  = PTHREAD_COND_INITIALIZER;
  static pthread_mutex_t     mutex = PTHREAD_MUTEX_INITIALIZER;

  timeval tv;
  gettimeofday(&tv, NULL);

  timespec ts;
  const int million=1000*1000;

  int new_usec= tv.tv_usec+1000*ms;
  ts.tv_sec  = tv.tv_sec + new_usec/million;
  ts.tv_nsec = (new_usec % million) * 1000;

  pthread_mutex_lock(&mutex);
  pthread_cond_timedwait(&cond, &mutex, &ts);
  pthread_mutex_unlock(&mutex);
}





/////////////////////////////////////////
// threads
struct thr::ThreadHandle{
  pthread_t handle;
};

extern "C"{
  typedef void* (*PtrFunC)(void*);
}

thr::ThreadHandle* thr::threadCreate(
  PtrFunThread pf,
  void* arg,
  ThreadCreationParam tcp)
{
  ThreadHandle* pth=new ThreadHandle;

  pthread_attr_t attr;
  /* initialize attr with default attributes */
  pthread_attr_init(&attr); 
  if (tcp==Detached)
    pthread_attr_setdetachstate(&attr, 
      PTHREAD_CREATE_DETACHED);
  else
    pthread_attr_setdetachstate(&attr, 
      PTHREAD_CREATE_JOINABLE);

  //A kludge needed for calling extern "C" function 
  //(pthread_create) with a C++ function
  int rv=pthread_create(&(pth->handle), &attr, 
    reinterpret_cast<PtrFunC>(pf), arg);
  if (rv!=0){
    delete pth;
    throw std::runtime_error("thr::create - could not create thread.");
  }
  return pth;
}


void thr::threadRelease(ThreadHandle* pth){
  delete pth;
}

void thr::threadExit(int value){
  static int rv=value;
  pthread_exit(&rv);
}


void thr::threadJoin(ThreadHandle* other){
  int rv=pthread_join(other->handle, 0);
  if (rv!=0)
    throw std::runtime_error("thr::threadJoin - error in pthread_join.");
}
void thr::threadDetach(ThreadHandle* other){
  int rv=pthread_detach(other->handle);
  if (rv!=0)
    throw std::runtime_error("thr::threadDetach - error in pthread_detach.");
}







/////////////////////////////////////////
// mutex

struct thr::MutexHandle{
  pthread_mutex_t handle;
};

thr::MutexHandle* thr::mutexCreate(){
  MutexHandle* pm =new MutexHandle;
  int rv=pthread_mutex_init(&(pm->handle), 0);
  if (rv){
    delete pm;
    throw std::runtime_error("thr::mutexCreate - error in pthread_mutex_init.");
  }
  return pm;
}

void thr::mutexDestroy(MutexHandle *pm){
  pthread_mutex_destroy(&(pm->handle));
  delete pm;
}

void thr::mutexEnter(MutexHandle *pm){
  pthread_mutex_lock(&(pm->handle));
}
void thr::mutexLeave(MutexHandle *pm){
  pthread_mutex_unlock(&(pm->handle));
}




/////////////////////////////////////////
// condition variable
struct thr::CondVarHandle{
  pthread_cond_t handle;
};


thr::CondVarHandle* thr::condVarCreate(){
  CondVarHandle* pcv=new CondVarHandle;
  int rv=pthread_cond_init(&pcv->handle, 0);
  if (rv){
    delete pcv;
    throw "thr::condVarCreate - "
          "error in pthread_cond_init.";
  }
  return pcv;
}

void thr::condVarDestroy(CondVarHandle* pcv){
  int rv=pthread_cond_destroy(&pcv->handle);
  if (rv) 
    assert(0);
}


void thr::condVarWait(CondVarHandle* pcv, MutexHandle* pm){
  pthread_cond_wait(&pcv->handle, &pm->handle);
}
bool thr::condVarTimedWait(
  CondVarHandle* pcv, 
  MutexHandle* pm,
  int t_ms)
{
  struct timeval now;
  gettimeofday(&now, 0);
  int usec=now.tv_usec+1000*t_ms;

  struct timespec timeout;
  timeout.tv_sec = now.tv_sec + usec/1000000;
  timeout.tv_nsec = (usec%1000000) * 1000;

  int retcode = pthread_cond_timedwait(
    &pcv->handle, &pm->handle, &timeout);

  return (retcode != ETIMEDOUT);
}

void thr::condVarSignal(CondVarHandle* pcv){
  pthread_cond_signal(&pcv->handle);
}
void thr::condVarBroadcast(CondVarHandle* pcv){
  pthread_cond_broadcast(&pcv->handle);
}


