#include "thr_primitives.hpp"

#include <windows.h>
#include <process.h>
#undef max

#include <map>
#include <limits>
#include <stdexcept>

#include <assert.h>


/////////////////////////////////////////
// general

void thr::sleep(int t_ms){
  Sleep(t_ms);
}


/////////////////////////////////////////
// threads

struct thr::ThreadHandle{
  PtrFunThread pf;
  void* parg;
  HANDLE handle;
  DWORD id;
};

//scaffolding for dispatching the __stdcall
//call needed by BeginThread to the real thread 
//function provided by user which is __cdecl
namespace{
  DWORD WINAPI ThreadFunc( LPVOID lpParam ){
    thr::ThreadHandle* pth=(thr::ThreadHandle*) lpParam;
    pth->pf(pth->parg);
    return 0; 
  } 
}  //end unnamed namespace

//ThreadCreationParam is not used in this implementation 
//since win32 threads seem not to support this feature 
//(i.e, as if they are said to be allways joinable)
thr::ThreadHandle* thr::threadCreate(
  PtrFunThread pf,
  void* parg,
  ThreadCreationParam )
{
  ThreadHandle* pt=new ThreadHandle;
  pt->parg=parg;
  pt->pf=pf;
  //a dirty hack is needed since w32 require 
  //that a thread entry function be __stdcall
  //while the user function will by default be __cdecl
  pt->handle=CreateThread(0,0,
    ThreadFunc, pt,
    0, &pt->id);
  if (pt->handle==0)
    throw std::runtime_error("thr::createThread - could not create thread.");
  return pt;
}


void thr::threadRelease(ThreadHandle* pth){
  delete pth;
}

void thr::threadExit(int value){
  ExitThread(value);
}

void thr::threadJoin(thr::ThreadHandle* other){
  static unsigned long rv;
  rv=WaitForSingleObject(
    other->handle, INFINITE);
}

// it seems there is no way to accomplish this 
// with win32 threads
void thr::threadDetach(thr::ThreadHandle* ){
  return;
}







/////////////////////////////////////////
// mutex

struct thr::MutexHandle{
  HANDLE handle;
};

thr::MutexHandle* thr::mutexCreate(){
  MutexHandle* pm= new MutexHandle;
  pm->handle=CreateMutex(0, FALSE, 0);
  if (pm->handle == 0){
    delete pm;
    throw std::runtime_error("thr::mutexCreate - error in CreateMutex.");
  }
  return pm;
}

void thr::mutexDestroy(thr::MutexHandle* pm){
  CloseHandle(pm->handle);
  delete pm;
}


void thr::mutexEnter(thr::MutexHandle* pm){
  WaitForSingleObject(pm->handle, INFINITE);
}
void thr::mutexLeave(thr::MutexHandle* pm){
  ReleaseMutex(pm->handle);
}




/////////////////////////////////////////
// condition waiting

// adapted form boost/libs/thread/src/,
// files condition.hpp and condition.cpp
// Copyright (C) 2001, William E. Kempf

struct thr::CondVarHandle{
  HANDLE m_gate;
  HANDLE m_queue;
  HANDLE m_mutex;

  // # threads that timed out and never made it to the m_queue
  unsigned m_gone;         
  // # threads m_blocked m_waiting for the condition
  unsigned long m_blocked; 
  // # threads m_waiting no longer m_waiting for the condition but still
  //   m_waiting to be removed from the m_queue
  unsigned m_waiting; 
                    
};




thr::CondVarHandle* thr::condVarCreate(){
  CondVarHandle* pcv= new CondVarHandle;

  pcv->m_gone=0;
  pcv->m_blocked=0;
  pcv->m_waiting=0;

  pcv->m_gate  = CreateSemaphore(0, 1, 1, 0);
  pcv->m_queue = CreateSemaphore(0, 0, 
    std::numeric_limits<long>::max(), 0);
  pcv->m_mutex = CreateMutex(0, 0, 0);

  if (!pcv->m_gate || !pcv->m_queue || !pcv->m_mutex) {
    int res = 0;
    if (pcv->m_gate){
      res = CloseHandle(pcv->m_gate);
      assert(res);
    }
    if (pcv->m_queue){
      res = CloseHandle(pcv->m_queue);
      assert(res);
    }
    if (pcv->m_mutex){
      res = CloseHandle(pcv->m_mutex);
      assert(res);
    }
    throw std::runtime_error("thr::condVarCreate() - could not allocate resources");
  }
  return pcv;
}


void thr::condVarDestroy(CondVarHandle* pcv){
  int res=0;
  res = CloseHandle(pcv->m_gate);
  assert(res);
  res = CloseHandle(pcv->m_queue);
  assert(res);
  res = CloseHandle(pcv->m_mutex);
  assert(res);
}


void thr::condVarWait(
  CondVarHandle* pcv, 
  MutexHandle* pm)
{
  // enter_wait
  int res = 0;
  res = WaitForSingleObject(pcv->m_gate, INFINITE);
  assert(res == WAIT_OBJECT_0);
  ++pcv->m_blocked;
  res = ReleaseSemaphore(pcv->m_gate, 1, 0);
  assert(res);

  // unlock mutex
  mutexLeave(pm);

  // do_wait 
  res = 0;
  res = WaitForSingleObject(pcv->m_queue, INFINITE);
  assert(res == WAIT_OBJECT_0);

  unsigned was_waiting=0;
  unsigned was_gone=0;

  res = WaitForSingleObject(pcv->m_mutex, INFINITE);
  assert(res == WAIT_OBJECT_0);
  was_waiting = pcv->m_waiting;
  was_gone = pcv->m_gone;
  if (was_waiting != 0){
    if (--pcv->m_waiting == 0){
      if (pcv->m_blocked != 0){
        res = ReleaseSemaphore(pcv->m_gate, 1, 0); // open m_gate
        assert(res);
        was_waiting = 0;
      }
      else if (pcv->m_gone != 0)
        pcv->m_gone = 0;
    }
  }
  else if (++pcv->m_gone == (std::numeric_limits<unsigned>::max() / 2))
  {
    // timeout occured, normalize the m_gone count
    // this may occur if many calls to wait with a timeout are made and
    // no call to notify_* is made
    res = WaitForSingleObject(pcv->m_gate, INFINITE);
    assert(res == WAIT_OBJECT_0);
    pcv->m_blocked -= pcv->m_gone;
    res = ReleaseSemaphore(pcv->m_gate, 1, 0);
    assert(res);
    pcv->m_gone = 0;
  }
  res = ReleaseMutex(pcv->m_mutex);
  assert(res);

  if (was_waiting == 1){
    for (/**/ ; was_gone; --was_gone) {
      // better now than spurious later
      res = WaitForSingleObject(pcv->m_queue, INFINITE);
      assert(res == WAIT_OBJECT_0);
    }
    res = ReleaseSemaphore(pcv->m_gate, 1, 0);
    assert(res);
  }
  
  
  // reacquire mutex
  mutexEnter(pm);
}
bool thr::condVarTimedWait(
  CondVarHandle* pcv, 
  MutexHandle* pm,
  int t_ms)
{
  // enter_wait
  int res = 0;
  res = WaitForSingleObject(pcv->m_gate, INFINITE);
  assert(res == WAIT_OBJECT_0);
  ++pcv->m_blocked;
  res = ReleaseSemaphore(pcv->m_gate, 1, 0);
  assert(res);

  // unlock mutex
  mutexLeave(pm);

  // do_timed_wait 
  assert(t_ms>=0);
  unsigned milliseconds=t_ms;

  res = WaitForSingleObject(pcv->m_queue, milliseconds);
  assert(res != WAIT_FAILED && res != WAIT_ABANDONED);

  bool ret = (res == WAIT_OBJECT_0);

  unsigned was_waiting=0;
  unsigned was_gone=0;

  res = WaitForSingleObject(pcv->m_mutex, INFINITE);
  assert(res == WAIT_OBJECT_0);
  was_waiting = pcv->m_waiting;
  was_gone = pcv->m_gone;
  if (was_waiting != 0) {
    if (!ret) { // timeout
      if (pcv->m_blocked != 0)
        --pcv->m_blocked;
      else
        ++pcv->m_gone; // count spurious wakeups
    }
    if (--pcv->m_waiting == 0) {
      if (pcv->m_blocked != 0) {
        res = ReleaseSemaphore(pcv->m_gate, 1, 0); // open m_gate
        assert(res);
        was_waiting = 0;
      }
      else if (pcv->m_gone != 0)
        pcv->m_gone = 0;
    }
  }
  else if (++pcv->m_gone == (std::numeric_limits<unsigned>::max() / 2))
  {
    // timeout occured, normalize the m_gone count
    // this may occur if many calls to wait with a timeout are made and
    // no call to notify_* is made
    res = WaitForSingleObject(pcv->m_gate, INFINITE);
    assert(res == WAIT_OBJECT_0);
    pcv->m_blocked -= pcv->m_gone;
    res = ReleaseSemaphore(pcv->m_gate, 1, 0);
    assert(res);
    pcv->m_gone = 0;
  }
  res = ReleaseMutex(pcv->m_mutex);
  assert(res);

  if (was_waiting == 1) {
    for (/**/ ; was_gone; --was_gone) {
      // better now than spurious later
      res = WaitForSingleObject(pcv->m_queue, INFINITE);
      assert(res ==  WAIT_OBJECT_0);
    }
    res = ReleaseSemaphore(pcv->m_gate, 1, 0);
    assert(res);
  }

  return ret;
}


void thr::condVarSignal(CondVarHandle* pcv){
  //notify_one
  unsigned signals = 0;

  int res = 0;
  res = WaitForSingleObject(pcv->m_mutex, INFINITE);
  assert(res == WAIT_OBJECT_0);

  if (pcv->m_waiting != 0) { // the m_gate is already closed
    if (pcv->m_blocked == 0){
      res = ReleaseMutex(pcv->m_mutex);
      assert(res);
      return;
    }

    ++pcv->m_waiting;
    --pcv->m_blocked;
    signals = 1;
  } else {
    res = WaitForSingleObject(pcv->m_gate, INFINITE);
    assert(res == WAIT_OBJECT_0);

    if (pcv->m_blocked > pcv->m_gone) {
      if (pcv->m_gone != 0) {
          pcv->m_blocked -= pcv->m_gone;
          pcv->m_gone = 0;
      }
      signals = pcv->m_waiting = 1;
      --pcv->m_blocked;
    } else {
      res = ReleaseSemaphore(pcv->m_gate, 1, 0);
      assert(res);
    }
  }

  res = ReleaseMutex(pcv->m_mutex);
  assert(res);

  if (signals) {
    res = ReleaseSemaphore(pcv->m_queue, signals, 0);
    assert(res);
  }

}
void thr::condVarBroadcast(CondVarHandle* pcv){
  //notify_all
  unsigned signals = 0;

  int res = 0;
  res = WaitForSingleObject(pcv->m_mutex, INFINITE);
  assert(res == WAIT_OBJECT_0);

  if (pcv->m_waiting != 0) { // the m_gate is already closed
    if (pcv->m_blocked == 0) {
      res = ReleaseMutex(pcv->m_mutex);
      assert(res);
      return;
    }

    pcv->m_waiting += (signals = pcv->m_blocked);
    pcv->m_blocked = 0;
  } else {
    res = WaitForSingleObject(pcv->m_gate, INFINITE);
    assert(res == WAIT_OBJECT_0);
    if (pcv->m_blocked > pcv->m_gone) {
      if (pcv->m_gone != 0) {
        pcv->m_blocked -= pcv->m_gone;
        pcv->m_gone = 0;
      }
      signals = pcv->m_waiting = pcv->m_blocked;
      pcv->m_blocked = 0;
    } else {
      res = ReleaseSemaphore(pcv->m_gate, 1, 0);
      assert(res);
    }

    res = ReleaseMutex(pcv->m_mutex);
    assert(res);

    if (signals){
      res = ReleaseSemaphore(pcv->m_queue, signals, 0);
      assert(res);
    }
  }

}


