#include "ui.hpp"
#include "ui_cmd.hpp"
#include "ui_report.hpp"

#include "../cli/cli_dev.hpp"
#include "../cli/cli_parser.hpp"
#include "../cli/cli_wrap.hpp"

#include "../cvsh/cvsh_createAlg.hpp"
#include "../cvsh/cvsh_createVS.hpp"
#include "../cvsh/cvsh_createVD.hpp"
#include "../img/img_data.hpp"
#include "../img/img_vector.hpp"
#include "../vs/vs_base.hpp"
#include "../vd/vd_base.hpp"
#include "../vd/vd_win.hpp"
#include "../alg/alg_base.hpp"
#include "../alg/alg_decoratorScale.hpp"

#include "../util/util_time.hpp"
#include "../util/util_string.hpp"

#include "../thr/thr_primitives.hpp"

#include "../win/root/win_base.hpp"
#include "../win/root/win_wrap.hpp"
#include "../win/event/win_event.hpp"
#include "../win/ann/win_ann.hpp"

#include <sstream>
#include <iostream>
#include <stdexcept>

#include <assert.h>
#include <ctype.h>

/////////////////////////
// local functions 

namespace{
  void handleError(
    cli_wrapAbstract& cli, 
    char const*msg)
  {
    std::ostringstream oss;
    oss <<"Caught exception in UIloop:\n";
    std::string indented;
    util_string::indent(msg, indented);
    oss <<indented <<std::endl;
    oss <<"  Invoke 'help' for more information." 
        <<std::endl;
    cli.msg(oss.str().c_str());
  }
}

/////////////////////////////////
// private implementation class

class ui_imp{
  enum {maxWinSize=768};
public:
  //construction
  ui_imp(
    cli_wrap& cli,
    std::vector<boost::shared_ptr<vs_base> >& vecSpVs,
    boost::shared_ptr<alg_base>& spWorker,
    std::vector<boost::shared_ptr<vd_base> >& vecSpVd,
    int minDelayMs);
  ~ui_imp();

public:
  //operations
  void loop(char const*);
  void config(char const*);
  void batch(int repetitions, int offset);

private:
  enum BatchParameter{BpShow, BpProcess, BpBatch};
  void batch(alg_decoratorScale& worker, int repetitions, 
      int offset, BatchParameter par);
  void processCmd(ui_cmd& cmd);
private:
  void changeSource(ui_cmd& cmd);
  void changeDestination(ui_cmd& cmd);
  void changeAlgorithm(ui_cmd& cmd);
  void doConfig(ui_cmd& cmd);
  void configWindows();
  void configWinMain();
  void configWinAux(alg_base& alg);
  void showWinAux(alg_base& alg);

private:
  //data
  cli_wrap& cli_;
  std::vector<boost::shared_ptr<vs_base> >& vecSpVs_;
  alg_decoratorScale worker_;
  alg_decoratorScale algNop_;
  std::vector<boost::shared_ptr<vd_base> >& vecSpVd_;
  int indexVdWin_;
  vd_win* pVdWin_;
  win_base* pWin_;
  int minDelayMs_;
  
  std::vector<boost::shared_ptr<win_base> > vecSpWin_;
  
  bool cfgProcess_;
  bool cfgWnd_;
};

//////////////////////////////////////////
// ui_imp construction
ui_imp::ui_imp(
  cli_wrap& cli,
  std::vector<boost::shared_ptr<vs_base> >& vecSpVs, 
  boost::shared_ptr<alg_base>& spWorker,
  std::vector<boost::shared_ptr<vd_base> >& vecSpVd,
  int minDelayMs)
:
  cli_(cli), 
  vecSpVs_(vecSpVs), 
  worker_(spWorker, maxWinSize),
  algNop_(alg_base::create("nop",img_fmt(10,10,1)), maxWinSize),
  vecSpVd_(vecSpVd),
  indexVdWin_(-1),
  pVdWin_(0),
  pWin_(0),
  minDelayMs_(minDelayMs),
  cfgProcess_(true),
  cfgWnd_(true)
{
  if (vecSpVs_.size()==0)
    throw std::runtime_error(
      "ui_imp::ui_imp - no video sources were provided");
  if (vecSpVd_.size()==0)
    throw std::runtime_error(
      "ui_imp::ui_imp - no video destinations were provided");
      
  // do we have a vd_win?
  for (unsigned i=0; i<vecSpVd_.size(); i++){
    if (vecSpVd_[i]->isWin()){
      indexVdWin_=i;
      pVdWin_=dynamic_cast<vd_win*>(vecSpVd_[i].get());
      pWin_=&pVdWin_->win();
    }
  }
}

ui_imp::~ui_imp(){
}

//////////////////////////////////////////
// ui_imp public members

void ui_imp::loop(char const* command){
  bool isFirstCommand=true;
  ui_cmd cmd(command);
  while (1){
    try{
      // get the next command ...
      if (isFirstCommand){
        isFirstCommand=false;
      } else{
        std::ostringstream oss;
        cmd.prompt(oss, vecSpVs_[0]->name(), 
          vecSpVs_[0]->tell(), vecSpVs_[0]->size());
        cli_.put(oss.str().c_str());
        cmd.get(cli_);
      }

      //report the command back to user
      std::ostringstream oss;
      oss <<"UIloop: got command " <<cmd <<std::endl;
      cli_.msg(oss.str().c_str());

      // ... check whether we are not 
      // told to close the show ...
      if (cmd.action()==ui_cmd::Quit)
        break;

      // .. and finally process the command
      processCmd(cmd);

    } catch(char const* msg){
      handleError(cli_, msg);
    } catch(std::exception& e){
      handleError(cli_, e.what());
    }
  }
}



void ui_imp::batch(int repetitions, int offset){
  cfgProcess_=true;
  batch(worker_, repetitions, offset, BpBatch);
}

// BpProcess writes results to all vd-s
// BpShow    writes results only to vd_win-s
void ui_imp::batch(alg_decoratorScale& worker, 
  int repetitions, int offset,
  BatchParameter par)
{
  //std::ostringstream oss;
  //oss <<"UIbatch: nFrames=" <<repetitions 
  //    <<", offset=" <<offset <<std::endl;
  //cli_.msg(oss.str().c_str());

  cfgWnd_=true;

  bool writeVdNonWin=(par==BpProcess || par==BpBatch);
  bool seekAfter    =(offset>0 && par==BpBatch);    //seek next
  bool seekBefore   =(offset<0 || 
    (offset>0 && par!=BpBatch));

  //source images are temporarily stored here
  img_vector vecImgSrc(vecSpVs_.size());
  int fpos=0;

  //special profiling timer used...
  util_time timer;
  util_time::Data t_fetch=0;
  util_time::Data t_process=0;
  util_time::Data t_store=0;
  
  const int timerMsDivider=timer.frequency()/1000;
  const int tDelayMin=minDelayMs_*timerMsDivider;
  util_time::Data tFrame=timer();

  util_time::Data t0=util_time::msDayUtc();

  for (int i=0; i<repetitions; i++){
    util_time::Data tf0=timer();

    //fetch
    for (unsigned j=0; j<vecSpVs_.size(); j++){
      vs_base& vs_j(*vecSpVs_[j]);
      if (seekBefore)
        vs_j.seek(vs_j.tell()+offset);
      vs_j.getFrame(vecImgSrc[j]);
      if (j==0){
        fpos=vs_j.tell();
      }
      if (seekAfter)
        vs_j.seek(vs_j.tell()+offset);
    }
    int msDayUtc=vecSpVs_[0]->msDayUtc();
    util_time::Data tff=timer();

    //transfer input events and 
    //check whether we should bail out
    win_event_vector events;
    bool stop=false;
    bool vdwinToggleRec=false;
    if (pWin_){
      const int systemModifiers=
        win_event_abstract::Shift+win_event_abstract::Ctrl;
      static int winEventPos=0;
      win_event e;
      while (pWin_->event(winEventPos, e)){
        //std::cerr <<"Got event: "        <<e.type() <<" " 
        //          <<e.modifiers() <<" "  <<e.par1() <<"!\n";
        bool intercept=false;
        if (e.type()==win_event::KeyboardAscii){
          if (e.par1()==27){
            intercept=true;
            stop=true;
          }
          if (e.modifiers()==systemModifiers){
            intercept=true;
            if (toupper(e.par1())=='S'){
              vdwinToggleRec=true;
            }
          }
        }
        if (!intercept){
          events.push(e);
          //std::cerr <<"Pushed event!\n" ;
        }
      }
    }
    
    //process 
    worker.process(vecImgSrc, events, msDayUtc, fpos);
    util_time::Data tfp=timer();

    //configure windows
    if (cfgWnd_){
      cfgWnd_=false;
      configWinMain();
      configWinAux(worker);
    }
    
    //store
    for (unsigned j=0; j<vecSpVd_.size(); j++){
      //get and check vd
      vd_base& vd=*vecSpVd_[j].get();
      assert(&vd!=0);

      if (vd.isWin()) {
        vd_win& vdwin=dynamic_cast<vd_win&>(vd);
        // inform vdWin to toggle recording
        if (vdwinToggleRec){
          bool state=vdwin.toggleRec();
          char const* txtState=state?"on.\n":"off.\n";
          cli_.msg("Toggled recording ");
          cli_.msg(txtState);
          vdwinToggleRec=false;
        }
        // inform vdWin if we are starting 
        // a new batch of frames
        if (i==0) {
          vdwin.resetTime();
        }
        //show image and annotations
        win_ann_abstract const* pAnn=worker.pAnnDst(0);
        if (pAnn){
          vdwin(worker.imgDst(0), *pAnn, msDayUtc, vecSpVs_[0]->tell());
        } else{
          vdwin(worker.imgDst(0), msDayUtc, vecSpVs_[0]->tell());
        }
      } else{
        //store image if needed
        if (writeVdNonWin){
          vd(worker.imgDstOrig(0), msDayUtc, vecSpVs_[0]->tell());
        }
      }
    }
    //show additional windows
    showWinAux(worker);
    util_time::Data tfs=timer();

    //update profile data
    t_fetch+=tff-tf0;
    t_process+=tfp-tff;
    t_store+=tfs-tfp;

    if (stop)
      repetitions=i+1;

    int tPassed=tfs-tFrame;
    int tWait=(tDelayMin-tPassed)/timerMsDivider;
    if (tWait>0){
      // std::cerr <<tWait <<" " <<tDelayMin <<" "<<tPassed <<"\n";
      thr::sleep(tWait-1);
    }
    tFrame+=tDelayMin;
  }
  util_time::Data t=util_time::msDayUtc()-t0;

  if (par!=BpShow){
    //generate overall profile
    std::ostringstream oss;
    ui_report prTotal;
    prTotal.push_back("fetch", t_fetch/repetitions);
    prTotal.push_back("process", t_process/repetitions);
    prTotal.push_back("store", t_store/repetitions);
    ui_profile(oss,
      "Overall profile:", 
      prTotal, timer.frequency());

    //generate processing profile
    ui_report prProc;
    worker.profile(prProc);
    if (prProc.size()){
      ui_profile(oss, 
        "Last frame processing profile:", 
        prProc, timer.frequency());
    }

    //generate transfer report
    ui_reportBatch(oss, repetitions, t, 1000);

    cli_.msg(oss.str().c_str());
  }
}


void ui_imp::config(char const* str){
  if (*str){
    std::string tmp;
    util_string::fixBackslashEscapes(str, tmp);

    std::istringstream iss(tmp);
    cli_wrap cli(iss, std::cout, std::cout);
    worker_.config(cli);

    configWinAux(worker_);
    showWinAux(worker_);
  }
}



//////////////////////////////////////////
// ui_imp private members

void ui_imp::processCmd(ui_cmd& cmd){

  //since we are called only from interactive mode,
  //one vd_base-s should actually be a vd_win
  assert(pVdWin_!=0);

  //action is an ActionLoc?
  if (cmd.action()==ui_cmd::Show ||
      cmd.action()==ui_cmd::Process)
  {
    //find out the count of repetitions
    int repetitions=1;
    if (cmd.location()!=ui_cmd::Addr){
      repetitions=cmd.param();
      if (repetitions==-1)
        repetitions=vecSpVs_[0]->size()-1;
      if (repetitions==0)
        return;
    }

    //find out offset for the next frame
    int offset;
    switch (cmd.location()){
      case ui_cmd::Next: offset=1; break;
      case ui_cmd::Prev: offset=-1; break;
      case ui_cmd::This: offset=0; break;
      case ui_cmd::Addr: offset=0; break;
      default:
        offset=0; assert(0);
    }

    //seek to the start frame
    if (cmd.location()==ui_cmd::Addr){
      for (unsigned j=0; j<vecSpVs_.size(); j++){
        vecSpVs_[j]->seek(cmd.param());
      }
    }

    //configure windows
    bool doesProcessing=(cmd.action()==ui_cmd::Process);
    if (doesProcessing!=cfgProcess_){
      cfgProcess_=doesProcessing;
    }

    //do the job
    if (doesProcessing){
      batch(worker_, repetitions, offset, BpProcess);
    } else{ //ui_cmd::Show
      batch(algNop_, repetitions, offset, BpShow);
    }
  } else{
    switch (cmd.action()){
    //action is an ActionSimple?
    case ui_cmd::Help:
      {
        std::ostringstream oss;
        ui_usage(oss);
        cli_.msg(oss.str().c_str());
      }
      break;
    //action is an ActionStr?
    case ui_cmd::Config:
      doConfig(cmd);
      // batch(algNop_, 1, 0, BpShow);
      break;
    case ui_cmd::Source:
      //std:: cerr <<cmd <<"\n";
      changeSource(cmd);
      cfgProcess_=true;
      batch(algNop_, 1, 0, BpShow);
      break;
    case ui_cmd::Destination:
      //std:: cerr <<cmd <<"\n";
      changeDestination(cmd);
      //cfgProcess_=true;
      //batch(algNop_, 1, 0, BpShow);
      break;
    case ui_cmd::Algorithm:
      changeAlgorithm(cmd);
      cfgProcess_=true;
      batch(algNop_, 1, 0, BpShow);
      break;
    default:
      assert(0);
    }
  }
}



void ui_imp::changeAlgorithm(ui_cmd& cmd){
  std::ostringstream oss;
  boost::shared_ptr<alg_base> spWorkerNew
    (cvsh_createAlg(oss, cmd.str(), vecSpVs_[0]->fmt()));
  cli_.msg(oss.str().c_str());
  worker_.reset(spWorkerNew);
  vecSpWin_.resize(worker_.countDst()-1);
}


// multiple sources supported, eg:
// $ source f=file.srs 1
void ui_imp::changeSource(ui_cmd& cmd){
  int index=cmd.param();
  if (index<0 || index>=static_cast<int>(vecSpVs_.size()))
    throw std::runtime_error("ui_imp::changeSource - illegal video source index");

  std::vector<cli_dev> devs(1);
  cli_parser::parseSrc(cmd.str(), devs[0]);

  std::vector<boost::shared_ptr<vs_base> > vecSpVs;
  std::ostringstream oss;
  cvsh_createVS(oss, devs, vecSpVs);
  cli_.msg(oss.str().c_str());

  vecSpVs_[index]=vecSpVs[0];
}

// multiple destinations supported, eg:
// $ destination f=file.srs 1
void ui_imp::changeDestination(ui_cmd& cmd){
  int index=cmd.param();
  if (index<0 || index>=static_cast<int>(vecSpVs_.size()))
    throw std::runtime_error("ui_imp::changeDestination - illegal video source index");

  std::vector<cli_dev> devs(1);
  cli_parser::parseDst(cmd.str(), devs[0]);

  std::vector<boost::shared_ptr<vd_base> > vecSpVd;
  std::ostringstream oss;
  cvsh_createVD(oss, devs, vecSpVd_[index]->fmt(), vecSpVd);
  cli_.msg(oss.str().c_str());

  vecSpVd_[index]=vecSpVd[0];
}

void ui_imp::doConfig(ui_cmd& cmd){
  std::ostringstream oss;
  oss <<"Configuring worker object ";
  if (*cmd.str()){
    oss <<"from string: " <<cmd.str() <<".\n";
    cli_.msg(oss.str().c_str());

    config(cmd.str());
  } else{
    oss <<"from standard input.\n";
    cli_.msg(oss.str().c_str());

    cli_wrap cli(std::cin, std::cout, std::cout);
    worker_.config(cli);

    configWinAux(worker_);
    showWinAux(worker_);
  }
}


void ui_imp::configWindows(){
  configWinMain();
  configWinAux(worker_);
}

void ui_imp::configWinMain(){
  std::string shortname;
  util_string::retractPath(vecSpVs_[0]->name(), shortname);

  // config main window
  if (pVdWin_){
    std::ostringstream oss;
    oss <<shortname.c_str() <<" (" 
        <<(cfgProcess_?worker_.name():"source")
        <<":0" <<")";
    pVdWin_->caption(oss.str().c_str());
  }
}
  
void ui_imp::configWinAux(alg_base& alg){
  std::string shortname;
  util_string::retractPath(vecSpVs_[0]->name(), shortname);

  // config auxiliary windows
  int countThis=alg.countDst()-1;
  int countTotal=std::max(countThis, worker_.countDst()-1);
  vecSpWin_.resize(countTotal);

  for (int idWin=0; idWin<countThis; ++idWin){
    int idAlgImg=idWin+1;
    
    std::ostringstream oss;
    oss <<shortname.c_str() <<" (" 
        <<(cfgProcess_?alg.captionDst(idAlgImg):"source")
        <<":" <<idAlgImg <<")";
    if (vecSpWin_[idWin].get()==0){
      vecSpWin_[idWin].reset(new win_wrap(
        oss.str().c_str(),
        alg.fmtDst(idAlgImg).width(), 
        alg.fmtDst(idAlgImg).height()));
    } else{
      vecSpWin_[idWin]->caption(oss.str().c_str());
    }
  }
  for (int idWin=countThis; idWin<countTotal; ++idWin){
    std::ostringstream oss;
    oss <<shortname.c_str() <<" (inactive)";
    if (vecSpWin_[idWin].get()!=0){
      vecSpWin_[idWin]->caption(oss.str().c_str());
    }
  }
}

void ui_imp::showWinAux(alg_base& worker){
  int count=worker.countDst()-1;
  for (int idWin=0; idWin<count; ++idWin){
    int idAlgImg=idWin+1;
    win_ann_abstract const* pAnn=worker.pAnnDst(idAlgImg);
    if (pAnn){
      vecSpWin_[idWin]->setImgAnn(worker.imgDst(idAlgImg), *pAnn);
    } else{
      vecSpWin_[idWin]->setImg(worker.imgDst(idAlgImg));
    }
  }
}


///////////////////////////////////////
// ui delegates all the work to ui_imp

ui::ui(
  cli_wrap& cli,
  std::vector<boost::shared_ptr<vs_base> >& vecSpVs,
  boost::shared_ptr<alg_base>& spWorker,
  std::vector<boost::shared_ptr<vd_base> >& vecSpVd,
  int minDelayMs)
:
  pimp(new ui_imp(cli, vecSpVs,spWorker,vecSpVd, minDelayMs))
{}
ui::~ui(){}

void ui::loop(char const* s){
  pimp->loop(s);
}
void ui::batch(int rep, int off){
  pimp->batch(rep,off);
}
void ui::config(char const* s){
  pimp->config(s);
}
