#include "util_avi.hpp"

#include "../util/util_bmp.hpp"

#include <assert.h>
#include <stdlib.h>

#include <string>
#include <iostream>
#include <algorithm>
#include <stdexcept>

// Exclude rarely-used stuff from Windows headers
#define VC_EXTRALEAN		
#ifndef STRICT
#define STRICT
#endif
#include <windows.h>

#define NOMSACM 
#define NOMMREG 
#include <vfw.h>

#if defined(_MSC_VER)
  #undef min
  #undef max
  #if _MSC_VER<1300
    #define min _MIN
    #define max _MAX
  #endif
#endif


namespace util{

//////////////////////////////////////////
//        AviInit
// used for initialisation of vfw dll
class AviInit{
public:
  AviInit(){  AVIFileInit(); }
  ~AviInit(){ AVIFileExit(); }
};

static AviInit ai_global;



//////////////////////////////////////////
//        AviInfo
// aggregates the data common 
// to reading and writing
struct AviInfo{
  long timeStart;			// cached start, end, length
  long timeEnd;
  long timeLength;

  long frameEnd;
  long frameStart;
  long frameLength;

  unsigned long msecPerPixel;// scale for video
  
  AVICOMPRESSOPTIONS coptions;
  AVISTREAMINFO sinfo;

  PAVIFILE file;
  PAVISTREAM pstream;
  PGETFRAME pgf;

  util_bmp_BI bi;
  int sz_bi;
  int sz_bits;

  int frame;
  int time;
};

}//end namespace util


//////////////////////////////////////////
//        AviWrite
// encapsulates writing to a new avi file


util::AviWrite::AviWrite(
  char const *name, 
  int w, int h, int szpix,
  int szBits, int rate)
{
  int szBitsNeeded=util_bmp_BIH::sizeofBits(szpix*8,w,h);
  if (szBitsNeeded != szBits)
    throw std::runtime_error("util::AviWrite::AviWrite - incorrect szBits");

  pai_=new AviInfo;
  pai_->sz_bi=pai_->bi.init(szpix*8,w,h);
  pai_->sz_bits=util_bmp_BIH::sizeofBits(szpix*8,w,h);

  int rv=AVIFileOpen(&pai_->file, name, (OF_WRITE | OF_CREATE), 0L);
  if (0!=rv){
    rv-=AVIERR_UNSUPPORTED;
    throw std::runtime_error("util::AviWrite::AviWrite - can't open AVI file!");
  }

  RECT rc={0, 0, w, h};
  AVISTREAMINFO& strhdr=pai_->sinfo;
  strhdr.fccType=streamtypeVIDEO; 
  strhdr.fccHandler=0; 
  strhdr.dwFlags=0; 
  strhdr.dwCaps=0; 
  strhdr.wPriority=0; 
  strhdr.wLanguage=0; 
  strhdr.dwScale=1; 
  strhdr.dwRate=rate; 
  strhdr.dwStart=0; 
  strhdr.dwLength=0; 
  strhdr.dwInitialFrames=0; 
  strhdr.dwSuggestedBufferSize=pai_->sz_bits; 
  strhdr.dwQuality=0; 
  strhdr.dwSampleSize=0; 
  strhdr.rcFrame=rc; 
  strhdr.dwEditCount=0; 
  strhdr.dwFormatChangeCount=0; 
  strcpy(strhdr.szName, "Generator: util_avi by Sinisa Segvic"); 

  if (0!=AVIFileCreateStream(pai_->file, &pai_->pstream, &strhdr))
    throw std::runtime_error("util::AviWrite::AviWrite - can't open stream!");

  pai_->pgf=0;

  pai_->timeStart=0;
  pai_->timeEnd=0;
  pai_->timeLength=0;

  pai_->frameStart=0;
  pai_->frameEnd=0;
  pai_->frameLength=0;

  pai_->frame=0;
  pai_->time=0;

  memset(&pai_->coptions, 0, sizeof(pai_->coptions));

  if (0!=AVIStreamSetFormat(pai_->pstream, 0, &pai_->bi, pai_->sz_bi))
    throw std::runtime_error("util::AviWrite::AviWrite - got error setting file format!");
}

util::AviWrite::~AviWrite(){
	if (pai_->pstream)
    AVIStreamRelease(pai_->pstream);

  if (pai_->file)
	  AVIFileRelease(pai_->file);

  delete pai_;
}

void util::AviWrite::operator()(void const* buf, int t){
  if (t<0){
    int dt=1000/pai_->sinfo.dwRate;
    pai_->time+=dt;
  } else {
    pai_->time=t;
  }

// AVIStreamWrite - the documented purpose of 2nd argument
//                  differs from the way it is used in
//                  vfwdk/samples/writeavi/writeavi.c sample!
  long samples, bytes;
	HRESULT hr = AVIStreamWrite(
    pai_->pstream,
    pai_->frame,          // time of this frame?
		1,		              // number of samples to write
    (char*)(buf),
		pai_->sz_bits,       // size of this frame
    AVIIF_KEYFRAME,	    // flags....
    &samples, &bytes);
  if (hr != AVIERR_OK)
    throw std::runtime_error("util::AviWrite::addFrame() - got error writing frame!");

  ++pai_->frame;
}



//////////////////////////////////////////
//        AviRead
// encapsulates reading from an existing avi file

util::AviRead::AviRead(char const *name):
  pai_(new AviInfo)
{
  // Open up a file through the AVIFile handlers.
  int rv=AVIFileOpen(&pai_->file, name, OF_READ, 0L);
  std::cerr<< rv <<":" <<AVIERR_BADFORMAT <<"," <<REGDB_E_CLASSNOTREG <<"\n";
  if (0!=rv){
    rv-=AVIERR_UNSUPPORTED;
    throw std::runtime_error("util::AviRead::AviRead - can't open AVI file!");
  }

  //openStream - uzimam samo jedan stream iz file-a
  if (AVIFileGetStream(pai_->file, &pai_->pstream, streamtypeVIDEO, 0) != AVIERR_OK)
    throw std::runtime_error("util::AviRead::AviRead - can't open stream!");

  //initStream: initialize the streams of a loaded file -- 
  //the compression options, the DrawDIB handles, and the scroll bars.
  if (AVIStreamInfo(pai_->pstream, &pai_->sinfo, sizeof(pai_->sinfo)) != AVIERR_OK)
    throw std::runtime_error("util::AviRead::AviRead - can't init stream!");

	//
 	// Initialize the compression options to some default stuff
	// !!! Pick something better
	//
  memset(&pai_->coptions, 0, sizeof(pai_->coptions));
	pai_->coptions.fccType = pai_->sinfo.fccType;
  pai_->coptions.dwFlags = AVICOMPRESSF_VALID |
                         AVICOMPRESSF_KEYFRAMES | 
                         AVICOMPRESSF_DATARATE;
  pai_->coptions.fccHandler = 0;
  pai_->coptions.dwQuality = (DWORD)ICQUALITY_DEFAULT;
  pai_->coptions.dwKeyFrameEvery = 7;	// !!! ask compressor?
  pai_->coptions.dwBytesPerSecond = 60000;

  //
	// We're finding the earliest and latest start and end points for
	// our scrollbar.
	//
  pai_->timeStart = std::min((long)0x7fff, AVIStreamStartTime(pai_->pstream));
  pai_->timeEnd   = std::max(0L, AVIStreamEndTime(pai_->pstream));
  pai_->frameStart= AVIStreamTimeToSample(pai_->pstream, pai_->timeStart);
  pai_->frameEnd  = AVIStreamTimeToSample(pai_->pstream, pai_->timeEnd);
  pai_->frameLength   = std::max(0L, AVIStreamLength(pai_->pstream));
  //assert(pai_->frameLength==pai_->frameEnd-pai_->frameStart);

  //
	// Initialize video streams for getting decompressed frames to display
	//
  if ((pai_->pgf = AVIStreamGetFrameOpen(pai_->pstream, 0))==0)
    throw std::runtime_error("util::AviRead::AviRead - can't initialise stream frame");
	
  // space between frames for video streams
  const int HSPACE=4;
  DWORD dw = (pai_->sinfo.rcFrame.right - pai_->sinfo.rcFrame.left) + HSPACE;
  pai_->msecPerPixel = (1000000 * pai_->sinfo.dwScale)/ (dw * pai_->sinfo.dwRate);
  pai_->timeLength = pai_->timeEnd - pai_->timeStart;

  //read format
  long sz_fmt;
  if (AVIStreamReadFormat(pai_->pstream, 0, 0, &sz_fmt)!=AVIERR_OK)
    throw std::runtime_error("util::AviRead::AviRead - error reading file format!");
  if (sz_fmt>static_cast<int>(sizeof(util_bmp_BI)))
    throw std::runtime_error("util::AviRead::AviRead - unsupported format size!");
  if (AVIStreamReadFormat(pai_->pstream, 0, &pai_->bi, &sz_fmt)!=AVIERR_OK)
    throw std::runtime_error("util::AviRead::AviRead - error reading file format!");
  pai_->bi.detectNColours();
  pai_->bi.check();
  pai_->sz_bi=sz_fmt;
  pai_->sz_bits=util_bmp_BIH::sizeofBits(
    pai_->bi.bpp(), pai_->bi.w(), pai_->bi.h());
  /*NOTUSED*/
  //int rate=pai_->sinfo.dwRate/pai_->sinfo.dwScale;
}

util::AviRead::~AviRead(){
	if (pai_->pgf) 
    AVIStreamGetFrameClose(pai_->pgf);

  if (pai_->pstream)
    AVIStreamRelease(pai_->pstream);

  if (pai_->file)
	  AVIFileRelease(pai_->file);

  delete pai_;
}



int util::AviRead::countFrames(){
  return pai_->frameLength;
}


int util::AviRead::szBits(){
  return pai_->sz_bits;
}

int util::AviRead::szPixel(){
  return pai_->bi.bpp()/8;
}
int util::AviRead::width(){
  return pai_->bi.w();
}
int util::AviRead::height(){
  return pai_->bi.h();
}

void* util::AviRead::getFrame(int frame){
  assert (pai_->pgf);
  assert (frame<pai_->frameEnd);
  assert (pai_->frameStart==0);//proba, ss on 26.03.99

  void* lp=AVIStreamGetFrame(pai_->pgf, frame-pai_->frameStart);
  return static_cast<char*>(lp)+pai_->sz_bi;
}


int util::AviRead::time(int frame){
  return AVIStreamSampleToTime(pai_->pstream, frame);
}

