#include <assert.h>
#include <string.h>
#include <stdexcept>
#include "util_bmp.hpp"


/////////////////////////////////////////////////////////////
//  util_bmp_BIH implementation
util_bmp_BIH::util_bmp_BIH(){
  assert(sizeof(util_bmp_BIH)==40);

  util_endian::putLittle4(sizeof(util_bmp_BIH), biSize);

  util_endian::putLittle2(1,biPlanes);

  util_endian::putLittle4(0,biXPelsPerMeter);
  util_endian::putLittle4(0,biYPelsPerMeter);
}
void util_bmp_BIH::init(int bpp, int w, int h, int countEntries){
  util_endian::putLittle4(w,biWidth);
  util_endian::putLittle4(h,biHeight);
  util_endian::putLittle2(bpp,biBitCount);
  util_endian::putLittle4(sizeofBits(bpp,w,h),biSizeImage);
  util_endian::putLittle4(0,biClrImportant);
  util_endian::putLittle4(countEntries,biClrUsed);

  if (bpp==16){
    //util_endian::putLittle4(0,biClrUsed);//was 512!!!
    util_endian::putLittle4(BI_BITFIELDS_,biCompression);
  }else{
    util_endian::putLittle4(BI_RGB_,biCompression);
  }
}

void util_bmp_BIH::check(){
  if (sizeof(util_bmp_BIH)==util_endian::getLittle4(biSize))
    if (1==util_endian::getLittle2(biPlanes))
      if ( bpp()==8  || bpp()==16 || bpp()==24 || bpp()==32)
        return;
  throw std::runtime_error("util_bmp_BIH::check - unsupported format");
}

int util_bmp_BIH::bpp(){
  return util_endian::getLittle2(biBitCount);
}
int util_bmp_BIH::w(){
  return util_endian::getLittle4(biWidth);
}
int util_bmp_BIH::h(){
  return util_endian::getLittle4(biHeight);
}

int util_bmp_BIH::sizeofBits(int bpp, int w, int h){
  return calcSzScanline(w)*h*bpp/8;
}

// this one should match the analogous method
// in img_data in order to achieve simplicity and speed...
int util_bmp_BIH::calcSzScanline (int bytes){
//  int num = (bytes + 3) / 4;
//  return (num*4);
  return bytes;
}




/////////////////////////////////////////////////////////////
//  util_bmp_BI implementation

util_bmp_BI::util_bmp_BI(int ncols):
  nColours_(ncols)
{
  assert(sizeof(util_bmp_RgbUnion)==sizeof(util_bmp_RgbQuad));
}

int util_bmp_BI::init(int bpp, int w, int h, char const* palette, int ncols){
  //ncols -- count of palette entries 
  if (palette==0){//unspecified
    switch (bpp){
    case 8:{
      nColours_=256;
      util_bmp_RgbQuad rgb;
      rgb.rgbReserved=0;
      for (int i=0; i<256; i++){
        rgb.rgbBlue=rgb.rgbGreen=rgb.rgbRed=util_endian::ONE_BYTE(i);
        bmiColors[i].s=rgb;
      }
      break;
    }
    case 16:
      nColours_=3;
      util_endian::putLittle4(0x3fff,bmiColors[0].a);
      util_endian::putLittle4(0x4000,bmiColors[1].a);
      util_endian::putLittle4(0x8000,bmiColors[2].a);
      break;
    case 24:
    case 32:
      nColours_=0;
      break;
    default:
      throw std::runtime_error("util_bmp_BI::init - illegal count of bits per pixel");
    }
  } else {
    const unsigned int sz_palbuf=ncols*sizeof(util_bmp_RgbUnion);
    if (sz_palbuf>sizeof(bmiColors))
      throw std::runtime_error("util_bmp_BI::init - illegal palette size");
    ::memcpy(bmiColors, palette, sz_palbuf);
    nColours_=ncols;
  }
  bmiHeader.init(bpp, w, h, nColours_);
  return nColours_*sizeof(util_bmp_RgbUnion)+sizeof(util_bmp_BIH);
}


void util_bmp_BI::setNColours(int c){
  nColours_=c;
}
void util_bmp_BI::detectNColours(){
  switch (bpp()){
  case 8:
    nColours_=256; break;
  case 16:
    nColours_=3; break;
  case 24:
  case 32:
    nColours_=0; break;
  }
}


void util_bmp_BI::check(){
  bmiHeader.check();
  const unsigned int sz_palbuf=nColours_*sizeof(util_bmp_RgbUnion);
  if (sz_palbuf>sizeof(bmiColors))
    throw std::runtime_error("util_bmp_BI::check - unsupported format");
}

int util_bmp_BI::bpp(){
  return bmiHeader.bpp();
}
int util_bmp_BI::w(){
  return bmiHeader.w();
}
int util_bmp_BI::h(){
  return bmiHeader.h();
}
util_bmp_RgbUnion *util_bmp_BI::palette(){
  return bmiColors;
}
int util_bmp_BI::nColours(){
  return nColours_;
}
void util_bmp_BI::greyScalePalette(unsigned char* dst){
  for (int i=0; i<nColours(); i++){
    util_bmp_RgbQuad rq=bmiColors[i].s;
    int colour=static_cast<unsigned int>(rq.rgbRed)+
               static_cast<unsigned int>(rq.rgbGreen)+
               static_cast<unsigned int>(rq.rgbBlue);
    dst[i]=static_cast<char>((2*colour+3)/6);
  }
}


/////////////////////////////////////////////////////////////
//  util_bmp_BFH implementation

util_bmp_BFH::util_bmp_BFH(){
  util_endian::putLittle2(BITMAP, bfType);
  util_endian::putLittle2(0, bfReserved1);
  util_endian::putLittle2(0, bfReserved2);
}

int util_bmp_BFH::init(int sz_bi, int sz_bits){
  int sz_fhdr=sizeof(util_bmp_BFH);
  util_endian::putLittle4(sz_fhdr+sz_bi+sz_bits, bfSize);
  util_endian::putLittle4(sz_fhdr+sz_bi, bfOffBits);
  return sz_fhdr;
}

void util_bmp_BFH::check(){
  if (util_endian::getLittle2(bfType)!=BITMAP)
    throw std::runtime_error("util_bmp_BFH::check - unsupported format");
}
int util_bmp_BFH::sz_bi(){
  return util_endian::getLittle4(bfOffBits)-sizeof(util_bmp_BFH);
}
int util_bmp_BFH::sz_bits(){
  return util_endian::getLittle4(bfSize)-util_endian::getLittle4(bfOffBits);
}

//  static 
bool util_bmp_BFH::checkMagic(unsigned char const first, unsigned char const second){
  util_endian::ONE_BYTE magic[2];
  util_endian::putLittle2(BITMAP, magic);
  return (magic[0]==first &&
          magic[1]==second);
}

