#include "util_os.hpp"


#include <string>
#include <sstream>

#include <time.h>
#include <stdio.h>

#include <sys/types.h>
#include <sys/stat.h>

#if defined(__unix__) 
  #include <unistd.h>
  #include <execinfo.h>    /* for backtrace */
#endif

#if defined(_WIN32) 
  #include <windows.h>
#endif


struct EnvDetect{
  std::string os_;
  std::string cpu_;
  std::string cxx_;
  std::string version_;
  std::string pid_;
private:
  EnvDetect();
public:
  static EnvDetect& instance();
};

EnvDetect::EnvDetect(){
  ///////////////////////
  // detect OS
  os_=
#if defined(__WIN32__) || defined(_WIN32)
    "Win32";
#elif defined(__linux__)
    "Linux";
#elif defined(__APPLE__)
    "MacOS";
#else
    "unknown os";
#endif  

#if defined(__unix__) || defined(__MACH__)
  os_+= " (unix)";
#endif


  ///////////////////////
  // compiler version and backend cpu
  int major=0;
  int minor=0;
  cpu_="unknown cpu";
#if defined(__INTEL_COMPILER)
  cxx_="icc";
  major=__INTEL_COMPILER/100;
  minor=__INTEL_COMPILER%100;
  cpu_="i386";
#elif defined(__GNUC__)
  cxx_="g++";
  major=__GNUC__;
  minor=__GNUC_MINOR__;
# if defined(__i386__)
    cpu_="i386";
#   if defined(__tune_pentium__)
      cpu_+=" (pentium)";
#   endif
# elif defined(__x86_64__)
    cpu_="x86_64";
# elif defined(__POWERPC__)
    cpu_="Power PC";
# endif
#endif
#if defined(_MSC_VER)
  cxx_="msvc";
  major=_MSC_VER/100;
  minor=_MSC_VER%100;
#if defined(_M_IX86)
  cpu_="i386";
  if (_M_IX86/100>=5){
    cpu_+=" (pentium)";
  }
#endif
#endif
#if defined(__BORLANDC__)
  cxx_="borland";
#endif
  std::ostringstream oss;
  oss <<major <<"." <<minor;
  version_=oss.str();



  ///////////////////////
  // detect pid
  int mypid= 
#if defined(__WIN32__) || defined(_WIN32)
    GetCurrentProcessId();
#elif defined(__linux__)
    getpid();
#endif
  oss.str("");
  oss <<mypid;
  pid_=oss.str();
}






void util_os::printBacktrace(){
#if defined(__linux__)
  void *addresses[50];
  int c = backtrace(addresses, sizeof(addresses)/sizeof(void*));
  char **strings = backtrace_symbols(addresses,c);
  printf("backtrace returned: %d\n", c);
  for(int i = 0; i < c; i++) {
    printf("%d: %p", i, addresses[i]);
    printf("%s\n", strings[i]);
  }
#else
  printf("Backtrace available only on linux :-(\n");
#endif
}






EnvDetect& EnvDetect::instance(){
  static EnvDetect singleton;
  return singleton;
}












std::string util_os::env::os(){
  return EnvDetect::instance().os_;
}
std::string util_os::env::cxx(){
  return EnvDetect::instance().cxx_;
}
std::string util_os::env::version(){
  return EnvDetect::instance().version_;
}
std::string util_os::env::cpu(){
  return EnvDetect::instance().cpu_;
}
std::string util_os::env::pid(){
  return EnvDetect::instance().pid_;
}

std::string util_os::env::report(
  int argc, char* argv[])
{
  struct stat ss;
  stat(argv[0],&ss);
  std::string strTime=ctime(&ss.st_mtime);

  time_t now;
  time(&now);
  std::string strNow=ctime(&now);

  std::ostringstream oss;
  oss <<"The session:\n";
  oss <<"  Program name: " <<argv[0] <<std::endl;
  oss <<"  Built with "   
    <<cxx() <<" v"
    <<version();
  oss <<" for " <<cpu() 
    <<" running " <<os() 
    <<std::endl;
  oss <<"  Modification time: " <<strTime; 
  oss <<"  Current time:      " <<strNow;
  oss <<"  Process ID:        " <<pid()   <<std::endl;

  for (int i=1; i<argc; ++i){
    oss <<"    arg " <<i <<": " <<argv[i] <<std::endl;
  }

  return oss.str();

}

