#ifndef util_registry_hpp
#define util_registry_hpp

#include <map>
#include <string>
#include <typeinfo>
// #include <iostream>
#include <stdexcept>


/******************************************************
*
** util_registry.hpp by Sinisa Segvic:
*
** provides functionality needed 
** for implementing "factory method"
*
** member templates are inline and feature
** dummy default argument due to MSVC limitations
** http://support.microsoft.com/default.aspx?scid=kb;EN-US;241949
*
*******************************************************/


namespace util{

////////////////////////////
// Abstract creator classes

class RegistryCreatorBase{
public:
  virtual ~RegistryCreatorBase(){}
  virtual char const* name_derived()=0;
private:
  // heap allocations are forbidden!
  static void* operator new(size_t size);
};

template <class Base>
class RegistryCreatorAbstract:
  public RegistryCreatorBase
{
  friend class Registry;
private:
  virtual char const* name_derived()=0;
  typedef typename Base::CtorArg CtorArg;
  virtual Base* create(const CtorArg& param)=0;
};


/////////////////////////////
//the Registry class

class Registry{
  friend class Iterator;
//audience: Registry
private:
  class Key{
    std::string rep_;
  public:
    Key(std::string s): rep_(s){}
    Key(char const* s): rep_(s){}
    char const* c_str(){return rep_.c_str();}
    bool operator<(const Key& o) const {return rep_.compare(o.rep_)<0;}
    bool operator==(const Key& o) const{return rep_.compare(o.rep_)==0;}
    bool operator!=(const Key& o) const{return !(*this==o);}
  };
  Registry(){}
  typedef std::map<Key, RegistryCreatorBase*> CreatorMap;
  typedef std::map<Key, CreatorMap>           RegistryMap;
  RegistryMap map_;

  //throws for non-existing name_base 
  CreatorMap& getRegistry(char const *name_base);
  
//audience: RegistryCreatorAbstract ctor
public:
  void registerCreator(
    char const* name_base, 
    RegistryCreatorBase* p);


//audience: clients
public: //we are singleton
  static Registry& instance();

public: // enumeration
  class Iterator{
    friend class Registry;
    CreatorMap::iterator rep_;  
    Iterator(CreatorMap::iterator it):rep_(it){}
  public:
    bool operator==(const Iterator& o) const{return rep_==o.rep_;}
    bool operator!=(const Iterator& o) const{return !operator==(o);}
  public:
    RegistryCreatorBase& operator*() {return *rep_->second;}
    RegistryCreatorBase* operator->() {return &(operator*());}
    void operator++(){++rep_;}
  };
  template<class Base>
  Iterator begin(){
    CreatorMap& cm=getRegistry(typeid(Base).name());
    return Iterator(cm.begin());
  }
  template<class Base>
  Iterator end(){
    CreatorMap& cm=getRegistry(typeid(Base).name());
    return Iterator(cm.end());
  }


public: // indirect creation 
  //  returns end() for a non-existent name_derived
  template<class Base>
  Iterator find(
    char const* name_derived)
  {
    // std::cerr <<"In find() for " 
    //   <<typeid(Base).name() <<"->"
    //   <<name_derived <<"\n";
    CreatorMap& cm=getRegistry(typeid(Base).name());
    // std::cerr <<"  CreatorMap= " <<&cm <<"\n";

    CreatorMap::iterator it=cm.find(name_derived);
    return Iterator(it);
  }

  //  throws for unexistent it
  template<class Base>
  Base* create(
    Iterator it, 
    const typename Base::CtorArg& par)
  {
    if (it==end<Base>()){
      std::string err("util::registry::create<");
      err+=typeid(Base).name();
      err+="> - unsupported type";
      throw std::runtime_error(err);
    }
    RegistryCreatorAbstract<Base>& creator=
      dynamic_cast<RegistryCreatorAbstract<Base>&>(*it);
    return creator.create(par);
  }

public: // direct creation 
  //  throws for unexistent procedure name_derived
  template<class Base>
  Base* create(
    char const* name_derived, 
    const typename Base::CtorArg& par)
  {
    Iterator it=find<Base>(name_derived);
    if (it==end<Base>()){
      std::string err("util::registry::create<");
      err+=typeid(Base).name();
      err+="> - unsupported type: ";
      err+=name_derived;
      throw std::runtime_error(err);
    }
    return create<Base>(it, par);
  }

};


////////////////////////////
// The concrete creator class
class Registry;
template <class Base, class Derived>
class RegistryCreator:
  public RegistryCreatorAbstract<Base>
{
public:
  RegistryCreator(){
    //std::cerr <<"Registering creator for "
    //          <<Derived::s_name() <<"\n";
    Registry::instance().registerCreator(
      typeid(Base).name(), this);
  }
private:
  virtual char const* name_derived(){
    return Derived::s_name();
  }
  typedef typename Base::CtorArg CtorArg;
  virtual Base* create(const CtorArg& p){
    return new Derived(p);
  }
};


} // end namespace util

#endif

