#include "util_w32.hpp"

#include <windows.h>

#include "../img/img_data.hpp"
#include "../img/img_util.hpp"
#include "../img/access/img_access_copy.hpp"
#include "../util/util_bmp.hpp"

#include <algorithm>
#include <stdexcept>

#include <stdio.h>

#include <windows.h>

#if defined(_MSC_VER)
  #undef min
  #undef max
  #if _MSC_VER<1300
    #define min _MIN
    #define max _MAX
  #endif
#endif



////////////////////////////////////
// OS version detection

bool util_w32::winNT4or98(){
  //from MSDN Jan 2001:
  OSVERSIONINFO osvi;
  osvi.dwOSVersionInfoSize = 
    sizeof(OSVERSIONINFO);
  ::GetVersionEx (&osvi);

  bool bIsWindows98orLater = 
   (osvi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) &&
   ( (osvi.dwMajorVersion > 4) ||
   ( (osvi.dwMajorVersion == 4) && (osvi.dwMinorVersion > 0) ) );
  bool bIsNT4orLater= 
   (osvi.dwPlatformId == VER_PLATFORM_WIN32_NT) &&
   (osvi.dwMajorVersion >= 4);
  
  return bIsWindows98orLater || bIsNT4orLater;
}

bool util_w32::winNT4(){
  //from MSDN Jan 2001:
  OSVERSIONINFO osvi;
  osvi.dwOSVersionInfoSize = 
    sizeof(OSVERSIONINFO);
  ::GetVersionEx (&osvi);

  bool bIsNT4orLater= 
   (osvi.dwPlatformId == VER_PLATFORM_WIN32_NT) &&
   (osvi.dwMajorVersion >= 4);
  
  return bIsNT4orLater;
}





///////////////////////////////////////
// Drawing and capturing

void util_w32::captureDesktopRGB(
  void* buf,
  int x, int y,
  int w, int h)
{
  //prepare bitmapinfo structure for GetDIBits
  static BITMAPINFO bi={ 
    { sizeof(BITMAPINFOHEADER),
      0, 0, 1, 24,
      BI_RGB, 0, 0, 0, 0, 0},
    {{0,0,0,0}}
  };
  bi.bmiHeader.biWidth = w;
  bi.bmiHeader.biHeight= h;

  //allocate resources
  HDC     screen=CreateDC("DISPLAY", NULL, NULL, NULL); 
  HDC     memDC=CreateCompatibleDC(screen);
  HBITMAP bmp=CreateCompatibleBitmap(screen, w, h);
  if (bmp == NULL){
    DeleteDC(memDC);
    DeleteDC(screen);
    throw std::runtime_error("util_w32::captureDesktop(): could not allocate resources!");
  }

  //copy live data
  SelectObject(memDC, bmp);
  BitBlt(memDC, 0, 0, w, h, screen, x, y, SRCCOPY);

  // Convert bitmap into DIB
  GetDIBits(screen, bmp, 0, h, buf, &bi, DIB_RGB_COLORS);

  // clean-up resources
  DeleteObject(bmp);
  DeleteDC(memDC);
  DeleteDC(screen);
}

namespace{

int min4(int i0, int i1, int i2, int i3){
  int min01=std::min(i0, i1);
  int min23=std::min(i2, i3);
  int min0123=std::min(min01, min23);

  return min0123;
}

bool drawDIBits(
  HDC hdc, void const* pbits,
  int xsize, int ysize, int szpix,
  int posx, int posy)
{
  util_bmp_BI bi;
  bi.init(8*szpix, xsize, ysize);
  bi.check();

  RECT rc;
  ::GetClipBox(hdc,&rc);
  int rch=abs(rc.top-rc.bottom);
  int rcw=abs(rc.left-rc.right);

  //printf("util_w32::drawDIBits-->clipbox(%d, %d, %d, %d)\n",
  //  rc.top,rc.bottom, rc.left,rc.right);

  int destx=std::max((int)rc.left, posx);
  int desty=std::max((int)rc.top, posy);
  int destwt=min4(rcw, xsize, rc.right-posx, posx+xsize-rc.left);
  int destht=min4(rch, ysize, rc.bottom-posy,posy+ysize-rc.top);

  //fix error in SetDIBitsToDevice
  if (destwt<=0 || destht<=0) 
    return false;

  int sourcex=std::min(std::max((int)rc.left-posx, 0), xsize);
  int sourcey=std::min(std::max(posy+ysize-(int)rc.bottom, 0),ysize);

  //printf("util_w32::drawDIBits-->SetDIBitsToDevice(%d, %d, %d, %d, %d, %d, %d, %d)\n",
  //  destx, desty, destwt, destht, sourcex, sourcey, 0, ysize);

  // 1-1 preslikavanje (no stretching)
  int retval = ::SetDIBitsToDevice(
    hdc,
    destx, desty,
    destwt, destht,
    sourcex, sourcey,
    0, ysize,
    pbits,
    static_cast<BITMAPINFO*>(static_cast<void*>(&bi)),
    DIB_RGB_COLORS);
  if (!retval){
    DWORD e=::GetLastError();
    if (e){
      //printf("util_w32::drawDIBits-->got error(%d)\n",e);
      throw std::runtime_error("util_w32::drawDIBits - error in SetDIBitsToDevice");
    }
  }

  return true;
}
} //end unnamed namespace




bool util_w32::drawImg(
  HDC hdc, const img_data& img,
  int posx, int posy)
{
  const img_fmt& fmt=img.fmt();
  int w=fmt.width();
  int h=fmt.height();
  int szpix=fmt.szPixel();

  if (szpix<1 || (szpix>4 && szpix!=6))
    throw "util_w32::drawImg"
          "illegal szPixel()";

  img_data imgTmp;

  //check for odd width
  if (w%4!=0){
    w+=4-w%4;
  }

  //check for szpix==2
  if (szpix==2){
    szpix=1;
  }
  if (szpix==6){
    szpix=3;
  }

  // copy to tmp
  imgTmp.reformat(img_fmt(w,h,szpix));
  img_access_copy(img, imgTmp);

  //check for endianness
  if (szpix==4 || szpix==3){
    img_util::flipEndian(imgTmp);
  }

  // no chech for fmt.bottomUp() which is true ...

  void const* pbits=imgTmp.bits();

  bool rv=drawDIBits(hdc, pbits, w,h,szpix, posx,posy);
  return rv;
}

