#include "vd_win.hpp"

#include "../win/root/win_wrap.hpp"
#include "../win/ann/win_ann_dump.hpp"

#include "../util/util_string.hpp"
#include "../util/util_time.hpp"

#include <assert.h>

#include <sstream>
#include <string>
#include <iomanip>
#include <stdexcept>
#include <iostream>


//////////////////////////////
// vd_win private data

class vd_win_imp{
public:
  vd_win_imp(const img_fmt& fmt);
  ~vd_win_imp();

  win_base& win(){return win_;}

  void caption(char const*);
  void operator()(
    const img_data& img, 
    int time=0, int frame=-1);
  void operator()(
    const img_data& img, 
    const win_ann_abstract& a, 
    int time=0, int frame=-1);

  void resetTime();
  bool toggleRec();
private:
  void formatCaption(int time, int frame);

private:
  std::string caption_;
  std::string outName_;
  win_wrap win_;

  int count_;
  int tmsPrevPrev_;
  int tmsPrev_;
  int tmsStart_;

  bool recording_;
};


vd_win_imp::vd_win_imp(const img_fmt& fmt):
  win_("", fmt.width(), fmt.height()),
  recording_(false)
{
  resetTime();
}

vd_win_imp::~vd_win_imp(){}



void vd_win_imp::caption(char const* s){
  std::string filename;
  util_string::retractPath(s, filename);
  caption_ = filename.c_str();
  win_.caption(filename.c_str());
}


void vd_win_imp::operator()(
  const img_data& img, 
  int tmsFrame, int frame)
{
  formatCaption(tmsFrame, frame);
  win_.setImg(img);
}
void vd_win_imp::operator()(
  const img_data& img, 
  const win_ann_abstract& a, 
  int tmsFrame, int frame)
{
  formatCaption(tmsFrame, frame);
  win_.setImgAnn(img, a);
  if (recording_){
    win_ann_dump::toTwoFiles(a,img,outName_.c_str());
  }
}


void vd_win_imp::resetTime(){
  tmsStart_=util_time::msDayUtc();
  tmsPrevPrev_=-1;
  tmsPrev_=-1;
  count_=0;
}
bool vd_win_imp::toggleRec(){
  recording_=!recording_;
  if (recording_){
    outName_=caption_;
    for (unsigned i=0; i<outName_.size(); i++)
      if (!isalnum(outName_[i]))
        outName_.erase(i);
  }
  return recording_;
}

void vd_win_imp::formatCaption(int tmsFrame, int frame){
  util_time::Data tmsNow=util_time::msDayUtc();

  if (frame==-1)
    frame=count_;
  if (tmsFrame==-1)
    tmsFrame=tmsNow-tmsStart_;
  double t_s=tmsFrame/1000.0;

  double dt=0;
  if (tmsPrevPrev_>=0){
    dt=(tmsNow-tmsPrevPrev_)/2;
  } else if (tmsPrev_>=0){
    dt=tmsNow-tmsPrev_;
  }
  double fps=dt?1000.0/dt:0;

  std::ostringstream oss;
  oss <<caption_ <<":";
  oss.setf(std::ios::left, std::ios::adjustfield);

  std::ostringstream oss1;
  oss1 <<"#" <<frame;
  oss <<' ' <<std::setw(5) <<oss1.str();

  oss1.str("");
  oss1.setf(std::ios::fixed, std::ios::floatfield);
  oss1.precision(1);
  oss1 <<"@" <<fps <<" fps";
  oss <<' ' <<std::setw(10) <<oss1.str();

  oss1.str("");
  oss1.precision(1);
  oss1 <<"("  <<t_s <<" s)";
  oss <<' ' <<std::setw(10) <<oss1.str();

  win_.caption(oss.str().c_str());

  count_++;
  tmsPrevPrev_=tmsPrev_;
  tmsPrev_=tmsNow;
}


//////////////////////////////
// vd_win construction

vd_win::vd_win(const img_fmt& f, char const* name)
:
  vd_base(f),
  pimp(new vd_win_imp(f))
{
  caption(name);
}

vd_win::~vd_win(){
}



//////////////////////////////
// vd_win attributes
win_base& vd_win::win() const{
  return pimp->win();
}

//////////////////////////////
// vd_base operations

void vd_win::operator()(
  const img_data& img, 
  int tmsFrame, int frame)
{
  (*pimp)(img, tmsFrame, frame);
  
  if (! (img.fmt() == fmt())){
    reformat(img.fmt());
  }
}

//////////////////////////////
// vd_win operations
void vd_win::size(int x, int y){
  win().size(x,y);
}
void vd_win::position(int x, int y){
  win().position(x,y);
}


void vd_win::caption(char const* s){
  pimp->caption(s);
}


void vd_win::operator()(
  const img_data& img, 
  const win_ann_abstract& a, 
  int tmsFrame, int frame)
{
  (*pimp)(img, a, tmsFrame, frame);

  if (! (img.fmt() == fmt())){
    reformat(img.fmt());
  }
}


bool vd_win::toggleRec(){
  return pimp->toggleRec();
}
void vd_win::resetTime(){
  pimp->resetTime();
}


