#include "vs_file_many.hpp"

#include "vs_file_util.hpp"

#include "../img/img_data.hpp"

#include "../util/util_string.hpp"

#include <boost/shared_ptr.hpp>


#include <algorithm>
#include <string>
#include <sstream>
#include <vector>
#include <fstream>
#include <stdexcept>
#include <map>
#include <iostream>

#include <assert.h>
#include <stdio.h>


// private helpers

namespace{
  void createVSFile(
    const std::string& path, 
    boost::shared_ptr<vs_file>& pvs)
  {
    try{
      // std::cerr <<"vs_file_many: created vs_file: " <<path <<"\n";
      pvs.reset(vs_file_util::create(path.c_str()));
    } catch(std::exception& e){
      std::ostringstream oss;
      oss <<"vs_file_many - error while creating ";
      oss <<path.c_str() <<" :\n" <<e.what();
      throw std::runtime_error(oss.str());
    }
  }


  struct MyThrower{
    MyThrower(char const* name){
      context_="vs_file_many (";
      context_+=name;
      context_+=") - ";
    }

    void operator()(const std::string &msg){
      throw std::runtime_error(context_+msg);
    }

    std::string context_;
  };
}


//////////////////////////////////////////////////
//vs_file_many private implementation

class vs_file_many_imp{
public:
  vs_file_many_imp(
    const std::vector<std::string>& vecStrFile);
  ~vs_file_many_imp();

  void getFrame(int frame, img_data& d);
  int fpos(int frame);
  
  int count(){return count_;}
  int msDayUtc(){return tmsDayUtc_;}
  const img_fmt& fmt(){return fmt_;}
  
private:
  void push(
    const std::string& str, 
    int from, int to);
  int find(int offset);

  img_fmt fmt_;
  int count_;
  int tmsDayUtc_;

  std::map<std::string, boost::shared_ptr<vs_file> > mapVs_;
  std::vector<int> vecFrom_;
  std::vector<int> vecTo_;
  std::vector<int> vecOffset_;
  std::vector<std::string> vecPath_;
};



vs_file_many_imp::vs_file_many_imp(
  const std::vector<std::string>& vecStrFile)
:
  tmsDayUtc_(0)
{
  count_=0;
  for (unsigned int i=0; i<vecStrFile.size(); ++i){
    MyThrower error(vecStrFile[i].c_str());
    
    // parse the vs spec
    std::vector<int> vecFrom;
    std::vector<int> vecTo;
    std::string thisFile;
    int n=vs_file_many::parseRange(vecStrFile[i], 
      thisFile,vecFrom,vecTo);
    std::string thisPath(thisFile);

    // check the file    
    FILE* pf=fopen(thisPath.c_str(), "r");
    if (!pf) 
      error("access error");
    fclose(pf);

        
    // push components of the vs spec
    if (n==0){
      push(thisPath, 0,1);
    } else{
      for (int j=0; j<n; ++j){
        if (vecTo[j]<0) 
          error("full range specification required");
        if (vecFrom[j]>=vecTo[j]) 
          error("correct range specification required");

        push(thisPath, vecFrom[j], vecTo[j]);
      }
    }
  }
  vecOffset_.push_back(count_);

  if (count_==0)
    throw "vs_file_many_imp::vs_file_many_imp"
          " - got an empty sequence";

  // find out the format of the first vs_file          
  boost::shared_ptr<vs_file> pvs;
  createVSFile(vecPath_[0], pvs);
  fmt_=pvs->fmt();  
}
vs_file_many_imp::~vs_file_many_imp(){
  //std::cerr <<"vs_file_many: in destructor\n";
}


void vs_file_many_imp::push(
  const std::string& path,
  int from, int to)
{
  mapVs_[path]=boost::shared_ptr<vs_file>();

  vecPath_.push_back(path);
  vecFrom_.push_back(from);
  vecTo_.push_back(to);
  vecOffset_.push_back(count_);

  count_ += to-from;
}

int vs_file_many_imp::find(int offset){
  assert(offset<count_);
  assert(vecPath_.size()+1==vecOffset_.size());
  for (unsigned i=0; i<vecPath_.size(); i++){
    if (
      offset>=vecOffset_[i] &&
      offset< vecOffset_[i+1])
    {
      return i;
    }
  }
  assert(0);
  return -1;
}

void vs_file_many_imp::getFrame(int frame, img_data& d)
{
  int i=find(frame);
  const std::string& path=vecPath_[i];

  // lazy construction with conditional caching  
  boost::shared_ptr<vs_file> pvs(mapVs_[path]);
  if (!pvs.get()){
    createVSFile(path, pvs);
    if (mapVs_.size()<250){
      // std::cerr <<"vs_file_many: cached " <<path <<"\n";
      // makes mapVs_[path] another image of *pvs
      mapVs_[path]=boost::shared_ptr<vs_file>(pvs);
    }
  }

  int pos=fpos(frame);
  pvs->seek(pos);
  pvs->getFrame(d);
  tmsDayUtc_=pvs->msDayUtc();
}

int vs_file_many_imp::fpos(int frame){
  int i=find(frame);
  return frame-vecOffset_[i]+vecFrom_[i];
}




//////////////////////////////////////////////////
//vs_file_many static methods

int vs_file_many::parseRange(
  const std::string& str, 
  std::string& path, 
  std::vector<int> vecFrom, 
  std::vector<int> vecTo)
{
  size_t pos=path.find_first_of('#');
  if (pos==std::string::npos){
    path=str;
    return 0;
  }
    
  path=str.substr(0,pos);
  bool done=false;
  while (!done){
    int from=0, to=-1;
    
    size_t pos2=path.find_first_of('#', pos+1);
    if (pos2==std::string::npos){
      pos2=path.size();
      done=true;
    }

    std::string range=path.substr(pos,pos2-pos);
    size_t pos_to=range.find_first_of("-");
    if (pos_to==std::string::npos){
      pos_to=range.size();
    } else{
      std::string strTo(range.begin()+pos_to+1, range.end());
      std::istringstream(strTo) >>to;
    }

    std::string strFrom(range.begin()+1, range.begin()+pos_to);
    std::istringstream(strFrom) >>from;

    vecFrom.push_back(from);
    vecTo.push_back(to);
    
    pos=pos2;
  }
  return static_cast<int>(vecFrom.size());
}



//////////////////////////////////////////////////
//vs_file_many implementation

vs_file_many::vs_file_many(
  const std::vector<std::string>& vecStrFile)
:
  pimp(new vs_file_many_imp(vecStrFile))
{
}

vs_file_many::~vs_file_many(){
}



//////////////////////////////////////////////////
//attributes
int vs_file_many::count(){
  return pimp->count();
}  

int vs_file_many::msDayUtc(){
  return pimp->msDayUtc();
}  

const img_fmt& vs_file_many::fmt(){
  return pimp->fmt();
}  

//////////////////////////////////////////////////
//operations
void vs_file_many::getFrameSimple(img_data& d){
  pimp->getFrame(frame(), d);
}  


