#include "vs_grab_linux1394.hpp"

#include "../img/img_data.hpp"
#include "../util/util_debug.hpp"
#include "../util/util_registry.hpp"

#include <libraw1394/raw1394.h>
#include <libdc1394/dc1394_control.h>

#include <boost/smart_ptr.hpp>

#include <vector>
#include <stdexcept>

#include <unistd.h>
#include <sys/time.h>
#include <assert.h>
#include <string.h>

#include <iostream>

/////////////////////////////////////
// static registry creator object

namespace{
  util::RegistryCreator<vs_grab,vs_grab_linux1394> creator;
}

/////////////////////////////////////
// logging

#ifdef NDEBUG
  typedef util::NoLog LogClass;
#else
  typedef util::Log LogClass;
#endif
namespace{
  //the logging object
  LogClass log_("/tmp/vs_grab_linux1394.log");
}

namespace{
  inline time_t timeToMs(struct timeval& tv){
    return (tv.tv_sec % (24*60*60))*1000 + tv.tv_usec/1000;
  }
    
  inline time_t curMsDayUtc(){
    static struct timeval tv;
    static struct timezone tz;
    gettimeofday(&tv, &tz);
    return timeToMs(tv);
  }

  inline void milisleep(int ms){
    timespec ts;
    ts.tv_sec=ms/1000;
    ts.tv_nsec=ms*1000000;
    nanosleep(&ts,0);
  }

  inline int myabs(int a){
    return a<0?-a:a;
  }

  const int Inf_=std::numeric_limits<int>::max();
}


/******************************************************
 ** singleton class encapsulating 
 ** all cameras on a given port
 ******************************************************/

class Cam1394;

class Lib1394{
  Lib1394();
public:
  static Lib1394& instance();
  ~Lib1394();

public:
  void connect(int port,int index, int mode, int w, int szpix);
  void disconnect(int port, int index);
  Cam1394& cam(int port, int index);

private:
  void checkPort(int port);
  void checkPortIndex(int port, int index);
private:
  //active camera connections (port x index)
  std::vector<std::vector<boost::shared_ptr<Cam1394> > > map_;
  
  std::vector<raw1394_portinfo> pinfo_;
};


/******************************************************
 ** class encapsulating single camera connection
 ******************************************************/


class Cam1394{
public:
  Cam1394(
    int port, int index, int channel, int mode,
    int w, int sz);
  ~Cam1394();

  int width() const {return width_;}
  int height() const {return height_;}
  int szpix() const {return szpix_;}

  void getFrame(unsigned char* bits);
  int msDayUtc(){return msDayUtc_;}
  int szBits(){return width()*height()*szpix();}

private:
  void cleanup();

private:
  bool negotiateFormatAndMode(int w, int sz);
  void setHighestFrameRate();
  void checkCameraParameters();

  static int fixQuadlet(quadlet_t val);
  static bool queryBit(quadlet_t val, int bit);

  static inline void yuv2rgb(
      const int& y, const int& u, const int& v,
      int &r, int &g, int& b);
  static inline void rgb2yuv(
    const int &r, const int &g, const int& b,
    int& y, int& u, int& v);

private:
  const bool doesdma_;

  raw1394handle_t dch_;
  nodeid_t node_;
  dc1394_cameracapture cap_;
  dc1394_camerainfo info_;
  dc1394_miscinfo misc_;
  dc1394_feature_set features_;
  time_t msDayUtc_;

  // data from IIDC 1394 DCAM Spec
  // enum fmts{VGA=0,SVGA1,SVGA2, R3,R4,R5, STILL,PARTIAL};
  enum fmts{
    FMT_VGA=0,
    FMT_SVGA1=1,
    FMT_SVGA2=2, 
    FMT_R3,
    FMT_R4,
    FMT_R5, 
    FMT_STILL,
    FMT_PARTIAL};
  enum modes_vga{
    M160x120xYUV444=0,
    M320x240xYUV422,
    M640x480xYUV411,
    M640x480xYUV422,
    M640x480xRGB,
    M640x480xMONO8,
    M640x480xMONO16,
    Mreserved};
  enum modes_svga1{
    M800x600xYUV422=0,
    M800x600xRGB,
    M800x600xMONO8,
    M1024x768xYUV422,
    M1024x768xRGB,
    M1024x768xMONO8,
    M800x600xMONO16,
    M1024x768xMONO16};
  enum modes_svga2{
    M1280x960xYUV422=0,
    M1280x960xRGB,
    M1280x960xMONO8,
    M1600x1200xYUV422,
    M1600x1200xRGB,
    M1600x1200xMONO8,
    M1280x960xMONO16,
    M1600x1200xMONO16};
  enum rates{FPS3_75=1, FPS7_5=2, FPS15=3, FPS30=4, FPS60=5};

  int format_;
  int mode_;
  int rate_;
  int rate_ms_;

  int width_;
  int height_;
  int szpix_;
};



/******************************************************
 ** Lib1394 public members
 ******************************************************/

Lib1394& Lib1394::instance(){
  static Lib1394 theLib;
  return theLib;
}


Lib1394::Lib1394(){
  log_ <<"In Lib1394::Lib1394 \n";
  raw1394handle_t rh = raw1394_new_handle();
  if (!rh) 
    throw "vs_grab_linux1394::Lib1394::Lib1394"
          " - couldn't open libraw (drivers?).";
  int n=raw1394_get_port_info(rh, 0, 0);
  pinfo_.resize(n);
  map_.resize(n);
  raw1394_get_port_info(rh, &pinfo_[0], n);
  raw1394_destroy_handle(rh);
}

Lib1394::~Lib1394(){
  log_ <<"In Lib1394::~Lib1394 \n";
  log_.flush();

  //check for dangling connections
  for (unsigned port=0; port<map_.size(); ++port){
    for (unsigned index=0; index<map_[port].size(); ++index){
      if (map_[port][index]){
        log_ <<"Lib1394::~Lib1394 - fixing connection " 
             <<port <<"," <<index <<"\n";
        log_.flush();
        disconnect(port,index);
      }
    }
  }

  log_ <<"Exiting Lib1394::~Lib1394 \n";
  log_.flush();
}


void Lib1394::connect(
  int port, int index, int mode,
  int w, int szpix){
  checkPort(port);
  if (index < static_cast<int>(map_[port].size()) && 
      map_[port][index].get())
  {
    throw "vs_grab_linux1394::Lib1394::connect"
          " - the specified camera is already connected.";
  }

  //TODO: provide more sensible choice for channel
  Cam1394* pcam=new Cam1394(port, index, index+1, mode, w, szpix);
  if (pcam){
    map_[port].resize(index+1);
    map_[port][index].reset(pcam);
  }
}

void Lib1394::disconnect(int port, int index){
  checkPortIndex(port, index);
  map_[port][index].reset();
}

Cam1394& Lib1394::cam(int port, int index){
  checkPortIndex(port, index);
  return *map_[port][index];
}

void Lib1394::checkPort(int port){
  if (port>= static_cast<int>(map_.size())){
    throw "vs_grab_linux1394::Lib1394::disconnect"
          " - illegal port.";
  }
}
void Lib1394::checkPortIndex(int port, int index){
  checkPort(port);
  if (index>= static_cast<int>(map_[port].size())){
    throw "vs_grab_linux1394::Lib1394::disconnect"
          " - illegal index.";
  }
}



/******************************************************
 ** Cam1394 public members
 ******************************************************/

Cam1394::Cam1394(
  int port, int index, int channel, int mode, 
  int width, int szpix)
:
  doesdma_(!mode)
{
  log_ <<"Entering Cam1394::ctor (" <<port <<"," <<index <<")\n";

  dch_ = dc1394_create_handle(port);
  if (!dch_) 
    throw "vs_grab_linux1394::Cam1394::ctor"
          " - couldn't open libdc (drivers?).";

  int numcams;
  nodeid_t *camNodes = dc1394_get_camera_nodes(dch_, &numcams, 0);
  int numnodes=raw1394_get_nodecount(dch_);
  if (index>=numcams){
    throw "vs_grab_linux1394::Cam1394::ctor"
          " - the camera index does not exist.";
  }

  node_=camNodes[index];
  dc1394_free_camera_nodes(camNodes);
  if(node_ == numnodes-1){
    throw "vs_grab_linux1394::Cam1394::ctor"
          " - libraw1394 is in the incorrect state:"
          " please reinitialize the ieee1394 drivers using gscanbus.";
  }

  
  // get info and features
  if (dc1394_get_camera_feature_set(dch_,node_,&features_) 
      !=DC1394_SUCCESS) 
    throw "vs_grab_linux1394::Cam1394::ctor - "
          "unable to get feature set";
  /// dc1394_print_feature_set(&features_);
  
  if (dc1394_get_camera_info(dch_,node_,&info_) 
      !=DC1394_SUCCESS) 
    throw "vs_grab_linux1394::Cam1394::ctor - "
          "unable to get camera info";
  /// dc1394_print_camera_info(&info_); 
  
  if (dc1394_get_camera_misc_info(dch_,node_,&misc_) 
      !=DC1394_SUCCESS) 
    throw "vs_grab_linux1394::Cam1394::ctor - "
          "unable to get camera misc info";
  // check that the ISO transfer is off
  if (misc_.is_iso_on) {
    dc1394_stop_iso_transmission(dch_, node_);
  }
   
  bool fmtok=negotiateFormatAndMode(width,szpix);
  if (!fmtok) 
    throw "vs_grab_linux1394::Cam1394::ctor - "
          "unable to negotiate format and mode";
  setHighestFrameRate();

  //setup capture
  int rv;
  if (doesdma_){
    int drop_frames=1;
    int num_buffers=3;
    rv=dc1394_dma_setup_capture(dch_, node_, channel,
       format_, mode_, SPEED_400, rate_, 
       num_buffers, drop_frames, 
       cap_.dma_device_file, &cap_);
  } else{
    rv=dc1394_setup_capture(dch_, node_, channel,
       format_, mode_, SPEED_400, rate_, &cap_);
  }
  if (rv!=DC1394_SUCCESS) 
    throw "vs_grab_linux1394::Cam1394::ctor - "
          "unable to setup camera capture parameters";
  msDayUtc_=curMsDayUtc();

  // start transmission
  if (dc1394_start_iso_transmission(dch_,node_)
      !=DC1394_SUCCESS) 
  {
    cleanup();
    throw "vs_grab_linux1394::Cam1394::ctor - "
          "unable to start iso transmission";
  }

  // wait for VIDERE cam to start sending data
  milisleep(350);

  // this might be checked while debugging
  // checkCameraParameters();
  log_ <<"Exiting Cam1394::Cam1394\n";
}

Cam1394::~Cam1394(){
  log_ <<"Entering Cam1394::~Cam1394\n";
  log_.flush();
  dc1394_stop_iso_transmission(dch_, node_);
  log_ <<"Stopped ISO transmission\n";
  log_.flush();
  cleanup();
  log_ <<"Exiting Cam1394::~Cam1394\n";
  log_.flush();
}

void Cam1394::cleanup(){
  if (doesdma_){
    dc1394_dma_unlisten(dch_, &cap_);
    dc1394_dma_release_camera(dch_,&cap_);
  } else{
    dc1394_release_camera(dch_,&cap_);
  }
  dc1394_destroy_handle(dch_);
}

void Cam1394::getFrame(unsigned char* pdst) {
  /// time_t msEntry=curMsDayUtc();

  // capture next frame:
  unsigned char const* psrc;
  if (doesdma_){
    for(;;){
      dc1394_dma_single_capture(&cap_);
      psrc = cap_.dma_ring_buffer + 
             cap_.dma_frame_size * cap_.dma_last_buffer;

      time_t msNow=curMsDayUtc();
      time_t msFill=timeToMs(cap_.filltime);
      /// std::cerr <<" wait_ms=" <<msNow-msEntry <<"\n";
      time_t msDelta=msNow-msFill;
                
      if (msDelta<rate_ms_/2){
        msDayUtc_=timeToMs(cap_.filltime);
        ///time_t msDayUtcLast=msDayUtc_;
        ///std::cerr <<"Got frame on " <<msDayUtc_ 
        ///          <<" delta=" <<msDayUtc_-msDayUtcLast 
        ///          <<" age_ms=" <<msDelta
        ///          <<"\n";
        log_ <<"Cam1394::getFrame - flushed frame (" 
             <<msDelta <<")\n";
        break;
      }

      /// std::cerr <<"  missed frame age_ms=" <<msDelta <<"\n";

      // a fix for intermittent freeezes on FC3
      // on cycab-hf3 at Irisa Rennes
      milisleep(1);
      
      if (dc1394_dma_done_with_buffer(&cap_) != DC1394_SUCCESS)
        throw "vs_grab_linux1394::Cam1394::getFrame: "
              "error releasing dma frame buffer";
    }
      

  } else{
    dc1394_single_capture(dch_, &cap_);
    psrc = reinterpret_cast<unsigned char*> (cap_.capture_buffer);
    msDayUtc_=curMsDayUtc();
  }

  assert (cap_.frame_width == width_);
  assert (cap_.frame_height == height_);
  pdst+=(height_-1)*width_*szpix_;

  // source: coriander v0.24, src/conversions.c
  switch(mode_) {
    // YUV444 to RGB or GRAY pixel conversion
    // with a vertical flip operation
    case MODE_160x120_YUV444: {
      int r, g, b, y0, u0, v0;
      for (int l = 0; l < height_; l++) {
        for (int i = 0; i < width_; i++) {
          u0 = (*psrc++) - 128; y0 = *psrc++; v0 = (*psrc++) - 128;
          switch (szpix_){
          case 1:
            *pdst++=y0;
            break;
          case 3:
            yuv2rgb(y0, u0, v0, r, g, b); 
            *pdst++ = r; *pdst++ = g; *pdst++=b;
            break;
          default:
            assert(0);
          }
        }
        pdst-=2*width_*szpix_;
      }
      break;
    }
    // YUV422 to RGB or GRAY pixel conversion
    // with a vertical flip operation
    case MODE_320x240_YUV422:
    case MODE_640x480_YUV422:
    case MODE_800x600_YUV422:
    case MODE_1024x768_YUV422:
    case MODE_1280x960_YUV422:
    case MODE_1600x1200_YUV422: {
      int r, g, b, y0, y1, u, v;
      for (int l = 0; l < height_; l++) {
        for (int i = 0; i < width_/2; i++) {
          u = (*psrc++) - 128; y0 = *psrc++; 
          v = (*psrc++) - 128; y1 = *psrc++;
          switch (szpix_){
          case 1:
            *pdst++=y0;
            *pdst++=y1;
            break;
          case 3:
            yuv2rgb(y0, u, v, r, g, b); 
            *pdst++ = r; *pdst++ = g; *pdst++=b;
            yuv2rgb(y1, u, v, r, g, b); 
            *pdst++ = r; *pdst++ = g; *pdst++=b;
            break;
          default:
            assert(0);
          }
        }
        pdst-=2*width_*szpix_;
      }
      break;
    }
    // YUV411 to RGB or GRAY pixel conversion
    // with a vertical flip operation
    case MODE_640x480_YUV411:{
      int y0, y1, y2, y3, u, v, r, g, b;
      for (int l = 0; l < height_; l++) {
        for (int i = 0; i < width_/4; i ++) {
          u = (*psrc++) - 128; y0 = *psrc++; y1 = *psrc++; 
          v = (*psrc++) - 128; y2 = *psrc++; y3 = *psrc++;
          switch (szpix_){
          case 1:
            *pdst++=y0;
            *pdst++=y1;
            *pdst++=y2;
            *pdst++=y3;
            break;
          case 3:
            yuv2rgb(y0, u, v, r, g, b); 
            *pdst++ = r; *pdst++ = g; *pdst++=b;
            yuv2rgb(y1, u, v, r, g, b); 
            *pdst++ = r; *pdst++ = g; *pdst++=b;
            yuv2rgb(y2, u, v, r, g, b);
            *pdst++ = r; *pdst++ = g; *pdst++=b;
            yuv2rgb(y3, u, v, r, g, b);
            *pdst++ = r; *pdst++ = g; *pdst++=b;
            break;
          default:
            assert(0);
          }
        }
        pdst-=2*width_*szpix_;
      }
      break;
    }
    //pure rgb24 
    case MODE_640x480_RGB:
    case MODE_800x600_RGB:
    case MODE_1024x768_RGB:
    case MODE_1280x960_RGB:
    case MODE_1600x1200_RGB:{
      for (int l = 0; l < height_-1; l++) {
        memcpy(pdst, psrc, width_*szpix_);
        psrc+=width_*szpix_;
        pdst-=width_*szpix_;
      }
      break;
    }
    // grey level pixels
    case MODE_640x480_MONO:
    case MODE_800x600_MONO:
    case MODE_1024x768_MONO:
    case MODE_1280x960_MONO:
    case MODE_1600x1200_MONO:
    case MODE_FORMAT7_0:
    case MODE_FORMAT7_1:
    case MODE_FORMAT7_2:
    case MODE_FORMAT7_3:
    case MODE_FORMAT7_4:
    case MODE_FORMAT7_5:
    case MODE_FORMAT7_6:
    case MODE_FORMAT7_7:{
      for (int l = 0; l < height_; l++) {
        memcpy(pdst, psrc, width_*szpix_);
        psrc+=width_*szpix_;
        pdst-=width_*szpix_;
      }
      break;
    }
    default:
      throw "vs_grab_linux1394::Cam1394::getFrame"
            "capture mode is not supported";
  }

  if (doesdma_)
    if (dc1394_dma_done_with_buffer(&cap_) != DC1394_SUCCESS)
      throw "vs_grab_linux1394::Cam1394::getFrame"
            "error releasing dma frame buffer";
}


/******************************************************
 ** Cam1394 startup routines
 ******************************************************/

bool Cam1394::negotiateFormatAndMode(
  int width, int szpix)
{
  // check for supported formats and modes:
  quadlet_t qFmt;
  if (dc1394_query_supported_formats(dch_, node_, &qFmt) 
      != DC1394_SUCCESS)
    throw "vs_grab_linux1394::Cam1394::ctor - "
          "cannot query supported formats";
  bool ok=true;
  quadlet_t qModeVga=0, qModeSvga1=0, qModeSvga2=0;
  if (queryBit(qFmt,FMT_VGA)){
    ok &= (dc1394_query_supported_modes(dch_, node_, 
           FORMAT_VGA_NONCOMPRESSED, &qModeVga) == DC1394_SUCCESS);
  }
  if (queryBit(qFmt,FMT_SVGA1)){
    ok &= (dc1394_query_supported_modes(dch_, node_, 
           FORMAT_SVGA_NONCOMPRESSED_1, &qModeSvga1) == DC1394_SUCCESS);
  }
  if (queryBit(qFmt,FMT_SVGA2)){
    ok &= (dc1394_query_supported_modes(dch_, node_, 
           FORMAT_SVGA_NONCOMPRESSED_2, &qModeSvga2) == DC1394_SUCCESS);
  }
  if (!ok)
    throw "vs_grab_linux1394::Cam1394::ctor - "
          "cannot query supported modes";

  /// std::cerr <<width <<":" <<szpix <<"\n";;
  /// std::cerr <<std::hex;
  /// std::cerr <<"fmts:  " <<qFmt <<"\n";
  /// std::cerr <<"modes_vga:  " <<qModeVga <<"\n";
  /// std::cerr <<"modes_svga1:" <<qModeSvga1 <<"\n";
  /// std::cerr <<"modes_svga2:" <<qModeSvga2 <<"\n";
  /// std::cerr <<std::dec;

  ok=false;
  if (width>=1600){
    format_=FORMAT_SVGA_NONCOMPRESSED_2;
    width_=1600;
    height_=1200;
    if (szpix==0 || szpix==1){
      szpix_=1;
      if (queryBit(qModeSvga2,M1600x1200xMONO8)){
        mode_=MODE_1600x1200_MONO;
        ok=true;
      }
    } else
    if (szpix==0 || szpix==3 || szpix==6){
      szpix_=3;
      if (queryBit(qModeSvga2,M1600x1200xRGB)){
        mode_=MODE_1600x1200_RGB;
        ok=true;
      } else if (queryBit(qModeSvga2,M1600x1200xYUV422)){
        mode_=MODE_1600x1200_YUV422;
        ok=true;
      }
    } else
    if (szpix==0 || szpix==2){
      szpix_=2;
      if (queryBit(qModeSvga2,M1600x1200xMONO16)){
        mode_=MODE_1600x1200_MONO16;
        ok=true;
      }
    }
  }
  if (!ok && width>=1280){
    format_=FORMAT_SVGA_NONCOMPRESSED_2;
    width_=1280;
    height_=960;
    if (szpix==0 || szpix==1){
      szpix_=1;
      if (queryBit(qModeSvga2,M1280x960xMONO8)){
        mode_=MODE_1280x960_MONO;
        ok=true;
      }
    } else
    if (szpix==0 || szpix==3 || szpix==6){
      szpix_=3;
      if (queryBit(qModeSvga2,M1280x960xRGB)){
        mode_=MODE_1280x960_RGB;
        ok=true;
      } else if (queryBit(qModeSvga2,M1280x960xYUV422)){
        mode_=MODE_1280x960_YUV422;
        ok=true;
      }
    } else
    if (szpix==0 || szpix==2){
      szpix_=2;
      if (queryBit(qModeSvga2,M1280x960xMONO16)){
        mode_=MODE_1280x960_MONO16;
        ok=true;
      }
    }  
  }
  if (!ok && width>=1024){
    format_=FORMAT_SVGA_NONCOMPRESSED_1;
    width_=1024;
    height_=768;
    if (szpix==0 || szpix==1){
      szpix_=1;
      if (queryBit(qModeSvga1,M1024x768xMONO8)){
        mode_=MODE_1024x768_MONO;
        ok=true;
      }
    } else
    if (szpix==0 || szpix==3 || szpix==6){
      szpix_=3;
      if (queryBit(qModeSvga1,M1024x768xRGB)){
        mode_=MODE_1024x768_RGB;
        ok=true;
      } else if (queryBit(qModeSvga1,M1024x768xYUV422)){
        mode_=MODE_1024x768_YUV422;
        ok=true;
      }
    } else
    if (szpix==0 || szpix==2){
      szpix_=2;
      if (queryBit(qModeSvga1,M1024x768xMONO16)){
        mode_=MODE_1024x768_MONO16;
        ok=true;
      }
    }  
  }
  if (!ok && width>=800){
    format_=FORMAT_SVGA_NONCOMPRESSED_1;
    width_=800;
    height_=600;
    if (szpix==0 || szpix==1){
      szpix_=1;
      if (queryBit(qModeSvga1,M800x600xMONO8)){
        mode_=MODE_800x600_MONO;
        ok=true;
      }
    } else
    if (szpix==0 || szpix==3 || szpix==6){
      szpix_=3;
      if (queryBit(qModeSvga1,M800x600xRGB)){
        mode_=MODE_800x600_RGB;
        ok=true;
      } else if (queryBit(qModeSvga1,M800x600xYUV422)){
        mode_=MODE_800x600_YUV422;
        ok=true;
      }
    } else
    if (szpix==0 || szpix==2){
      szpix_=2;
      if (queryBit(qModeSvga1,M800x600xMONO16)){
        mode_=MODE_800x600_MONO16;
        ok=true;
      }
    }    
  }
  if (!ok && width>=640){
    format_=FORMAT_VGA_NONCOMPRESSED;
    width_=640;
    height_=480;
    if (szpix==0 || szpix==1){
      szpix_=1;
      if (queryBit(qModeVga,M640x480xMONO8)){
        mode_=MODE_640x480_MONO;
        ok=true;
      } else if (queryBit(qModeVga,M640x480xYUV422)){
        mode_=MODE_640x480_YUV411;
        ok=true;
      } else if (queryBit(qModeVga,M640x480xYUV411)){
        mode_=MODE_640x480_YUV422;
        ok=true;
      }
    } else
    if (szpix==2){
      szpix_=2;
      if (queryBit(qModeVga,M640x480xMONO16)){
        mode_=MODE_640x480_MONO16;
        ok=true;
      } else if (queryBit(qModeVga,M640x480xYUV422)){
        mode_=MODE_640x480_YUV411;
        ok=true;
      } else if (queryBit(qModeVga,M640x480xYUV411)){
        mode_=MODE_640x480_YUV422;
        ok=true;
      }
    } else
    if (szpix==3 || szpix==6){
      szpix_=3;
      if (queryBit(qModeVga,M640x480xRGB)){
        mode_=MODE_640x480_RGB;
        ok=true;
      } else if (queryBit(qModeVga,M640x480xYUV422)){
        mode_=MODE_640x480_YUV422;
        ok=true;
      } else if (queryBit(qModeVga,M640x480xYUV411)){
        mode_=MODE_640x480_YUV411;
        ok=true;
      }
    }  
  }
  if (!ok && width>=320){
    format_=FORMAT_VGA_NONCOMPRESSED;
    width_=320;
    height_=240;
    if (szpix==0 || szpix==1 || szpix==3 || szpix==6){
      szpix_=(szpix==1)?1:3;
      if (queryBit(qModeVga,M320x240xYUV422)){
        mode_=MODE_320x240_YUV422;
        ok=true;
      }
    } 
  }
  if (!ok && width>=160){
    format_=FORMAT_VGA_NONCOMPRESSED;
    width_=160;
    height_=120;
    if (szpix==0 || szpix==1 || szpix==3 || szpix==6){
      szpix_=(szpix==1)?1:3;
      if (queryBit(qModeVga,M160x120xYUV444)){
        mode_=MODE_160x120_YUV444;
        ok=true;
      }
    } 
  }
  //std::cerr <<width_ <<":" <<szpix_ <<"=" <<ok <<"\n";;
  //std::cerr <<format_ <<"#" <<mode_ <<"\n";;

  //last resorts:
  if (!ok && szpix>0){
    ok=negotiateFormatAndMode(width, 0);
  }
  if (!ok && szpix==0 && width!=640){
    ok=negotiateFormatAndMode(640, 0);
  }
  if (!ok && szpix==0 && width!=1280){
    ok=negotiateFormatAndMode(1280, 0);
  }
  return ok;
}

void Cam1394::setHighestFrameRate(){
  quadlet_t val;
  if (dc1394_query_supported_framerates(dch_, 
        node_, format_, mode_, &val) != DC1394_SUCCESS)
    throw "vs_grab_linux1394::Cam1394::ctor - "
          "cannot query supported framerates\n";

  //negotiate framerate 
  rate_=-1;
  if (queryBit(val,FPS3_75)){
    rate_=FRAMERATE_3_75;
    rate_ms_=267;
  }
  if (queryBit(val,FPS7_5)){
    rate_=FRAMERATE_7_5;
    rate_ms_=133;
  }
  if (queryBit(val,FPS15)){
    rate_=FRAMERATE_15;
    rate_ms_=67;
  }
  if (queryBit(val,FPS30)){
    rate_=FRAMERATE_30;
    rate_ms_=33;
  }
  if (queryBit(val,FPS60)){
    rate_=FRAMERATE_60;
    rate_ms_=17;
  }
  if (rate_==-1)
    throw "vs_grab_linux1394::Cam1394::ctor - "
          "did not obtain any applicable framerate\n";
}

void Cam1394::checkCameraParameters(){
  quadlet_t val, val2;
  dc1394bool_t ison;

  //get format
  dc1394_get_video_format(dch_, node_, &val); 
  dc1394_get_video_mode(dch_, node_, &val);
  dc1394_get_video_framerate(dch_, node_, &val);
  
  //get status
  dc1394_get_iso_channel_and_speed(dch_, node_, &val, &val2);
  dc1394_get_iso_status(dch_, node_, &ison);

  return;
}

/******************************************************
 ** Cam1394 private operations
 ******************************************************/

bool Cam1394::queryBit(
  quadlet_t val, 
  int i)
{
  assert(i<32);
  int bitmask=1<<i;
  int byte=fixQuadlet(val);
  bool rv=(byte & bitmask)!=0;
  //std::cerr <<std::hex <<val <<"#" <<byte <<"#" <<bitmask 
  //          <<"=" <<std::dec <<rv <<"\n";
  return rv;
}
int Cam1394::fixQuadlet(quadlet_t val){
  //this might change on a big endian machine!
  int src=val;
  int dst=0;
  for (int i=0; i<32; i++){
    dst <<= 1;
    if (src&1)
      dst |= 1;
    src >>= 1;
  }
  return dst;
}

// source: coriander v0.24, src/conversions.c
inline void Cam1394::yuv2rgb(
  const int& y, const int& u, const int& v,
  int &r, int &g, int& b)
{
  r = y + ((v*1436) >>10);
  g = y - ((u*352 + v*731) >> 10);
  b = y + ((u*1814) >> 10);
  r = r < 0 ? 0 : r;
  g = g < 0 ? 0 : g;
  b = b < 0 ? 0 : b;
  r = r > 255 ? 255 : r;
  g = g > 255 ? 255 : g;
  b = b > 255 ? 255 : b;
}
inline void Cam1394::rgb2yuv(
  const int &r, const int &g, const int& b,
  int& y, int& u, int& v)
{
  y = (306*r + 601*g + 117*b)  >> 10;
  u = ((-172*r - 340*g + 512*b) >> 10)  + 128;
  v = ((512*r - 429*g - 83*b) >> 10) + 128;
  y = y < 0 ? 0 : y;
  u = u < 0 ? 0 : u;
  v = v < 0 ? 0 : v;
  y = y > 255 ? 255 : y;
  u = u > 255 ? 255 : u;
  v = v > 255 ? 255 : v;
}









/******************************************************
 ** vs_grab_linux1394
 ******************************************************/

// Construction
vs_grab_linux1394::vs_grab_linux1394(const CtorArg& arg):
  port_(arg.board),
  index_(arg.channel),
  fmt_(10,10,1)
{
  Lib1394::instance().connect(
    arg.board, arg.channel, arg.mode,
    arg.fmt.width(), arg.fmt.szPixel());
}
vs_grab_linux1394::~vs_grab_linux1394(){
  Lib1394::instance().disconnect(port_, index_);
}


// vs_base operations
const img_fmt& vs_grab_linux1394::fmt(){
  Cam1394& cam(Lib1394::instance().cam(port_, index_));
  fmt_=img_fmt(cam.width(), cam.height(), cam.szpix());
  return fmt_;
}
void vs_grab_linux1394::getFrameSimple(img_data& img){
  Cam1394& cam(Lib1394::instance().cam(port_, index_));
  img.reformat(fmt());
  assert(img.fmt().szBits() == cam.szBits());
  cam.getFrame((unsigned char*)img.bits());
}
int vs_grab_linux1394::msDayUtc(){
  Cam1394& cam(Lib1394::instance().cam(port_, index_));
  return cam.msDayUtc();
}

// run-time identification
char const* vs_grab_linux1394::s_name(){
  return "linux1394";
}
char const* vs_grab_linux1394::name(){
  return s_name();
}

