#include "vs_grab_linuxV4L.hpp"

#include "../img/img_data.hpp"
#include "../img/img_util.hpp"
#include "../util/util_time.hpp"
#include "../util/util_registry.hpp"

#include <boost/smart_ptr.hpp>
#include <boost/lexical_cast.hpp>

#include <linux/videodev.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <fcntl.h>


#include <iostream>
#include <algorithm>
#include <vector>
#include <string>
#include <stdexcept>
#include <limits>

#include <assert.h>

#ifdef __INTEL_COMPILER
  unsigned int __invalid_size_argument_for_IOC=0xffffffff;
#endif

/******************************************************
 ** static registry creator object
 ******************************************************/

namespace{
  util::RegistryCreator<vs_grab,vs_grab_linuxV4L> creator;
} // unnamed namespace end

/******************************************************
 ** diagnostic helpers
 ******************************************************/

namespace{
  inline void checkParam(
    int par, int limitLo, int limitHi,
    char const* method)
  {
    if (par<limitLo || par>=limitHi){
      std::string str("vs_grab_linuxV4L::");
      str+=method;
      str+=" - got illegal argument: ";
      str+=boost::lexical_cast<std::string>(par);
      std::cerr <<str <<"\n";
      assert(0);
      throw std::runtime_error(str);
    }
  }
  inline void assertParam(int par, int val,
    char const* method)
  {
    checkParam(par, val, val+1, method);
  }
} // unnamed namespace end


/******************************************************
 ** diagnostic helpers
 ** print methods for ioctl structs 
 ******************************************************/


namespace{

void print_capability(video_capability& cap)
{
  std::cerr <<"Driver caps for: " <<cap.name <<std::endl;
  std::cerr 
    <<"    [ can capture to memory:\t\t\t| " 
    <<((cap.type & VID_TYPE_CAPTURE)?    "X" : " ") 
    <<" ]" <<std::endl;
  std::cerr 
    <<"    [ has a tuner of some form:\t\t\t| " 
    <<((cap.type & VID_TYPE_TUNER)?      "X" : " ") 
    <<" ]" <<std::endl;
  std::cerr 
    <<"    [ has teletext capability:\t\t\t| " 
    <<((cap.type & VID_TYPE_TELETEXT)?   "X" : " ") 
    <<" ]" <<std::endl;
  std::cerr 
    <<"    [ can overlay images onto frame buffer:\t| " 
    <<((cap.type & VID_TYPE_OVERLAY)?    "X" : " ") 
    <<" ]" <<std::endl;
  std::cerr
    <<"    [ overlay is chromakeyed:\t\t\t| " 
    <<((cap.type & VID_TYPE_CHROMAKEY)?  "X" : " ") 
    <<" ]" <<std::endl;
  std::cerr
    <<"    [ overlay clipping is supported:\t\t| " 
    <<((cap.type & VID_TYPE_CLIPPING)?   "X" : " ") 
    <<" ]" <<std::endl;
  std::cerr
    <<"    [ overlay overwrites frame buffer memory:\t| " 
    <<((cap.type & VID_TYPE_FRAMERAM)?   "X" : " ") 
    <<" ]" <<std::endl;
  std::cerr
    <<"    [ hardware supports image scaling:\t\t| " 
    <<((cap.type & VID_TYPE_SCALES)?     "X" : " ") 
    <<" ]" <<std::endl;
  std::cerr
    <<"    [ image capture is greyscale only:\t\t| " 
    <<((cap.type & VID_TYPE_MONOCHROME)? "X" : " ") 
    <<" ]" <<std::endl;
  std::cerr
    <<"    [ capture can be of only part of the image:\t| " 
    <<((cap.type & VID_TYPE_SUBCAPTURE)? "X" : " ") 
    <<" ]" <<std::endl;

  std::cerr <<"  channels: "  <<cap.channels  <<std::endl;
  std::cerr <<"  audios: "    <<cap.channels  <<std::endl;
  std::cerr <<"  maxwidth: "  <<cap.maxwidth  <<std::endl;
  std::cerr <<"  maxheight: " <<cap.maxheight <<std::endl;
  std::cerr <<"  minwidth: "  <<cap.minwidth  <<std::endl;
  std::cerr <<"  minheight: " <<cap.minheight <<std::endl;
}


void print_channel(video_channel& ch)
{
  std::cerr <<"Channel caps for #" <<ch.channel <<std::endl;
  std::cerr <<"  name: "   <<ch.name    <<std::endl;
  std::cerr <<"  tuners: " <<ch.tuners  <<std::endl;
  std::cerr <<"  flags: "  <<std::endl;
  std::cerr 
    <<"   [ channel has tuners:\t\t\t| "
    <<((ch.flags & VIDEO_VC_TUNER)? "X" : " ")
    <<" ]" <<std::endl;
  std::cerr 
    <<"   [ channel has audio: \t\t\t| "
    <<((ch.flags & VIDEO_VC_AUDIO)? "X" : " ")
    <<" ]" <<std::endl;
  std::cerr 
    <<"   [ input is a tv input:\t\t\t| "
    <<((ch.flags & VIDEO_TYPE_TV)? "X" : " ")
    <<" ]" <<std::endl;
  std::cerr 
    <<"   [ input is a camera: \t\t\t| " 
    <<((ch.flags & VIDEO_TYPE_CAMERA)? "X" : " ")
    <<" ]" <<std::endl;
  std::cerr <<"  norm: " <<ch.norm <<std::endl;
}


void print_picture(video_picture& vpic){
  std::cerr <<"Picture settings" <<std::endl;
  std::cerr <<"  brightness: "<<vpic.brightness<<std::endl;
  std::cerr <<"  hue: "       <<vpic.hue       <<std::endl;
  std::cerr <<"  colour: "    <<vpic.colour    <<std::endl;
  std::cerr <<"  contrast: "  <<vpic.contrast  <<std::endl;
  std::cerr <<"  whiteness: " <<vpic.whiteness <<std::endl;
  std::cerr <<"  depth: "     <<vpic.depth     <<std::endl;
  std::cerr <<"  palette: "   <<vpic.palette   <<std::endl;
}
void print_window(video_window& vcwin){
  std::cerr <<"Window settings" <<std::endl;
  std::cerr <<"  x-offset: "  <<vcwin.x         <<std::endl;
  std::cerr <<"  y-offset: "  <<vcwin.y         <<std::endl;
  std::cerr <<"  width: "     <<vcwin.width     <<std::endl;
  std::cerr <<"  height: "    <<vcwin.height    <<std::endl;
  std::cerr <<"  chromakey: " <<vcwin.chromakey <<std::endl;
  std::cerr <<"  flags: "     <<vcwin.flags     <<std::endl;
}
void print_capture(video_capture& vcpt){
  std::cerr <<"Capture settings" <<std::endl;
  std::cerr <<"  x-offset: "  <<vcpt.x         <<std::endl;
  std::cerr <<"  y-offset: "  <<vcpt.y         <<std::endl;
  std::cerr <<"  width: "     <<vcpt.width     <<std::endl;
  std::cerr <<"  height: "    <<vcpt.height    <<std::endl;
  std::cerr <<"  decimation: "<<vcpt.decimation<<std::endl;
  std::cerr <<"  flags: "     <<vcpt.flags     <<std::endl;
}


void print_mbuf(video_mbuf& mbuf)
{
  std::cerr <<"Memory buffer settings" <<std::endl;
  std::cerr <<"  size: "   <<mbuf.size       <<std::endl;
  std::cerr <<"  frames: " <<mbuf.frames     <<std::endl;
  std::cerr <<"  offset: " <<mbuf.offsets[1] <<std::endl;
}

void print_fmt(char const* intro, const img_fmt& f)
{
  std::cerr <<intro <<std::endl;
  std::cerr <<"  width: "  <<f.width()   <<std::endl;
  std::cerr <<"  height: " <<f.height()  <<std::endl;
  std::cerr <<"  szpix: "  <<f.szPixel() <<std::endl;
}

void print_mmap(char const* intro, video_mmap& vm)
{
  std::cerr <<intro <<std::endl;
  std::cerr <<"  frame: "  <<vm.frame   <<std::endl;
  std::cerr <<"  height: " <<vm.height  <<std::endl;
  std::cerr <<"  width: "  <<vm.width   <<std::endl;
  std::cerr <<"  format: " <<vm.format  <<std::endl;
}

} // unnamed namespace end


/******************************************************
 ** helper struct: FrameBuffer
 ******************************************************/
struct FrameBuffer{
  unsigned char* bits;
  int   time;
};

/******************************************************
 ** private implementation class, 
 ** encapsulates a single v4l device
 ******************************************************/

class V4LDevice{
  friend class V4LDeviceMgr;
  typedef unsigned char* PBITS;
private:
  V4LDevice(int board, int channel, int mode, const img_fmt& f);
  ~V4LDevice();
public:
  const img_fmt& fmt();
  void getFrame(unsigned char*& , int& time);
private: // getFrame operations
  void getFrameMmap(unsigned char*& , int& time);
  void getFrameRead(unsigned char*& , int& time);
  void startCapture(int fb);
  bool frameArriving(int fb);
  void waitCapture(int fb);
private: // frame obsolescence threshold [ms]
  enum {tFreshMs_=100};
public:
  int board(){return board_;}
  void setChannel(int channel, int mode);
  void setFmt(const img_fmt& fmt);
private:
  void getFmt();
private:
  int board_;
  bool useDMA_;
  img_fmt fmt_;

  std::string device_;

  int handle_;
  video_capability vcap_;
  video_channel    vchn_;
  video_picture    vpic_;
  video_window     vcwin_;
  video_capture    vcpt_;
  video_mbuf       vmbuf_;
  video_mmap       vmap_;

  PBITS bufMmap_;
  FrameBuffer fb_[2];
  int fbIndex_;
};

/******************************************************
 ** private singleton class for managing v4l devices
 ******************************************************/
class V4LDeviceMgr{
  V4LDeviceMgr(){}
  ~V4LDeviceMgr();
public:
  static V4LDeviceMgr& instance(){
    static V4LDeviceMgr mgr;
    return mgr;
  }
public:
  V4LDevice& connect(
    int board, int channel, int mode, 
    const img_fmt& f);
  void release(int board);
private:
  std::vector<V4LDevice*> map_;
};

V4LDeviceMgr::~V4LDeviceMgr(){
  //check for dangling connections
  for (unsigned i=0; i<map_.size(); i++){
    if (map_[i]){
      std::cerr <<"V4LDeviceMgr::~V4LDeviceMgr - "
                  "fixing connection " <<i <<"\n";
      release(i);
    }
  }
}

V4LDevice& V4LDeviceMgr::connect(
  int board, int channel, int mode, const img_fmt& f)
{
  if (board<static_cast<int>(map_.size())){
    if (map_[board]){
      throw "vs_grab_linuxV4L::V4LDeviceMgr::connect"
            " - the specified device is already in use.";
    }
  } else{
    map_.resize(board+1);
  }
  assert(board<static_cast<int>(map_.size()));

  map_[board]=new V4LDevice(board, channel, mode, f);
  return *map_[board];
}

void V4LDeviceMgr::release(int index){
  assert(map_[index]);
  delete map_[index];
  map_[index]=0;
}


   
/******************************************************
 ** V4LDevice members
 ******************************************************/

V4LDevice::V4LDevice(
  int board, int channel, int mode,
  const img_fmt& f)
:
  board_(board), 
  fmt_(320,240,1),
  fbIndex_(0)
{
  device_="/dev/video";
  device_+=boost::lexical_cast<std::string>(board);
  handle_ = open(device_.c_str(),O_RDWR);
  assertParam(handle_==-1, 0, "V4LDevice::V4LDevice");

  // check caps
  int rv=ioctl(handle_, VIDIOCGCAP, &vcap_);
  assertParam(rv, 0, "V4LDevice::V4LDevice::VIDIOCGCAP");
  print_capability(vcap_);
  assertParam(vcap_.type && VID_TYPE_CAPTURE, 1, 
    "V4LDevice::V4LDevice::VID_TYPE_CAPTURE");

  // set video source
  setChannel(channel, mode);
  
  // set the format
  setFmt(f);

  // check for DMA
  useDMA_=true;
  if (useDMA_){
    ioctl(handle_,VIDIOCGMBUF,&vmbuf_) ;
    print_mbuf(vmbuf_);
    useDMA_= vmbuf_.frames >= 2;
  }

  // init frame buffer structs
  fb_[0].time=0;
  fb_[1].time=0;
  if (useDMA_){
    // map driver buffers
    bufMmap_ = (unsigned char *)mmap(0,vmbuf_.size,
      PROT_READ|PROT_WRITE,MAP_SHARED, handle_,0);
    for (int i=0; i<2; ++i){
      fb_[i].bits=bufMmap_+vmbuf_.offsets[i];
    }
  } else{
    fb_[0].bits=static_cast<unsigned char*>(malloc(fmt_.szBits()));
    fb_[1].bits=0;
  }
}

V4LDevice::~V4LDevice(){
  if (useDMA_){
    munmap(bufMmap_, vmbuf_.size);
  } else{
    free(fb_[0].bits);
  }
  close(handle_);
}

void V4LDevice::setChannel(int ch, int mode)
{
  int rv=ioctl(handle_, VIDIOCGCHAN, &vchn_);
  assertParam(rv, 0, "V4LDevice::V4LDevice::VIDIOCGCHAN");

  vchn_.channel=ch;  
  vchn_.norm=mode;
  //print_channel(vchn_);

  rv=ioctl(handle_, VIDIOCSCHAN, &vchn_);
  rv=ioctl(handle_, VIDIOCGCHAN, &vchn_);
  assertParam(rv, 0, "V4LDevice::V4LDevice::VIDIOCGCHAN");
  //print_channel(vchn_);
}


void V4LDevice::getFmt()
{
  int rv=ioctl(handle_, VIDIOCGPICT, &vpic_);
  assertParam(rv, 0, "V4LDevice::V4LDevice::VIDIOCGPICT");
  print_picture(vpic_);

}


void V4LDevice::setFmt(
  const img_fmt& f)
{
  // overlay window size is specified here (not used)
  if (false){
    int rv=ioctl(handle_, VIDIOCGWIN, &vcwin_);
    assertParam(rv, 0, "V4LDevice::setFmt::VIDIOCGWIN");
    print_window(vcwin_);
  }

  // the desired capture area is defined here:
  if (vcap_.type & VID_TYPE_SUBCAPTURE){
    int rv=ioctl(handle_, VIDIOCGCAPTURE, &vcpt_);
    assertParam(rv, 0, "V4LDevice::V4LDevice::VIDIOCGCAPTURE");
  
    vcpt_.x=0;  
    vcpt_.y=0;  
    vcpt_.width= vcap_.maxwidth;  
    vcpt_.height=vcap_.maxheight;  
    vcpt_.decimation=0;
    //check the single field grab possibility (NOT TESTED!)
    if (vcpt_.height<vcap_.maxheight/static_cast<unsigned>(2)){
      vcpt_.flags=VIDEO_CAPTURE_ODD;
    }

    rv=ioctl(handle_, VIDIOCSCAPTURE, &vcpt_);
    rv=ioctl(handle_, VIDIOCGCAPTURE, &vcpt_);
    assertParam(rv, 0, "V4LDevice::V4LDevice::VIDIOCGCAPTURE");
  }

  // the desired type and size of the obtained image is specified here
  // (camera might do the transform from the captured image!)
  int pixelType=0;
  int pixelSize=0;
  switch (f.szPixel()){
  case 1:
    pixelType=VIDEO_PALETTE_GREY;
    pixelSize=1;
    break;
  case 2:
    pixelType=VIDEO_PALETTE_GREY;
    pixelSize=1;
    break;
  case 3:
    pixelType=VIDEO_PALETTE_RGB24;
    pixelSize=3;
    break;
  case 4:
    pixelType=VIDEO_PALETTE_RGB32;
    pixelSize=4;
    break;
  }
  assertParam(pixelType*pixelSize==0,0,
    "V4LDevice::setFmt::VIDEO_PALETTE"); 

  // cvsh image format
  int w=std::min(f.width(), vcap_.maxwidth);
  int h=std::min(f.height(), vcap_.maxheight);
  fmt_=img_fmt(w,h, pixelSize);
  print_fmt("The requested format", fmt_);

  // v4l image format
  vmap_.frame  = 0 ;
  vmap_.width  = fmt_.width();
  vmap_.height = fmt_.height();
  vmap_.format = pixelType;
  print_mmap("The resulting mmap", vmap_);
}

// private helpers
void V4LDevice::startCapture(int fb){
  fb_[fb].time=util_time::msDayUtc();
  vmap_.frame=fb;
  ioctl(handle_,VIDIOCMCAPTURE,&vmap_);
}

bool V4LDevice::frameArriving(int fb){
  int dt=util_time::msDayUtc()-fb_[fb].time;
  return dt<tFreshMs_;
}

void V4LDevice::waitCapture(int fb){
  assert (fb_[fb].time>0);
  ioctl(handle_,VIDIOCSYNC,&fb); 
}


const img_fmt& V4LDevice::fmt(){
  return fmt_;
}

void V4LDevice::getFrame(
  unsigned char*& pBits, 
  int& msDayUtc)
{
  if (useDMA_){
    getFrameMmap(pBits, msDayUtc);
  } else{
    getFrameRead(pBits, msDayUtc);
  }
}

void V4LDevice::getFrameMmap(
  unsigned char*& pBits, 
  int& msDayUtc)
{
  int fbIndexOther=(fbIndex_+1)%2;
  if (frameArriving(fbIndex_)){
    startCapture(fbIndexOther);
    waitCapture(fbIndex_);
  } else{
    std::cerr <<"V4LDevice::getFrameMmap init\n";
    startCapture(fbIndex_);
    waitCapture(fbIndex_);
    startCapture(fbIndexOther);
  }
  pBits=fb_[fbIndex_].bits;
  msDayUtc=fb_[fbIndex_].time;
  fb_[fbIndex_].time=0;
  fbIndex_=fbIndexOther;
}

void V4LDevice::getFrameRead(
  unsigned char*& pBits, 
  int& msDayUtc)
{
  int t0=util_time::msDayUtc();
  int sz=fmt_.width()*fmt_.height()*fmt_.szPixel();
  int rv=read(handle_, fb_[0].bits, sz);
  assertParam(rv, sz, "V4LDevice::getFrameRead::read");
  int t1=util_time::msDayUtc();

  pBits=fb_[0].bits;
  msDayUtc=(t0+t1)/2;
}





/******************************************************
 ** vs_grab_linuxV4L
 ******************************************************/

// Construction
vs_grab_linuxV4L::vs_grab_linuxV4L(
  const CtorArg& arg)
:
  cam(V4LDeviceMgr::instance().
    connect(arg.board, arg.channel, arg.mode, arg.fmt))
{
}

vs_grab_linuxV4L::~vs_grab_linuxV4L(){
  V4LDeviceMgr::instance().release(cam.board());
}

// vs_base Attributes
char const* vs_grab_linuxV4L::name(){
  return s_name();
}

// vs_base operations
const img_fmt& vs_grab_linuxV4L::fmt(){
  return cam.fmt();
}

void vs_grab_linuxV4L::getFrameSimple(img_data& img){
  const img_fmt& f=fmt();
  img.reformat(f);

  unsigned char* pbits;
  cam.getFrame(pbits, msLastFrame_);
  // img.setBits(pbits, f.szBits());

  // transfer bits with vertical flip
  int szLine=f.width()*f.szPixel();
  unsigned char* psrc=pbits+(f.height()-1)*szLine;
  for (int i=0; i<fmt().height(); ++i){
    img.setLine(i, psrc);
    psrc-=szLine;
  }

  // flip component bytes in individual pixels
  if (fmt().szPixel()==3){
    img_util::flipEndian3(img);
  } else if (fmt().szPixel()==4){
    img_util::flipEndian4(img);
  }

}



// vs_grab attributes
char const* vs_grab_linuxV4L::s_name(){
  return "linuxV4L";
}

