#include "vs_grab_macx1394.hpp"
#include "vs_grabUtil.hpp"

#include "../img/img_data.hpp"
#include "../util/util_time.hpp"
#include "../util/util_registry.hpp"

#define TARGET_API_MAC_CARBON 1
#include <Carbon/Carbon.h>
#include <QuickTime/QuickTime.h>

#include <string>
#include <sstream>
#include <iostream>
#include <stdexcept>

#include <assert.h>


// implementation details
//  - only yuv422 cameras are supported (deffective drivers from apple)
//  - the board and mode ctor arguments are currently unused


////////////////////////////////////
// static registry creator object

namespace{
  util::RegistryCreator<vs_grab,vs_grab_macx1394> creator;
} // unnamed namespace end


//////////////////////////////
// MacOS helpers

namespace{
  void CE(OSErr err, char const* str){
    if (err!=noErr){
      std::ostringstream oss;
      oss <<"MAC OS-X error on " <<str <<": " <<err;
      throw std::runtime_error(oss.str());
    }
  }
  std::string p2cxx(ConstStr255Param pstr){
    char cstr[64];
    p2cstrcpy(cstr,pstr); 
    return std::string (cstr);
  }
  std::string fc2cxx(OSType l){
    char buf[sizeof(OSType)];
    *((OSType*)buf)=l;
    return std::string (buf, sizeof(OSType));
  }
  std::ostream& operator<<(std::ostream& os, const CompTimeValue& tv){
    os <<"t(" <<tv.hi <<":"  <<tv.lo <<')';
    return os;
  }
  std::ostream& operator<<(std::ostream& os, const TimeRecord& t){
    os <<t.value <<"/" <<t.scale <<":" <<t.base;
    return os;
  }
}



///////////////////////
// simple loger component

class Logger{
  std::ostream& os_;
public:
  std::ostream& os(){return os_;}
  Logger(std::ostream& os):
    os_(os)
  {}
};
template <class T> 
Logger& operator<<(Logger& log, const T& t){
  if (false)
    log.os() <<t;
  return log;
}
Logger log_(std::cerr);




////////////////////////////////////////
//private implementation class

class MacXfwcam{
  img_fmt fmt_;
  VideoDigitizerComponent vdc_;
  ComponentInstance clk_;
  int countInputs_;
  int w_;
  int h_;
  int szpix_;
  int szdata_;
  TimeRecord tr0_;
  TimeRecord tr_;
public:
  MacXfwcam(int w, int h, int szpix, int ch);
  ~MacXfwcam();
  const img_fmt& fmt(){return fmt_;}
  void acquire(void* buf);
  int msDayUtc();
};
MacXfwcam::~MacXfwcam(){
  log_ <<"In ~MacXfwcam\n";
}
MacXfwcam::MacXfwcam(int w, int h, int szpix, int ch):
  fmt_(0,0,1)
{
  //open components
  CE(OpenADefaultComponent(
      clockComponentType, systemMicrosecondClock, &clk_), 
    "OpenADefaultComponent");
  CE(OpenADefaultComponent(
      videoDigitizerComponentType, vdSubtypeIIDC, &vdc_), 
    "OpenADefaultComponent");

  //channel inputs
  short input;
  CE(VDGetNumberOfInputs(vdc_,&input), "VDGetNumberOfInputs");
  countInputs_=input+1;
  if (countInputs_==0)
    throw std::runtime_error("MacXfwcam::MacXfwcam - no connected cameras");
  log_ <<"MacXfwcam inputs\n";
  log_ <<"  count:    " <<countInputs_ <<"\n";
  if (ch>=countInputs_) ch=countInputs_-1;
  CE(VDSetInput(vdc_, ch), "VDSetInput");
  CE(VDGetInput(vdc_, &input), "VDGetInput");
  log_ <<"  channel:  " <<input <<"\n";
  UInt64 fl1,fl2;
  CE(VDGetUniqueIDs(vdc_, &fl1,&fl2), "VDGetUniqueIDs");
  std::ostringstream oss;
  oss <<std::hex <<fl1 <<":" <<fl2;
  log_ <<"  camera id:" <<oss.str() <<"\n";

  //check codec, set codec and rect
  int codecIndex=-1;
  VDCompressionListHandle hclp= (VDCompressionListHandle)NewHandle(0);
  CE(VDGetCompressionTypes (vdc_, hclp), "VDGetCompressionTypes");
  int sz=GetHandleSize((Handle)hclp)/sizeof(VDCompressionList);
  log_ <<"MacXfwcam CompressionTypes:" <<sz <<"\n";
  VDCompressionList* clp=*hclp;
  for (int i=0; i<sz; ++i){
    if (clp[i].cType==k422YpCbCr8CodecType){
      codecIndex=i;
    }
    log_ <<"  codec type     " <<i <<": " <<fc2cxx(clp[i].cType) <<"\n";
    log_ <<"  codec typeName " <<i <<": " <<p2cxx(clp[i].typeName) <<"\n";
    log_ <<"  codec name     " <<i <<": " <<p2cxx(clp[i].name) <<"\n";
  }
  if (codecIndex==-1){
    throw std::runtime_error("MacXfwcam::MacXfwcam - camera does not support"
                             "the only codec supported by apple (2yuv)");
  }
  DisposeHandle((Handle)hclp);
  Rect rcSrc={0,0,w,h};
  CE(VDSetCompression(vdc_, k422YpCbCr8CodecType, 24, &rcSrc, 0,0,0),"VDSetCompression");

  //get image description
  ImageDescriptionHandle hid=(ImageDescriptionHandle)NewHandle(0);
  CE(VDGetImageDescription (vdc_,hid), "VDGetImageDescription");;
  ImageDescription* pid=*hid;
  w_ = pid->width;
  h_ = pid->height;
  if (szpix==1) 
    szpix_=1;
  else
    szpix_=3;
  szdata_= pid->dataSize;
  DisposeHandle((Handle)hid);
  log_ <<"The source image is: " 
    <<w_ <<"x" <<h_ <<" (" <<szdata_ <<")\n";

  //set time base
  TimeBase tb=NewTimeBase ();
  SetTimeBaseMasterClock(tb, (Component)clk_, 0);
  SetTimeBaseRate(tb, Long2Fix(1));
  TimeRecord trz;
  GetTimeBaseTime(tb, 0, &trz);
  CE(ClockGetTime (clk_,&trz), "ClockGetTime");
  SetTimeBaseTime(tb, &trz);
  CE(VDSetTimeBase (vdc_, tb), "VDSetTimeBase"); 

  //start grabbing and jump over the first two frames
  fmt_=img_fmt(w_,h_,szpix_);
  img_data img(fmt_);
  log_ <<"Started grabbing";
  CE(VDSetCompressionOnOff(vdc_, TRUE),"VDSetCompressionOnOff");
  for (int i=0; i<2; ++i) {
    log_ <<".";
    unsigned char qfc, sim;
    long  szdata;
    Ptr pdata;
    CE(VDCompressOneFrameAsync(vdc_),"VDCompressOneFrameAsync");
    do{
      CE(VDCompressDone(
          vdc_, &qfc, &pdata, &szdata, &sim, &tr_),
        "VDCompressDone");
    } while (qfc==0);
    CE(VDReleaseCompressBuffer(vdc_, pdata),"VDReleaseCompressBuffer");
  }
  log_ <<".\n";

  tr0_=tr_;
}

void MacXfwcam::acquire(void* pdst){
  unsigned char qfc, sim;
  long  szdata;
  Ptr pdata;

  while (1){
    log_ <<".";
    CE(VDCompressOneFrameAsync(vdc_),"VDCompressOneFrameAsync");
    do{
      CE(VDCompressDone(
          vdc_, &qfc, &pdata, &szdata, &sim, &tr_),
        "VDCompressDone");
    } while (qfc==0);
    assert(szdata==szdata_);

    TimeScale ts1k=1000;
    ConvertTimeScale (&tr_, ts1k);

    TimeRecord trclk;
    CE(ClockGetTime (clk_,&trclk), "ClockGetTime");
    ConvertTimeScale (&trclk, ts1k);

    int delta=trclk.value.lo-tr_.value.lo;
    // SubtractTime (&trclk,&tr_);
    // int delta=trclk.value.lo;

    if (delta<50) break;
    
    CE(VDReleaseCompressBuffer(vdc_, pdata),"VDReleaseCompressBuffer");
  }

  if (szpix_==1)
    vs_grabUtil::yuv422ToGray(w_,h_,pdata, pdst);
  else
    vs_grabUtil::yuv422ToRgb(w_,h_,pdata, pdst);

  CE(VDReleaseCompressBuffer(vdc_, pdata),"VDReleaseCompressBuffer");

  vs_grabUtil::flipVertical(w_,h_,szpix_, pdst);
}

int MacXfwcam::msDayUtc(){
  TimeRecord tr=tr_;
  SubtractTime (&tr,&tr0_);
  return tr.value.lo;
}







/******************************************************
 ** vs_grab_macx1394
 ******************************************************/

// Construction
vs_grab_macx1394::vs_grab_macx1394(
  const CtorArg& arg)
:
  pcam(new MacXfwcam(
    arg.fmt.width(), arg.fmt.height(), 
    arg.fmt.szPixel(), arg.channel))
{
}

vs_grab_macx1394::~vs_grab_macx1394(){
}

// vs_base Attributes
char const* vs_grab_macx1394::name(){
  return s_name();
}

// vs_base operations
const img_fmt& vs_grab_macx1394::fmt(){
  return pcam->fmt();
}
void vs_grab_macx1394::getFrameSimple(img_data& img){
  img.reformat(fmt());
  pcam->acquire(img.bits());
}
int vs_grab_macx1394::msDayUtc(){
  return pcam->msDayUtc();
}

// vs_grab attributes
char const* vs_grab_macx1394::s_name(){
  return "macx1394";
}


