#include "vs_grab_w32MCI.hpp"

#include "../util/util_bmp.hpp"
#include "../util/util_registry.hpp"

#include "../img/img_data.hpp"
#include "../img/img_util.hpp"

#include <assert.h>
#include <time.h>
#include <limits.h>

#include <string>
#include <map>
#include <stdexcept>

// Exclude rarely-used stuff from Windows headers
#define VC_EXTRALEAN		
#ifndef STRICT
#define STRICT
#endif
#include <windows.h> 

#define NOMSACM 
#define NOMMREG 
#include <vfw.h>

/******************************************************
**        vs_grab_w32MCI.CPP  by Sinisa Segvic
** Implementacijski file za vs_grab_w32MCI klasu.
*******************************************************/



/////////////////////////////////////
// static registry creator object

namespace{
  util::RegistryCreator<vs_grab,vs_grab_w32MCI> creator;
}



/////////////////////////////////////
// BITMAPINFO wrapper
class BiWrap{
public:
  BiWrap(int w, int h, int bpp);
  int size();
public:
  BITMAPINFOHEADER bih;
  RGBQUAD bmiColors[256];
};

BiWrap::BiWrap(
  int w, int h, int bpp)
{
  assert(sizeof(bih)==40);

  bih.biSize=sizeof(bih);
  bih.biPlanes=1;
  bih.biXPelsPerMeter=0;
  bih.biYPelsPerMeter=0;

  bih.biWidth=w;
  bih.biHeight=h;
  bih.biBitCount=static_cast<WORD>(bpp);
  bih.biSizeImage=w*h*bpp/8;
  bih.biClrImportant=0;

  switch (bpp){
  case 8:{
    bih.biCompression=BI_RGB;
    bih.biClrUsed=256;
    for (int i=0; i<256; i++){
      bmiColors[i].rgbBlue=
      bmiColors[i].rgbGreen=
      bmiColors[i].rgbRed=
        static_cast<BYTE>(i);
      bmiColors[i].rgbReserved=0;
    }
    break;
  }
  case 16:
    bih.biCompression=BI_BITFIELDS;
    bih.biClrUsed=3;
    bmiColors[0].rgbReserved=bmiColors[0].rgbRed=0;
    bmiColors[0].rgbGreen=0x3f;
    bmiColors[0].rgbBlue=0xff;
    bmiColors[1].rgbReserved=bmiColors[1].rgbRed=0;
    bmiColors[1].rgbGreen=0x40;
    bmiColors[1].rgbBlue=0x00;
    bmiColors[2].rgbReserved=bmiColors[2].rgbRed=0;
    bmiColors[2].rgbGreen=0x80;
    bmiColors[2].rgbBlue=0x00;
    break;
  case 24:
  case 32:
    bih.biCompression=BI_RGB;
    bih.biClrUsed=0;
    break;
  default:
    throw std::runtime_error("BiWrap::BiWrap - illegal bpp!");
  }
}

int BiWrap::size(){
  return sizeof(bih)+bih.biClrUsed*sizeof(RGBQUAD);
}

///////////////////////////////////////////////
// data provided by driver via callback functions
// is stored in Win32MCIdata objects for each MCI session

struct Win32MCIdata{
  int nID;
  char* lpStatusText;
  int nErrId;
  char* lpErrorText;
  void* ptrBits;
  int sizeofBits;
};

///////////////////////////////////////////////
// encapsulation of static callback functions

class Win32MCIinfo{
  typedef std::map<HWND,Win32MCIdata> Map;
  static Map map_;
public:
  void checkIn(HWND capw);
  void checkOut(HWND capw);
public:
  static Win32MCIdata& data(HWND);
private:
  static long statusCB(HWND, int, char *);
  static long errorCB(HWND, int, char *);
  static long frameCB(HWND, LPVIDEOHDR);
};

void Win32MCIinfo::checkIn(HWND capw){
  map_.insert(Map::value_type(capw, Win32MCIdata()));
  // Register the status callback function. 
  int rv=capSetCallbackOnStatus(capw, statusCB);
  // Register the error callback function 
  rv=capSetCallbackOnError(capw, errorCB);
  // Register the error callback function 
  rv=capSetCallbackOnFrame(capw, frameCB);
}

void Win32MCIinfo::checkOut(HWND capw){
  map_.erase(capw);
  int rv=capSetCallbackOnStatus(capw, 0);
  rv=capSetCallbackOnError(capw, 0);
  rv=capSetCallbackOnFrame(capw, 0);
}

Win32MCIdata& Win32MCIinfo::data(HWND h){
  assert(map_.find(h)!=map_.end());
  return map_[h];
}

long Win32MCIinfo::statusCB(
  HWND hWnd, int n, char *s)
{
  Win32MCIdata& d=data(hWnd);
  d.nID=n;
  d.lpStatusText=s;
  SetWindowText(hWnd, s);
  //SetWindowText(GetParent(hWnd), s);
  return 1;
}

long Win32MCIinfo::errorCB(HWND hWnd, int n, char *s){
  Win32MCIdata& d=data(hWnd);
  d.nErrId=n;
  d.lpErrorText=s;
  SetWindowText(hWnd, s);
  //SetWindowText(GetParent(hWnd), s);
  return 1;
}

long Win32MCIinfo::frameCB(HWND hWnd, LPVIDEOHDR lpVHdr){
  Win32MCIdata& d=data(hWnd);
  d.ptrBits=lpVHdr->lpData;
  d.sizeofBits=lpVHdr->dwBytesUsed;
  return 1;
}

// the only Win32MCIinfo object...
Win32MCIinfo info_;
// and its static data member
Win32MCIinfo::Map Win32MCIinfo::map_;



///////////////////////////////////////////////
// Worker class

class vs_grab_w32MCIimp{
  HWND parent_;
  HWND capw_;

  CAPDRIVERCAPS caps_; 
  CAPSTATUS     status_;
  CAPTUREPARMS  params_;

  img_fmt    fmt_;
  BiWrap     bi_;
  BITMAPINFO *pbi_;
  int        sz_bi_;

  char name_[128];
  char version_[64];

public:
//construction
  vs_grab_w32MCIimp(HWND parent, int w, int h, int bpp, 
    int board, int channel);
  ~vs_grab_w32MCIimp();

//attributes
  const img_fmt& fmt();

//operations
  bool boxDisplay();
  bool boxSource();
  bool boxFormat();

  void liveOn();
  void liveOff();

  int getFrame(void* buf);

private:
  void initWnd();
  void closeWnd();
  void getDriverFormat();
};


////////////////////////////////////////
// worker construction

vs_grab_w32MCIimp::vs_grab_w32MCIimp(
  HWND parent, int w, int h, int bpp, 
  int board, int channel)
:
  parent_(parent),
  fmt_(w, h, bpp/8),
  bi_(w, h, bpp)
{
  //check for parameter qui pro quo
  assert(bpp<=32);

  // int flags=WS_VISIBLE;
  int flags=0;
  if (parent_!=0)
    flags |= WS_CHILD;
  capw_ = capCreateCaptureWindow ("Greb greb greb", 
    flags, 0, 0, w, h, parent_, board);
  if (capw_==0)
    throw "vs_grab_w32MCIimp::vs_grab_w32MCIimp"
          " - couldn't open capture window.";

  int rv=capDriverGetName(capw_, name_, sizeof(name_)); 
  rv=capDriverGetVersion(capw_, version_, sizeof(version_)); 

  if (!capDriverConnect (capw_, board)){
    closeWnd();
    throw "vs_grab_w32MCI::vs_grab_w32MCI"
          " - couldn't connect specified capture driver.";
  }

  //pametniji popusta...
  pbi_=0;
  sz_bi_=bi_.size();
  if (!capSetVideoFormat(capw_, &bi_, sz_bi_)){
    getDriverFormat();
    if (!capSetVideoFormat(capw_, pbi_, sz_bi_)){
      delete pbi_;
      closeWnd();
      throw "vs_grab_w32MCI::vs_grab_w32MCI - "
            "couldn't set video format.";
    }
  }
  initWnd();
  info_.checkIn(capw_);

  if (channel!=0){
    boxSource();
  }

} 
vs_grab_w32MCIimp::~vs_grab_w32MCIimp(){
  int rv=capDriverDisconnect (capw_); 
  rv=0; //warning killer
  delete pbi_;
  closeWnd();
}

void vs_grab_w32MCIimp::closeWnd(){
  info_.checkOut(capw_);
  int rv=DestroyWindow(capw_);
  rv=0; //warning killer
  capw_=0;
}

void vs_grab_w32MCIimp::initWnd(){
  int rv=capDriverGetCaps(capw_, &caps_, 
    sizeof(CAPDRIVERCAPS)); 
  rv=capGetStatus(capw_, &status_, 
    sizeof (CAPSTATUS)); 
  rv=capCaptureGetSetup(capw_, &params_, 
    sizeof(CAPTUREPARMS));

  rv=SetWindowPos(capw_, NULL, 
      //ovi brojevi imaju smisla
      //ako se radi o child prozoru
      status_.uiImageWidth, 
      30,
      //ovi su bitni i inace 
      status_.uiImageWidth, 
      status_.uiImageHeight, 
      SWP_NOZORDER|SWP_NOMOVE);
  rv=0; //warning killer
}


////////////////////////////////////////
// worker attributes

const img_fmt& vs_grab_w32MCIimp::fmt(){
  if (pbi_)
    fmt_=img_fmt(
      pbi_->bmiHeader.biWidth,
      pbi_->bmiHeader.biHeight,
      pbi_->bmiHeader.biBitCount/8);
  else
    fmt_=img_fmt(
      bi_.bih.biWidth,
      bi_.bih.biHeight,
      bi_.bih.biBitCount/8);
  return fmt_;
}


////////////////////////////////////////
// worker operations

void vs_grab_w32MCIimp::liveOn(){
  //sitnice koje zivot znace...
  int rv=capPreviewRate(capw_, 1);
  rv=capPreviewScale(capw_, 0);
  rv=capPreview(capw_, 1);
  rv=capOverlay(capw_, 0);
  rv=0; //warning killer
}
void vs_grab_w32MCIimp::liveOff(){
  //sitnice koje zivot znace...
  int rv=capPreview(capw_, 0);
  rv=capOverlay(capw_, 0);
  rv=0; //warning killer
}


int vs_grab_w32MCIimp::getFrame(void* buf){
  Win32MCIdata& d=info_.data(capw_);
  d.sizeofBits=0;
  capGrabFrameNoStop(capw_);
  if (!d.sizeofBits){
    throw "vs_grab_w32MCIimp::getFrame"
          " - error while grabbing image frame";
  }
  ::memcpy(buf, d.ptrBits, d.sizeofBits);
  return d.sizeofBits;
}


bool vs_grab_w32MCIimp::boxDisplay(){
  if (caps_.fHasDlgVideoDisplay)
    if (capDlgVideoDisplay(capw_))
       return true;
  return false;
}
bool vs_grab_w32MCIimp::boxSource(){
  if (caps_.fHasDlgVideoSource)
    if (capDlgVideoSource(capw_))
       return true;
  return false;
}
bool vs_grab_w32MCIimp::boxFormat(){
  if (!caps_.fHasDlgVideoFormat ||
      !capDlgVideoFormat(capw_))
         return false;
  getDriverFormat();
  initWnd();
  return true;
}


void vs_grab_w32MCIimp::getDriverFormat(){
  delete pbi_;
  sz_bi_=capGetVideoFormatSize(capw_);
  pbi_=static_cast<BITMAPINFO*>(operator new(sz_bi_));
  capGetVideoFormat(capw_, pbi_, sz_bi_);
}





/******************************************************
**vs_grab_w32MCI
*******************************************************/

// Construction
vs_grab_w32MCI::vs_grab_w32MCI(
  const CtorArg& arg)
:
  pimp(new vs_grab_w32MCIimp(0, 
    arg.fmt.width(),arg.fmt.height(),arg.fmt.szPixel()*8, 
    arg.board, arg.channel))
{
}
vs_grab_w32MCI::~vs_grab_w32MCI(){
  delete pimp;
}



// vs_base Attributes
char const* vs_grab_w32MCI::name(){
  return s_name();
}

// vs_base operations
const img_fmt& vs_grab_w32MCI::fmt(){
  return pimp->fmt();
}
void vs_grab_w32MCI::getFrameSimple(img_data& img){
  img.reformat(fmt());
  int sz=pimp->getFrame(img.bits());
  assert(sz==img.fmt().szBits());
  sz=0; //warning killer

  // switch from bgr to rgb
  if (img.fmt().szPixel()==3)
    img_util::flipEndian3(img);
  if (img.fmt().szPixel()==4)
    img_util::flipEndian4(img);
}

// vs_grab attributes
char const* vs_grab_w32MCI::s_name(){
  return "w32MCI";
}

//MCI specific operations
void vs_grab_w32MCI::liveWindow(bool on){
  if (on){
    pimp->liveOn();
  } else{
    pimp->liveOff();
  }
}


void vs_grab_w32MCI::boxDisplay(){
  pimp->boxDisplay();
}
void vs_grab_w32MCI::boxSource(){
  pimp->boxSource();
}
void vs_grab_w32MCI::boxFormat(){
  pimp->boxFormat();
}


