#include "vs_grab_w32PXC.hpp"

#include "../util/util_w32.hpp"
#include "../img/img_data.hpp"
#include "../img/img_util.hpp"
#include "../util/util_time.hpp"
#include "../util/util_registry.hpp"

//#include <pxc/pxc.h>
//#include <pxc/iframe.h>

// Exclude rarely-used stuff from Windows headers
#define VC_EXTRALEAN		
#ifndef STRICT
#define STRICT
#endif
#include <windows.h> 

#if defined(_MSC_VER)
  #undef min
  #undef max
  #if _MSC_VER<1300
    #define min _MIN
    #define max _MAX
  #endif
#endif


#include <algorithm>

#include <assert.h>
#include <time.h>


/////////////////////////////////////
// static registry creator object

namespace{
  util::RegistryCreator<vs_grab,vs_grab_w32PXC> creator;
}


//////////////////////////////////////
// class encapsulating PXC grabber library 

class LibPXC{
private:
  LibPXC();
public:
  ~LibPXC();

  static LibPXC& instance();
  const PXC& lib();
private:
  char* libName_;
  PXC libpxc_;
};

LibPXC::LibPXC(){
  if (util_w32::winNT4()){
    libName_="pxc2_nt.dll";
  } else{
    libName_="pxc2_95.dll";
  }

  if (!imagenation_OpenLibrary(libName_,&libpxc_,sizeof(libpxc_)))
    throw "vs_grab_w32PXC::LibPXC::LibPXC"
          " - couldn't open PXC grabber library.";
}

LibPXC::~LibPXC(){
  imagenation_CloseLibrary(&libpxc_);
}

const PXC& LibPXC::lib(){
  return libpxc_;
}

LibPXC& LibPXC::instance(){
  //the object holding PXC frame library
  static LibPXC theLib;
  return theLib;
}




//////////////////////////////////////
// class encapsulating PXC frame library 

class LibFrame{
private:
  LibFrame();
public:
  ~LibFrame();

  static LibFrame& instance();
  const FRAMELIB& lib();
private:
  char* libName_;
  FRAMELIB frameLib_;
};

LibFrame::LibFrame(){
  libName_="frame_32.dll";

  if (!imagenation_OpenLibrary(libName_,&frameLib_,sizeof(frameLib_)))
    throw "vs_grab_w32PXC::LibFrame::LibFrame"
          " - couldn't open PXC frame library.";
}

LibFrame::~LibFrame(){
  imagenation_CloseLibrary(&frameLib_);
}

const FRAMELIB& LibFrame::lib(){
  return frameLib_;
}

LibFrame& LibFrame::instance(){
  //the object holding PXC frame library
  static LibFrame theLib;
  return theLib;
}


//////////////////////////////////////
// class encapsulating frame grabber
class FrameGrabber{
  const FrameGrabber& operator=(const FrameGrabber&);
public:
  FrameGrabber(
    int board, int channel, const img_fmt& f, 
    bool singleField);
  ~FrameGrabber();

//audience FrameBuffer
public:
  int widthPhysical(){return driverWidth_;}
  int heightPhysical(){return driverHeight_;}
  int pixelSizePhysical(){return szpix_;}
  int pixelType(){return pixeltype_;}
  bool singleField(){return singleField_;}

  long grab(FRAME* pf, bool queued);
  void waitGrab(long id);

//audience vs_grab:
public:
  int widthLogical(){return driverWidth_;}
  int heightLogical(){
    return singleField_?driverHeight_/2:driverHeight_;
  }
  int pixelSizeLogical(){return szpix_;}

private:
  void setFormatInterlaced(const img_fmt& f);
  void setFormatNonInterlaced(const img_fmt& f);
  void setCropLeftTop();
private:
  const PXC& libpxc_;

private:
  // frame grabber handle
  long fgh_;

  //grab mode
  bool singleField_;

  //grabber data
  int modelnumber_;
  int boardrev_;

  //type of attached video signal
  int videotype_;

  //image attributes
  int maxWidth_;
  int maxHeight_;
  int pixeltype_;

  //image attributes
  int driverWidth_;
  int driverHeight_;
  int szpix_;
};

FrameGrabber::FrameGrabber(
  int board, int channel, 
  const img_fmt& f, 
  bool singleField)
:
  libpxc_(LibPXC::instance().lib())
{
  //request a grabber...
  fgh_=libpxc_.AllocateFG(static_cast<short>(board));
  if (!fgh_)
    throw "vs_grab_w32PXC::FrameGrabber::FrameGrabber"
          " - couldn't open the requested grabber.";

  //set video source
  libpxc_.SetCamera(fgh_, static_cast<short>(channel), 0);

  modelnumber_ = libpxc_.GetModelNumber(fgh_);
  boardrev_    = libpxc_.ReadRevision(fgh_);
  int proba=libpxc_.GetFieldCount(fgh_);

  // find out image dimensions
  // wait for CCIR autodetect
  Sleep(100); 
  videotype_= libpxc_.VideoType(fgh_);
  switch(videotype_) {
  case 0:     /* no video */
    libpxc_.FreeFG(fgh_);
    throw "vs_grab_w32PXCimp::vs_grab_w32PXCimp"
          " - no video connected.";
    break;
  case 2:     /* CCIR */
    maxWidth_ = 768;
    maxHeight_ = 576;
    break;
  case 1:     /* NTSC */
    maxWidth_ = 640;
    maxHeight_ = 486;
    break;
  default:
    throw "vs_grab_w32PXCimp::vs_grab_w32PXCimp"
          " - unexpected videotype.";
  }

  proba=libpxc_.GetFieldCount(fgh_);

  //  Determine width_ and height_ based on f, maxWidth_ 
  //and maxHeight_. depending on whether we want to 
  //consider full (interlaced) sensor data or only the
  //first half of it (a single field).
  //  The latter is useful when analysing dynamic scenes  
  //with cameras which does not have a full shutter feature.
  if (singleField && f.height()<=maxHeight_/2){
    singleField_ =true;
    setFormatNonInterlaced(f);
  } else{
    singleField_ =false;
    setFormatInterlaced(f);
  }

  //find out pixel type
  szpix_=f.szPixel();
  switch(szpix_) {
  case 1:
    pixeltype_=PBITS_Y8; break;
  case 2:
    pixeltype_=PBITS_Y16; break;
  case 3:
    pixeltype_=PBITS_RGB24; break;
  case 4:
    pixeltype_=PBITS_RGB32; break;
  default:
    throw "vs_grab_w32PXCimp::FrameGrabber::FrameGrabber"
          " - unexpected pixeltype.";
  }

  //tell grabber whether we have an S-Video input
  if (channel==1){
    short cf=libpxc_.GetChromaControl(fgh_);
    if (cf==-1)
      throw "vs_grab_w32PXC::FrameGrabber::FrameGrabber"
            " - got error on GetChromaControl.";
    short rv=libpxc_.SetChromaControl(fgh_, cf|SVIDEO);
    if (rv==0)
      throw "vs_grab_w32PXC::FrameGrabber::FrameGrabber"
            " - got error on SetChromaControl.";
  }
}

void FrameGrabber::setCropLeftTop(){
  //set cropping from top and left
  const int cropValueX=2;
  int rv=libpxc_.SetLeft(fgh_,cropValueX);  //should return cropValue
  assert(rv==cropValueX);

  const int cropValueY=0;
  rv =libpxc_.SetTop(fgh_,cropValueY);     //should return cropValue
  assert(rv==cropValueY);
}

void FrameGrabber::setFormatNonInterlaced(const img_fmt& f){
  assert (singleField_ && f.height()<=maxHeight_/2);

  setCropLeftTop();

  // negotiate image dimensions
  driverWidth_ = std::min(maxWidth_,f.width());
  int rv=libpxc_.SetXResolution(
    fgh_,static_cast<short>(driverWidth_)); //should return driverWidth_
  if (rv!=driverWidth_){
    libpxc_.SetXResolution(fgh_,static_cast<short>(maxWidth_)); 
    driverWidth_=libpxc_.GetXResolution(fgh_); 
    assert(driverWidth_!=0);
  }
  // no cropping
  rv=libpxc_.SetWidth(fgh_,
    static_cast<short>(driverWidth_)); //should return driverWidth_
  assert(rv==driverWidth_);

  driverHeight_ = std::min(maxHeight_, 2*f.height());
  rv=libpxc_.SetYResolution(fgh_,
    static_cast<short>(driverHeight_)); //should return driverHeight_
  if (rv!=driverHeight_){
    libpxc_.SetYResolution(fgh_,
      static_cast<short>(maxHeight_)); 
    driverHeight_=libpxc_.GetXResolution(fgh_); 
    assert(driverHeight_!=0);
  }
  // no cropping!
  rv=libpxc_.SetHeight(fgh_,
    static_cast<short>(driverHeight_));  //should return driverHeight_
  assert(rv==driverHeight_);
}

void FrameGrabber::setFormatInterlaced(const img_fmt& f){
  setCropLeftTop();

  // negotiate image dimensions
  driverWidth_ = std::min(maxWidth_,f.width());
  int rv=libpxc_.SetXResolution(fgh_,
    static_cast<short>(driverWidth_)); //should return driverWidth_
  if (rv!=driverWidth_){
    libpxc_.SetXResolution(fgh_,
      static_cast<short>(maxWidth_));
    driverWidth_=libpxc_.GetXResolution(fgh_); 
    assert(driverWidth_!=0);
  }
  // no cropping
  rv=libpxc_.SetWidth(fgh_,
    static_cast<short>(driverWidth_)); //should return driverWidth_
  assert(rv==driverWidth_);

  driverHeight_ = std::min(maxHeight_, f.height());
  rv=libpxc_.SetYResolution(fgh_,
    static_cast<short>(driverHeight_)); //should return driverHeight_
  if (rv!=driverHeight_){
    libpxc_.SetYResolution(fgh_,
      static_cast<short>(maxHeight_)); 
    driverHeight_=libpxc_.GetXResolution(fgh_); 
    assert(driverHeight_!=0);
  }
  // no cropping
  rv=libpxc_.SetHeight(fgh_,
    static_cast<short>(driverHeight_));  //should return driverHeight_
  assert(rv==driverHeight_);
}

FrameGrabber::~FrameGrabber(){
  libpxc_.FreeFG(fgh_);
}

long FrameGrabber::grab(
  FRAME* pf, bool queued)
{
  int grabtype=(queued?QUEUED:0);
  if (singleField_){
    grabtype+=FIELD0+SINGLE_FLD;
  } else{
    //should be set to EITHER if a full 
    //shutter camera is connected
    grabtype+=FIELD0;
  }
  return libpxc_.Grab(fgh_, pf, 
    static_cast<short>(grabtype));
}
void FrameGrabber::waitGrab(long id)
{
  libpxc_.WaitFinished(fgh_, id);
}




//////////////////////////////////////
// class encapsulating single frame buffer

class FrameBuf{
  const FrameBuf& operator=(const FrameBuf&);
public:
  FrameBuf(FrameGrabber& fg);
  ~FrameBuf();

  bool framePending();
  void* getPendingFrame();
  void* getNewFrame();
  void prepareFrame();
  int transferFrame(void* p);

private:
  const PXC&      libpxc_;
  const FRAMELIB& libfrm_;

private:
  FrameGrabber& fg_;

  FRAME* pframe_;
  void* pbuf_;

  long id_;
  time_t tPrepared_;

  int szScanLine_;
  int szBuf_;
  bool singleField_;
};



FrameBuf::FrameBuf(FrameGrabber& fg)
:
  libpxc_(LibPXC::instance().lib()),
  libfrm_(LibFrame::instance().lib()),

  fg_(fg),

  pframe_(libpxc_.AllocateBuffer(
    static_cast<short>(fg_.widthPhysical()), 
    static_cast<short>(fg_.heightPhysical()), 
    static_cast<short>(fg_.pixelType()))),

  id_(0),
  tPrepared_(0),

  szScanLine_(fg_.widthPhysical()*fg_.pixelSizePhysical()),
  szBuf_(fg_.heightPhysical()*szScanLine_),
  singleField_(fg_.singleField())
{
  if (!pframe_)
    throw "vs_grab_w32PXCimp::FrameBuf::FrameBuf"
          " - couldn't allocate frame.";
  pbuf_=libfrm_.FrameBuffer(pframe_);
}

FrameBuf::~FrameBuf(){
  assert(pframe_);
  libpxc_.FreeFrame(pframe_);
}


bool FrameBuf::framePending(){
  return tPrepared_ && (util_time::msDayUtc()-tPrepared_)<=80;
}

void* FrameBuf::getPendingFrame(){
  assert (tPrepared_);
  fg_.waitGrab(id_);
  tPrepared_=0;
  return pbuf_;
}
void* FrameBuf::getNewFrame(){
  fg_.grab(pframe_, false);
  return pbuf_;
}

void FrameBuf::prepareFrame(){
  tPrepared_=util_time::msDayUtc();
  id_=fg_.grab(pframe_, true);
}

int FrameBuf::transferFrame(void* p){
  // copy lines upside-down
  int szTransfer=singleField_?szBuf_/2:szBuf_;
  char* pdst=static_cast<char*>(p);
  char* psrc=static_cast<char*>(pbuf_)+szTransfer-szScanLine_;

  if (singleField_){
    for (int i=fg_.heightLogical()-1; i>=0; --i) {
      ::memcpy(pdst, psrc, szScanLine_);
      psrc-=szScanLine_;
      pdst+=szScanLine_;
    }
  } else{
    for (int i=fg_.heightPhysical()-1; i>=0; --i) {
      ::memcpy(pdst, psrc, szScanLine_);
      psrc-=szScanLine_;
      pdst+=szScanLine_;
    }
  }
  assert(pdst-szTransfer==p);
  assert(psrc+szScanLine_==pbuf_);

//  ::memcpy(p, pbuf_, szTransfer);

  return szTransfer;
}


/******************************************************
 ** private implementation class
 ******************************************************/

class vs_grab_w32PXCimp{
public:
  vs_grab_w32PXCimp(
    int board, int channel, const img_fmt& f, 
    bool singleField);
  ~vs_grab_w32PXCimp();

public:
  const img_fmt& fmt();
  int getFrame(void*);
private:
  FrameBuf& fb(int i);

private:
  FrameGrabber grabber_;

  int fbIndex_;
  FrameBuf fb0_;
  FrameBuf fb1_;

  img_fmt fmt_;
};


vs_grab_w32PXCimp::vs_grab_w32PXCimp(
  int board, int channel, 
  const img_fmt& f, 
  bool singleField)
:
  grabber_(board, channel, f, singleField),
  fbIndex_(0),
  fb0_(grabber_),
  fb1_(grabber_),
  fmt_(grabber_.widthLogical(), 
       grabber_.heightLogical(), 
       grabber_.pixelSizeLogical())
{
}

vs_grab_w32PXCimp::~vs_grab_w32PXCimp(){
}

const img_fmt& vs_grab_w32PXCimp::fmt(){
  return fmt_;
}

int vs_grab_w32PXCimp::getFrame(void* p){

  int fbIndexOther=(fbIndex_+1)%2;

  if (fb(fbIndex_).framePending()){
    fb(fbIndexOther).prepareFrame();
    fb(fbIndex_).getPendingFrame();
  } else{
    fb(fbIndex_).getNewFrame();
    fb(fbIndexOther).prepareFrame();
  }

  int sz=fb(fbIndex_).transferFrame(p);

  fbIndex_=fbIndexOther;

  return sz;
}

FrameBuf& vs_grab_w32PXCimp::fb(int i){
  if (i==0)
    return fb0_;
  else
    return fb1_;
}


/******************************************************
 ** vs_grab_w32PXC
 ******************************************************/

// Construction
vs_grab_w32PXC::vs_grab_w32PXC(
  const CtorArg& arg)
:
  pimp(new vs_grab_w32PXCimp(
    arg.board, arg.channel, arg.fmt, 
    arg.mode==0))
{
  //enum FieldOption{ConsiderSingleField, AllwaysBothFields};
}

vs_grab_w32PXC::~vs_grab_w32PXC(){
  delete pimp;
}

// vs_base Attributes
char const* vs_grab_w32PXC::name(){
  return s_name();
}

// vs_base operations
const img_fmt& vs_grab_w32PXC::fmt(){
  return pimp->fmt();
}
void vs_grab_w32PXC::getFrameSimple(img_data& img){
  img.reformat(fmt());
  int sz=pimp->getFrame(img.bits());
  assert(sz==img.fmt().szBits());
  sz=0; //warning killer

  //switch from bgr to rgb
  if (img.fmt().szPixel()==3 ||
      img.fmt().szPixel()==4)
    img_util::flipEndian(img);
}

// vs_grab attributes
char const* vs_grab_w32PXC::s_name(){
  return "w32PXC";
}





