#include "win_ann.hpp"

#include "../../grid/draw/grid_draw.hpp"

#include <stdexcept>

#include <assert.h>

/////////////////////////////////////////////
// construction

win_ann::win_ann(
  const win_ann_abstract& o)
{
  (*this)=o;
}

const win_ann& win_ann::operator=(
  const win_ann_abstract& o)
{
  clear();

  sounds_.resize(o.nSounds());
  for (int i=0; i<o.nSounds(); i++){
    sounds_[i]=o.sound(i);
  }
  points_.resize(o.nPoints());
  for (int i=0; i<o.nPoints(); i++){
    points_[i]=o.point(i);
  }
  lines_.resize(o.nLines());
  for (int i=0; i<o.nLines(); i++){
    lines_[i]=o.line(i);
  }
  circles_.resize(o.nCircles());
  for (int i=0; i<o.nCircles(); i++){
    circles_[i]=o.circle(i);
  }
  contours_.resize(o.nContours());
  for (int i=0; i<o.nContours(); i++){
    contours_[i]=o.contour(i);
  }
  strings_.resize(o.nStrings());
  for (int i=0; i<o.nStrings(); i++){
    strings_[i]=o.string(i);
  }
  return *this;
}

namespace{
  inline grid::Pixel scaleConst(
    const grid::Pixel& pt, int s)
  {
    return grid::Pixel(pt.x()/s, pt.y()/s);
  }
  inline void scaleMutate(
    grid::Pixel& pt, int s)
  {
    pt=scaleConst(pt,s);
  }
}

void win_ann::assign(
  const win_ann_abstract& o, int s)
{
  clear();

  sounds_.resize(o.nSounds());
  for (int i=0; i<o.nSounds(); i++){
    sounds_[i]=o.sound(i);
  }
  points_.resize(o.nPoints());
  for (int i=0; i<o.nPoints(); i++){
    points_[i]=o.point(i);
    scaleMutate(points_[i].p, s);
  }
  lines_.resize(o.nLines());
  for (int i=0; i<o.nLines(); i++){
    lines_[i]=o.line(i);
    lines_[i].l=grid::Line(
      scaleConst(lines_[i].l.first(), s),
      scaleConst(lines_[i].l.second(), s));
  }
  circles_.resize(o.nCircles());
  for (int i=0; i<o.nCircles(); i++){
    circles_[i]=o.circle(i);
    scaleMutate(circles_[i].c,s);
    circles_[i].r/=s;
  }
  contours_.resize(o.nContours());
  for (int i=0; i<o.nContours(); i++){
    contours_[i]=o.contour(i);
    for (unsigned j=0; j<contours_[i].c.size(); ++j){
      scaleMutate(contours_[i].c[j],s);
    }
  }
  strings_.resize(o.nStrings());
  for (int i=0; i<o.nStrings(); i++){
    strings_[i]=o.string(i);
    scaleMutate(strings_[i].p,s);
  }
}


/////////////////////////////////////////////
// mutator operations

void win_ann::clear(){
  sounds_.clear();
  points_.clear();
  lines_.clear();
  circles_.clear();
  contours_.clear();
  strings_.clear();
}

void win_ann::addSound(
  int t_ms, int freq)
{
  sounds_.push_back(Sound());
  sounds_.back().t=t_ms;
  sounds_.back().f=freq;
}
void win_ann::addPixel(
  const grid::Pixel& point, 
  int col, char const* str,
  PixelStyle style) 
{
  points_.push_back(Point2D());
  points_.back().p= point;
  points_.back().colour=col;
  points_.back().style=style;
  points_.back().text=str;
}

void win_ann::addLine(
  const grid::Line& line, 
  int col, char const* str,
  LineStyle style, int width)
{
  lines_.push_back(Line());
  lines_.back().l=line;
  lines_.back().width=width;
  lines_.back().style=style;
  lines_.back().colour=col;
  lines_.back().text=str;
}

void win_ann::addCircle(
  const grid::Pixel& centre, 
  int radius,
  int col, char const* str,
  LineStyle ls, FillStyle fs, 
  int width)
{
  circles_.push_back(Circle());
  circles_.back().c=centre;
  circles_.back().r=radius;
  circles_.back().colour=col;
  circles_.back().text=str;
  circles_.back().ls=ls;
  circles_.back().fs=fs;
  circles_.back().width=width;
}

void win_ann::addContour(
  const grid::Contour& contour, 
  int col, char const* str,
  PixelStyle s)
{
  if (!contour.empty()) {
    contours_.push_back(Contour());
    contours_.back().c= contour;
    contours_.back().colour=col;
    contours_.back().text=str;
    contours_.back().style=s;
  }
}

void win_ann::addStr(
  const grid::Pixel& pt, 
  int col, char const* str,
  int sz)
{
  strings_.push_back(String());
  strings_.back().p= pt;
  strings_.back().colour=col;
  strings_.back().text=str;
  strings_.back().size=sz;
}








/////////////////////////////////////////////
// const operations

bool win_ann::empty() const {
  return 
   points_.empty() &&
   lines_.empty() && 
   circles_.empty() && 
   contours_.empty() &&
   sounds_.empty() &&
   strings_.empty();
}

int win_ann::nSounds() const{
  return static_cast<int>(sounds_.size());
}
const win_ann_abstract::Sound& 
win_ann::sound(int i) const{
  if (i<0 || i>=nSounds())
    throw std::runtime_error("win_ann::sound - illegal argument");
  return sounds_[i];
}

int win_ann::nPoints() const{
  return static_cast<int>(points_.size());
}
const win_ann_abstract::Point2D& 
win_ann::point(int i) const{
  if (i<0 || i>=nPoints())
    throw std::runtime_error("win_ann::point - illegal argument");
  return points_[i];
} 

int win_ann::nLines() const{
  return static_cast<int>(lines_.size());
}
const win_ann_abstract::Line& 
win_ann::line(int i) const{
  if (i<0 || i>=nLines())
    throw std::runtime_error("win_ann::line - illegal argument");
  return lines_[i];
} 

int win_ann::nCircles() const{
  return static_cast<int>(circles_.size());
}
const win_ann_abstract::Circle& 
win_ann::circle(int i) const{
  if (i<0 || i>=nCircles())
    throw std::runtime_error("win_ann::circle - illegal argument");
  return circles_[i];
} 

int win_ann::nContours() const{
  return static_cast<int>(contours_.size());
}
const win_ann_abstract::Contour& 
win_ann::contour(int i) const{
  if (i<0 || i>=nContours())
    throw std::runtime_error("win_ann::contour - illegal argument");
  return contours_[i];
} 

int win_ann::nStrings() const{
  return static_cast<int>(strings_.size());
}
const win_ann_abstract::String& 
win_ann::string(int i) const{
  if (i<0 || i>=nStrings())
    throw std::runtime_error("win_ann::string - illegal argument");
  return strings_[i];
} 

