#include "win_ann_util.hpp"

#include "win_ann_abstract.hpp"

#include "../../grid/root/grid_primitives.hpp"

#include <assert.h>


using grid::Pixel;
using grid::Line;



namespace{
  inline int myround(double f){
    return static_cast<int>(f<0?f-.5:f+.5);
  }
}

void win_ann_util::addAim(    
  win_ann_abstract& ann,
  const Pixel& p, int size, 
  int col, int width)
{
  const int size2=3*size/2;
  const int d1=4*size/5;  //80%  of size
  const int d2=4*size2/3; //133% of size2

  ann.addCircle(p, size,  col,"",
    win_ann_abstract::LS_Normal, 
    win_ann_abstract::FS_None, width);
  ann.addCircle(p, size2, col,"",
    win_ann_abstract::LS_Normal, 
    win_ann_abstract::FS_None, width);
  ann.addLine( Line( 
      Pixel(p.x(), p.y()+d1), 
      Pixel(p.x(), p.y()+d2) ), col,"",
    win_ann_abstract::LS_Normal, width);
  ann.addLine( Line(
      Pixel(p.x(), p.y()-d1), 
      Pixel(p.x(), p.y()-d2) ), col,"",
    win_ann_abstract::LS_Normal, width);
  ann.addLine( Line(
      Pixel(p.x()+d1, p.y()), 
      Pixel(p.x()+d2, p.y()) ), col,"",
    win_ann_abstract::LS_Normal, width);
  ann.addLine( Line(
      Pixel(p.x()-d1, p.y()), 
      Pixel(p.x()-d2, p.y()) ), col,"",
    win_ann_abstract::LS_Normal, width);
}


void win_ann_util::addCamera(    
  win_ann_abstract& ann,
  const Pixel& p, 
  double phi,
  int size, int col,
  char const* comment)
{
  const int width=3;

  ann.addCircle(p, size,  col,"",
    win_ann_abstract::LS_Normal, 
    win_ann_abstract::FS_None, width);

  Pixel lnOffset(
    myround(2*size*sin(phi)),
    myround(2*size*cos(phi)));
  ann.addLine(
    Line(p, p+lnOffset), col,"",
    win_ann_abstract::LS_Normal, width);

  int offy=cos(phi)>0? 2*size: -3*size;
  Pixel strOffset(-2*size,offy);
  ann.addStr(p+strOffset, col, comment, 1);
}


void win_ann_util::addRectangle(
  win_ann_abstract& ann,
  const Pixel& p00,
  const Pixel& p11,
  int col, int width)
{
  Pixel p01(p11.x(), p00.y());
  Pixel p10(p00.x(), p11.y());

  ann.addLine(Line(p00,p01), col, "",
    win_ann_abstract::LS_Normal, width);
  ann.addLine(Line(p01,p11), col, "",
    win_ann_abstract::LS_Normal, width);
  ann.addLine(Line(p11,p10), col, "",
    win_ann_abstract::LS_Normal, width);
  ann.addLine(Line(p10,p00), col, "",
    win_ann_abstract::LS_Normal, width);
}




void win_ann_util::addEgoAngles(win_ann_abstract& ann,
  double ax, double ay, double az,
  double x0, double y0, double f)
{
  // std::cerr <<"win_ann_util::addEgoAngles " <<x0 <<" " <<y0 <<"\n";
  int dax= myround(y0+f*tan(-ax)); // positive angle=positive offset
  int day= myround(x0+f*tan(ay));  // positive angle=negative offset
  int dazx=myround(0.25*f*cos(az));
  int dazy=myround(0.25*f*sin(az));
  
  Pixel c(myround(x0), myround(y0));
  int width=myround(2*x0);
  int height=myround(2*y0);
  
  ann.addLine(Line(Pixel(0,dax),Pixel(width,dax)), 
    0, "", win_ann_abstract::LS_Normal, 1);
  ann.addLine(Line(Pixel(day,0),Pixel(day,height)), 
    1, "", win_ann_abstract::LS_Normal, 1);
  ann.addLine(Line(c, c+Pixel(dazx,dazy)), 
    2, "", win_ann_abstract::LS_Normal, 1);
}

  
void win_ann_util::addEpipole(win_ann_abstract& ann,
  double e0, double e1, double e2,
  double x0, double y0, double f)
{
  double dx, dy;
  if (e2==0){
    double n=sqrt(e0*e0+e1*e1+e2*e2);
    dx=2*x0*e0/n;
    dy=2*x0*e1/n;
  } else{
    dx=e0/e2; 
    dy=e1/e2;
  }
  Pixel c(myround(x0), myround(y0));
  Pixel ptEpi(myround(x0+f*dx), myround(y0+f*dy));
  
  ann.addLine(grid::Line(c, ptEpi), 5, "", 
      win_ann_abstract::LS_Normal, 1);
  ann.addPixel(ptEpi, 5, "", win_ann_abstract::PS_BoldHuge);
}


