#ifndef win_base_hpp
#define win_base_hpp

/******************************************************
 *
 ** win_base by Sinisa Segvic
 *
 ** encapsulates an os dependent 
 ** implementation of a simple window
 *
 ** the concrete classes make a local copy of 
 ** all const& arguments, so that the caller may 
 ** delete them after functions return
 *
 ** window commands:
 **   <Ctrl>  + <RightClick>  ...  dumps image in .bmp format
 **             <RightClick>  ...  dumps annotation in .eps format
 **   <Shift> + <RightClick>  ...  dumps image+ann in .eps format
 *
 ** in order to support factory method,
 ** each concrete class should declare:
 **  - public constructor with a single CtorArg argument
 **  - static data member of the type 
 **    Registry<win_base>::Creator<win_base_concrete>
 **    which interfaces to the registry object
 **  - static member s_name() which should be 
 **    used to implement name()
 ******************************************************/

class img_data;
class win_ann_abstract;
class win_event_abstract;


class win_base{
  friend class win_wrap;
public:
  win_base();
  virtual ~win_base();
private:
  win_base(const win_base&);
  const win_base& operator=(const win_base&);

public:    
  virtual char const* name()=0;
  virtual int width() =0;
  virtual int height() =0;

  virtual void position(int, int)=0;
  virtual void size(int, int)=0;
  virtual void caption(char const*)=0;

//window output
public:
  // sets img, clears ann, redraws
  void setImg(const img_data&);  
  // sets img, sets ann, redraws
  void setImgAnn(const img_data&, const win_ann_abstract&);
  // leaves previous img, sets ann, redraws
  virtual void setAnn(const win_ann_abstract&);
  // beeps the sound
  virtual void flash()=0;

//window input
public:
  //gets the next event, returns false when unavailable
  virtual bool event(int& iterator, win_event_abstract&) const =0;

// output template methods (implemented by derived classes)
private:
  virtual void putImg(const img_data&)=0;
  virtual void putImgAnn(const img_data&, const win_ann_abstract&)=0;
  virtual void putAnn(const win_ann_abstract&)=0;

// win_base construction
public:
  // concrete classes ctor argument
  struct CtorArg{
    int width;
    int height; 
    char const* caption;
  public:
    CtorArg(int w, int h, char const* s):
      width(w), height(h), caption(s)
    {}
  };
  //factory for concrete classes 
  static win_base* create(
    const char* name, 
    const CtorArg& arg);
};

#endif
