#include "win_w32_worker.hpp"

#include "../root/win_base.hpp"
#include "../ann/win_ann.hpp"
#include "../ann/win_ann_dump.hpp"
#include "../event/win_event.hpp"
#include "../event/win_event_queue.hpp"

#include "../../img/img_data.hpp"
#include "../../img/img_util.hpp"
#include "../../img/io/img_io_save_bmp.hpp"
#include "../../util/util_w32.hpp"
#include "../../grid/draw/grid_draw_w32.hpp"
#include "../../grid/draw/grid_draw_fileEps.hpp"

#include "../../util/util_fs.hpp"

#include "../../thr/thr_base.hpp"

#include <windows.h>

#include <vector>
#include <string>
#include <iostream>
#include <stdexcept>

#include <assert.h>
#include <ctype.h>


////////////////////////////////////////
// static access to active workers

class WorkerManager{
  std::vector<win_w32_worker*> workers_;
public:
  win_w32_worker& get(HWND hwnd){
    int id=GetWindowLong(hwnd,GWL_USERDATA);
    if (id>=static_cast<int>(workers_.size())){
      assert(0);
    }
    win_w32_worker* p=workers_[id];
    assert(p);
    return *p;
  }
  void registerWorker(win_w32_worker* p, HWND hwnd){
    workers_.push_back(p);
    int id=static_cast<int>(workers_.size())-1;
    SetWindowLong(hwnd,GWL_USERDATA,id);
  }
  void invalidate(int i){
    workers_[i]=0;
  }
};
namespace{
  WorkerManager s_manager;
}

/////////////////////////////////////////////////
// Window Procedure called by Windows
// and the corresponding static helpers

namespace{
  void paintAnnotation(HWND hwnd, HDC hdc){
    win_w32_worker& self=s_manager.get(hwnd);;
    thr::Lock guard(self.mutex());

    if (!self.ann().empty()){
      grid::draw_w32 dc(1, 0, -1, self.height()-1, hdc);
      win_ann_dump::toDC(self.ann(),dc);
    }
  }
  void paintImg(HWND hwnd, HDC hdc){
    win_w32_worker& self=s_manager.get(hwnd);;
    thr::Lock guard(self.mutex());
    bool ok=util_w32::drawImg(hdc, self.img(), 0,0);
    if (!ok){
      SendMessage(hwnd, WM_ERASEBKGND, (WPARAM)hdc, 0);
    }
  }
  void doPaint(HWND hwnd){
    PAINTSTRUCT ps;     
    HDC hdc = BeginPaint(hwnd, &ps);
    paintImg(hwnd, hdc);
    paintAnnotation(hwnd, hdc);
    EndPaint(hwnd, &ps);
  }

  int getMod(){
    int mod=0;
    if (GetKeyState(VK_SHIFT)<0)
      mod|=win_event::Shift;
    if (GetKeyState(VK_CONTROL)<0)
      mod|=win_event::Ctrl;
    if (GetKeyState(VK_MENU)<0)
      mod|=win_event::Alt;
    return mod;
  }

  void processKey(HWND hwnd, int vkey){
    win_w32_worker& self=s_manager.get(hwnd);;
    thr::Lock guard(self.mutex());

    int type=win_event::KeyboardAscii;
    int ch=MapVirtualKey(vkey, 2);
    if (ch==0){
      if (vkey>=VK_F1 && ch<=VK_F12){
        ch=vkey-VK_F1;
        type=win_event::KeyboardFx;
      } else if (vkey>=VK_LEFT && ch<=VK_DOWN){
        ch=vkey-VK_LEFT;
        type=win_event::KeyboardArrow;
      } else if (vkey>=VK_INSERT && ch<=VK_DELETE){
        ch=vkey-VK_PRIOR;
        type=win_event::KeyboardEdit;
      } else if (vkey>=VK_END && ch<=VK_HOME){
        ch=2+1-(vkey-VK_END);
        type=win_event::KeyboardEdit;
      } else if (vkey>=VK_PRIOR && ch<=VK_NEXT){
        ch=4+vkey-VK_PRIOR;
        type=win_event::KeyboardEdit;
      }
    }
  
    self.addChar(type, ch, getMod());
  }
  void processClick(HWND hwnd, int button, int x, int y){
    win_w32_worker& self=s_manager.get(hwnd);;
    thr::Lock guard(self.mutex());
    self.addClick(button, x,y, getMod());
  }
  void processDoubleClick(HWND hwnd, int button, int x, int y){
    win_w32_worker& self=s_manager.get(hwnd);;
    thr::Lock guard(self.mutex());
    self.addClick(button, x,y, getMod()|win_event::DoubleClick);
  }
  void processRightClick(HWND hwnd, int , int ){
    win_w32_worker& self=s_manager.get(hwnd);;
    thr::Lock guard(self.mutex());

    //the file is named after the window caption...
    std::string fileName=self.text();
    for (unsigned i=0; i<fileName.size(); i++)
      if (!isalpha(fileName[i]) && !isdigit(fileName[i]))
        fileName.erase(i);
    std::string fullPath=util_fs::getLogDir();
    fullPath+=fileName;

    bool wantBmp=GetKeyState(VK_CONTROL)<0;
    if (wantBmp) {
      //bmp dump, no annotations
      fullPath+=".bmp";
      std::string unique=util_fs::
        uniqueFilename(fullPath.c_str());
      img_io_save_bmp w(unique.c_str());
      w(self.img());
    } else{ //wantEps
      fullPath+=".eps";
      std::string unique=util_fs::
        uniqueFilename(fullPath.c_str());

      //dump the annotation to the drawing object
      //dump the image as well if shift was pressed
      if (GetKeyState(VK_SHIFT)<0){
        grid::draw_fileEps dc(unique.c_str(), 
          self.width(), self.height(), self.img());
        win_ann_dump::toDC(self.ann(),dc);
      } else{
        grid::draw_fileEps dc(unique.c_str(), 
          self.width(), self.height());
        win_ann_dump::toDC(self.ann(),dc);
      }
    }

  }

  bool processMsg(
    HWND hWnd, unsigned int message, 
    WPARAM wParam, LPARAM lParam)
  {
    switch (message) {        
    case WM_USER:
      switch(wParam){
      case 0:
        DestroyWindow(hWnd);
        break;
      default:
        throw std::runtime_error("win_w32_worker::WindowProcedure - unknown WM_USER message.");
      }
      return true;
    case WM_USER+1:
      { //draw annotation
        HDC hdc=GetDC(hWnd);
        paintAnnotation(hWnd, hdc);
        ReleaseDC(hWnd, hdc);
      }
      break;
    case WM_PAINT:
      doPaint(hWnd);
      return true;
    case WM_SYSKEYDOWN:
    case WM_KEYDOWN:
      processKey(hWnd, (int)wParam);
      return true;
    case WM_LBUTTONDOWN:
      processClick(hWnd, 0, LOWORD(lParam), HIWORD(lParam));
      return true;
    case WM_MBUTTONDOWN:
      processClick(hWnd, 1, LOWORD(lParam), HIWORD(lParam));
      return true;
    case WM_RBUTTONDOWN:
      processRightClick(hWnd, LOWORD(lParam), HIWORD(lParam));
      return true;
    case WM_LBUTTONDBLCLK :
      processDoubleClick(
        hWnd, 0, LOWORD(lParam), HIWORD(lParam));
      return true;
    case WM_MBUTTONDBLCLK:
      processDoubleClick(
        hWnd, 1, LOWORD(lParam), HIWORD(lParam));
      return true;
    case WM_RBUTTONDBLCLK:
      processDoubleClick(
        hWnd, 2, LOWORD(lParam), HIWORD(lParam));
      return true;
    case WM_CLOSE:
      return true;
    case WM_DESTROY:
      ::PostQuitMessage (0);            
      return true;
    }
    return false;
  }

  LRESULT CALLBACK WindowProcedure(
    HWND hWnd, unsigned int message, 
    WPARAM wParam, LPARAM lParam)
  {
    //printf("Got msg %x\n", message);

    bool processed;
    try{
      processed=processMsg(hWnd, message, wParam, lParam);
    }catch(char const* msg){
       MessageBox(0, msg, "Window Procedure: got exception", IDOK);
       return 0;
    } catch (...){
       MessageBox(0, "Unknown exception!", "Window Procedure: got exception", IDOK);
       return 0;
    }
    if (processed)
      return 0;
    return ::DefWindowProc (hWnd, message, wParam, lParam );
  }
  
}



//////////////////////////////////////////////
// helper class providing the 
// registration of the Window Class

class WinClass{
  std::string name_;
public:
  WinClass (
    WNDPROC winProc, 
    char const * className);
  char const* name(){
    return name_.c_str();
  }
};

WinClass::WinClass(
  WNDPROC winProc, 
  char const * className)
{
  WNDCLASS myClass;
  name_=className;

  //check whether the name is free
  int i=0;
  while (GetClassInfo(GetModuleHandle(0), 
         name_.c_str(), &myClass))
  {
    name_[name_.size()-1]=static_cast<char>('0'+i);
    assert (i<10);
    i++;
  }
    
  myClass.style = CS_DBLCLKS;
  myClass.lpfnWndProc = winProc;            // window procedure: mandatory
  myClass.cbClsExtra = 0;
  myClass.cbWndExtra = 0;
  myClass.hInstance = GetModuleHandle(0);   // owner of the class: mandatory
  myClass.hIcon = 0;
  myClass.hCursor = ::LoadCursor (0, IDC_ARROW);       // optional
  myClass.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1); // optional
  myClass.lpszMenuName = 0;
  myClass.lpszClassName = name_.c_str();                   // mandatory

  if (!::RegisterClass (&myClass)){
    DWORD e=GetLastError();
    e=0; //warning killer
    assert(0);
  }
}

//////////////////////////////////////////////
// win_w32_worker implementation

win_w32_worker::win_w32_worker(
  int width, int height, char const* title)
:
  hwnd_(0),
  width_(width),
  height_(height),
  caption_(title)
{
  img_.reformat(img_fmt(width_,height_,3));
  img_util::fillWithRGBGradient(img_);
}
win_w32_worker::~win_w32_worker(){
}


/////////////////////////////////
// operations

void win_w32_worker::show(){
  ::ShowWindow (hwnd_, SW_SHOW);
  ::UpdateWindow (hwnd_);    
}
void win_w32_worker::hide(){
  ::ShowWindow (hwnd_, SW_HIDE);
}
void win_w32_worker::close(){
  ::SendMessage(hwnd_, WM_USER,0,0);
}

void win_w32_worker::position(int x, int y){
  ::SetWindowPos(hwnd_,0,x,y,0,0, 
    SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOZORDER);
}
void win_w32_worker::size(int cx, int cy){
  width_=cx;
  height_=cy;
  ::SetWindowPos(hwnd_,0,0,0,
    cx+borderx_, cy+bordery_, 
    SWP_NOACTIVATE|SWP_NOMOVE|SWP_NOZORDER);
}
void win_w32_worker::caption(char const* s){
  caption_=s;
  ::SetWindowText(hwnd_,s);
}
void win_w32_worker::beep(int t, int f){
  ::Beep(t, f);
}


void win_w32_worker::putImg(
  const img_data& src)
{
  {
    thr::Lock guard(mtx_);
    img_=src;
    ann_.clear();
  }
  update();
}

void win_w32_worker::putImgAnn(
  const img_data& src,
  const win_ann_abstract& a)
{
  {
    thr::Lock guard(mtx_);
    img_=src;
    ann_=a;
  }
  update();
}

void win_w32_worker::putAnn(
  const win_ann_abstract& a)
{
  {
    thr::Lock guard(mtx_);
    ann_=a;
  }
  update();
}

void win_w32_worker::update(){
  ::InvalidateRect(hwnd_,0,0);
  ::UpdateWindow (hwnd_);    
}

/////////////////////////////////
// user input

bool win_w32_worker::queryEvent(
  int& id, win_event_abstract& e)
{
  thr::Lock guard(mtx_);
  return log_.pop_back(id,e);
}

//called from guarded section
//of Window procedure
void win_w32_worker::addClick(
  int button, int x,int y, int mod)
{
  win_event::Type buttons[]={
    win_event::MouseLeft,
    win_event::MouseMiddle,
    win_event::MouseRight
  };

  int normy=img().fmt().height()-y;
  log_.push_front(win_event(buttons[button], mod, x, normy));
}
//called from guarded section
//of Window procedure
void win_w32_worker::addChar(
  int type, int ch, int mod)
{
  log_.push_front(win_event(
    (win_event::Type)type, mod, ch, 0));
}

/////////////////////////////////
// virtual members

int win_w32_worker::operator()(){
  static WinClass wc(WindowProcedure, "ssWinClass");

  HWND hwnd = ::CreateWindow (
    wc.name(),            // name of a registered window class
    caption_.c_str(),     // window caption
    WS_OVERLAPPEDWINDOW,  // window style
    CW_USEDEFAULT,        // x position
    CW_USEDEFAULT,        // y position
    CW_USEDEFAULT,        // width        
    CW_USEDEFAULT,        // height
    0,                    // handle to parent window
    0,                    // handle to menu
    GetModuleHandle(0),   // application instance
    0);                   // window creation data

  //fixing the link between the window and object
  s_manager.registerWorker(this, hwnd);

  RECT rcw, rcc;
  GetWindowRect(hwnd, &rcw);
  GetClientRect(hwnd, &rcc);
  borderx_=(rcw.right-rcw.left)-(rcc.right-rcc.left);
  bordery_=(rcw.bottom-rcw.top)-(rcc.bottom-rcc.top);
  ::SetWindowPos(hwnd,0,0,0,
    width_+borderx_, height_+bordery_, 
    SWP_NOACTIVATE|SWP_NOMOVE|SWP_NOZORDER);

  hwnd_=hwnd;
  cv_.signal();

  while (1){
    MSG msg;
    int status = ::GetMessage (&msg, 0, 0, 0);

    if (status == 0)//window dead
      break;
    if (status == -1){//error
      assert(0);
      break;
    }
    ::DispatchMessage(& msg);
  }
  return 0;
}
void win_w32_worker::cleanup(){
  close();
}






