#include "win_x_renderer_imlib.hpp"
#include "win_x_connection.hpp"

#include "../../img/img_data.hpp"
#include "../../img/img_util.hpp"
#include "../../img/access/img_access_copy.hpp"


#include <Imlib.h>

#include <boost/smart_ptr.hpp>

#include <stdio.h>

///////////////////////////////////////
// singleton wrapper around Imlib

class ImlibWrap{
private:
  ImlibWrap();
public:
  ~ImlibWrap();
  static ImlibWrap& instance();
  ImlibData* lib(){return lib_;}
  Display* dpy(){return dpy_;}
private:
  ImlibData* lib_; 
  Display* dpy_; 
};

ImlibWrap::ImlibWrap(){
  dpy_= win_x_connection::instance().display();
  lib_=Imlib_init(dpy_);

  if (!lib_)
    throw "ImlibWrap::ctor - "
          "Imlib_init returned 0.";
}

ImlibWrap::~ImlibWrap(){
}

ImlibWrap& ImlibWrap::instance(){
  static ImlibWrap theObject;
  return theObject;
}


///////////////////////////////////////
// private implementation

class win_x_renderer_imlibImp{
  int width_;
  int height_;
  ImlibImage *img_;
  ImlibData* lib_; 
  Display* dpy_; 
public:
  win_x_renderer_imlibImp();
  ~win_x_renderer_imlibImp();

  void setDefault(int w, int h);
  void setImg(const img_data& );
  void getImg(img_data& );
  void render(Drawable);
};

win_x_renderer_imlibImp::win_x_renderer_imlibImp()
{
  width_=0; 
  height_=0; 
  img_=0;
  lib_=ImlibWrap::instance().lib();
  dpy_=ImlibWrap::instance().dpy();
}

win_x_renderer_imlibImp::~win_x_renderer_imlibImp(){
  if (img_)
    Imlib_destroy_image(lib_, img_);
}

void win_x_renderer_imlibImp::setDefault(int w, int h){
  img_data img(img_fmt(w,h,3));
  img_util::fillWithRGBGradient(img);
  setImg(img);
}

void win_x_renderer_imlibImp::setImg(const img_data& img){
  int w=img.fmt().width();
  int h=img.fmt().height();

  //invalidate and (re)create imlib structure if needed
  //problem: each time the format changes, 
  //  imlib image must be recreated'
  //  img might not have szpix of 3,
  //  so we have to call Imlib_create_image_from_data
  //  with a dummy buffer of correct dimensions :-(
  if (w!=width_ || h!=height_){
    Imlib_destroy_image(lib_, img_);
    img_=0;
  }
  if (!img_) {
    void* dummy=calloc(w*h*3,1);
    img_=Imlib_create_image_from_data( lib_, 
        static_cast<unsigned char*>(dummy), 0, w, h);
    free(dummy);
    width_=w;
    height_=h;
  }

  //prepare tmp img for in-place manipulation
  img_data img_tmp( img_fmt(w,h,3), 
   (char*)img_->rgb_data, w*h*3);

  //copy input img to Imlib space
  img_access_copy(img, img_tmp);

  // fix bottom-up image
  img_util::flipVertical(img_tmp);

  // fix imlib endianness?
}

void win_x_renderer_imlibImp::getImg(img_data& img){
  if (!img_) {
    setDefault(320,240);
    assert(img_);
  }

  img.reformat(img_fmt(width_, height_, 3));

  //prepare tmp img for in-place manipulation
  img_data img_tmp( img_fmt(width_,height_,3), 
   (char*)img_->rgb_data, img.fmt().szBits());

  //copy Imlib img to output img
  img_access_copy(img_tmp, img);

  // fix bottom-up image
  img_util::flipVertical(img);

  // fix imlib endianness?
}

void win_x_renderer_imlibImp::render(Drawable d){
  if (!img_) {
    setDefault(320,240);
    assert(img_);
  }

  Imlib_changed_image(lib_, img_);
  Imlib_render(lib_, img_, width_, height_);

  //the unnecessary copy must be built due to the
  //inappropriate interface of libImlib...
  Pixmap p=Imlib_move_image(lib_, img_);
  XCopyArea(dpy_, p, d,
      DefaultGC(dpy_, DefaultScreen(dpy_)),
      0,0, width_, height_, 0,0);
  Imlib_free_pixmap(lib_, p);
}

//////////////////////////////////////////////
// win_x_renderer_imlib implementation

win_x_renderer_imlib::win_x_renderer_imlib():
  pimp(new win_x_renderer_imlibImp)
{}
win_x_renderer_imlib::~win_x_renderer_imlib(){}

void win_x_renderer_imlib::operator()(
  const img_data& img, Drawable d)
{
  pimp->setImg(img);
  pimp->render(d);
}

void win_x_renderer_imlib::operator()(
  int w, int h, Drawable d)
{
  pimp->setDefault(w,h);
  pimp->render(d);
}

