#include "win_x_thread.hpp"
#include "win_x_connection.hpp"
#include "win_x_renderer_imlib.hpp"

#include "../root/win_base.hpp"
#include "../ann/win_ann.hpp"
#include "../ann/win_ann_dump.hpp"
#include "../event/win_event.hpp"
#include "../event/win_event_queue.hpp"

#include "../../grid/draw/grid_draw_x.hpp"
#include "../../grid/draw/grid_draw_fileEps.hpp"

#include "../../img/img_data.hpp"
#include "../../img/io/img_io_save_bmp.hpp"
#include "../../img/access/img_access_copy.hpp"
#include "../../img/img_util.hpp"

#include "../../util/util_time.hpp"
#include "../../util/util_fs.hpp"


 
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <pthread.h>
#include <signal.h>

#include <unistd.h>
#include <sys/time.h>

#include <boost/smart_ptr.hpp>
#include <map>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>






/////////////////////////////////
// local static functions

#define NOLOG
#if defined(NOLOG) || defined(NDEBUG)
  #define sLog(X)
  #define sLog1(X,Y)
#else
  static void sLog(char const* msg){
    static time_t t0=-1;
    if (t0==-1) t0=util_time::msDayUtc();
    fprintf(stderr, "%ld:%ld- %s\n", 
      pthread_self(), util_time::msDayUtc()-t0, msg);
  }
  static void sLog1(char const* msg, int arg){
    static char buf[256];
    sprintf(buf, msg, arg);
    sLog(&buf[0]);
  }
#endif



class Guard{
  pthread_mutex_t& m_;
public:  
  Guard(pthread_mutex_t& m):m_(m){
    //sLog("Guard in");
    pthread_mutex_lock(&m_);
    //sLog("Guard OK");
  }
  ~Guard(){
    pthread_mutex_unlock(&m_);
    //sLog("Guard out");
  }
};


namespace{
void mysleep(int ms){
  static pthread_cond_t      cond  = PTHREAD_COND_INITIALIZER;
  static pthread_mutex_t     mutex = PTHREAD_MUTEX_INITIALIZER;

  timeval tv;
  gettimeofday(&tv, NULL);

  timespec ts;
  const int million=1000*1000;

  int new_usec= tv.tv_usec+1000*ms;
  ts.tv_sec  = tv.tv_sec + new_usec/million;
  ts.tv_nsec = (new_usec % million) * 1000;

  pthread_mutex_lock(&mutex);
  pthread_cond_timedwait(&cond, &mutex, &ts);
  pthread_mutex_unlock(&mutex);
}
}//unnamed namespace end





/////////////////////////////////
// The client window description

class WindowData{
  pthread_mutex_t mutex;

  // these members are read-only in the main thread
  Pixmap xpixmap;                          //Xlib pixmap
  Window wnd;                              //Xlib handle
  boost::shared_ptr<grid::draw_x> painter; //drawing object

  int wndx; // real position
  int wndy;
  int wndw; // real dimensions
  int wndh;
 
  win_event_queue log;

  // these members are read-only in the worker thread
  img_data img;
  win_ann ann;

  int szx; // specified dimensions 
  int szy; // (= image dimensions)
  char caption[256];
public:
  WindowData(int cx, int cy);
  ~WindowData();
private:
  WindowData(const WindowData& o);
  const WindowData& operator=(const WindowData&);
public:
  void setSize(int cx, int cy);
  void setCaption(char const* caption);
  bool queryEvent(int& it, win_event_abstract& e);
  void setImg(const img_data& d);
  void setAnn(const win_ann_abstract& a);
  void setImgAnn(const img_data&, const win_ann_abstract&);
public:
  void render(win_x_renderer& r);
  void drawAnn();
  void drawAnn(grid::draw& dc);
  void setGeometry(int x, int y, int w, int h);
  void pushEvent(const win_event& e);
  void getCaption(std::string& str);
  void getImg(img_data& d);
public:
  Window window(){Guard g(mutex);return wnd;}
  Pixmap pixmap(){Guard g(mutex);return xpixmap;}
  void setWindow(Window w){Guard g(mutex);wnd=w;}
  void setPixmap(Pixmap p){Guard g(mutex);xpixmap=p;}
  void setPainter(grid::draw_x* p);
  int width(){return szx;}
  int height(){return szy;}
  int wndWidth(){return wndw;}
  int wndHeight(){return wndh;}
};

WindowData::WindowData(int cx, int cy){
  pthread_mutex_init(&mutex,0);
  caption[0]='X';
  caption[1]=0;
  setSize(cx,cy);
  wnd=(Window)-1;
}
WindowData::~WindowData(){
  pthread_mutex_destroy(&mutex);
}


// interface to main program
void WindowData::setSize(int w, int h){
  Guard g(mutex);
  img.reformat(img_fmt(w,h,3));
  img_util::fillWithRGBGradient(img);
  ann.clear();
  szx=w;
  szy=h;
}
void WindowData::setCaption(char const* title){
  Guard g(mutex);
  const int sz=sizeof(caption);
  strncpy(&caption[0], title, sz-1);
  caption[sz-1]=0;
}
bool WindowData::queryEvent(
  int& it, win_event_abstract& e)
{
  Guard g(mutex);
  return log.pop_back(it,e);
}
void WindowData::setImg(const img_data& d){
  Guard g(mutex);
  ann.clear();
  img=d;
}
void WindowData::setAnn(const win_ann_abstract& a){
  Guard g(mutex);
  ann=a;
}
void WindowData::setImgAnn(
  const img_data& d, 
  const win_ann_abstract& a)
{
  Guard g(mutex);
  img=d;
  ann=a;
}

// interface to worker thread
void WindowData::render(win_x_renderer& r){
  Guard g(mutex);
  r(img, xpixmap);
}
void WindowData::drawAnn(){
  Guard g(mutex);
  grid::draw_x* p=painter.get();
  assert(p);
  win_ann_dump::toDC(ann,*p);
}
void WindowData::drawAnn(grid::draw& dc){
  Guard g(mutex);
  win_ann_dump::toDC(ann,dc);
}
void WindowData::getCaption(std::string& str){
  Guard g(mutex);
  str=&caption[0];
}
void WindowData::getImg(img_data& d){
  Guard g(mutex);
  d=img;
}
void WindowData::setPainter(grid::draw_x* p){
  Guard g(mutex);
  painter.reset(p);
}
void WindowData::setGeometry(int x, int y, int w, int h){
  Guard g(mutex);
  wndx = x; 
  wndy = y;
  wndw = w; 
  wndh = h;
  assert(painter.get()!=0);
  painter->reformat(1,0, -1,wndh-1); 
}
void WindowData::pushEvent(const win_event& e){
  Guard g(mutex);
  log.push_front(e);
}








//thread entry function declaration
namespace{
  extern "C" void* WorkerThread(void* ptr);
}



////////////////////////////////
// win_x_thread_imp 

class win_x_thread_imp{
  typedef std::map<int, boost::shared_ptr<WindowData> > WindowMap;
  // X data
  Display* dpy_; //X connection
  Window wndInput_;
  int current_;
  WindowMap map_; //clients' windows

  //thread data
  bool notDone_;
  pthread_t tid_;
  pthread_cond_t cond_;
  pthread_mutex_t mutex_;

  win_x_renderer_imlib renderer_; //rendering object

public:
  //client messages
  enum ClientMsg{
    Update, Render,
    SetSize, SetPosition, SetCaption,
    CreateWnd, DestroyWnd, ShutDown
  };
//audience: win_x
public:
  win_x_thread_imp();
  ~win_x_thread_imp();

  int createWindow(int cx, int cy);
  void destroyWindow(int id);

  // window operations
  void position(int id, int x, int y);
  void size(int id, int x, int y);
  void caption(int id, char const* title);
  void flash(int);

  // user input
  bool queryEvent(int id, int& it, win_event_abstract& e);

  // output
  void putImg(int id, const img_data& img);
  void putImgAnn(
    int id, const img_data& img,
    const win_ann_abstract& a);
  void putAnn( int id, 
    const win_ann_abstract& a);

//audience: workerthread
public:
  //window access
  WindowData& wnd(int id);
  int index(Window w);
  void remove(int id);

  //synchronization
  void waitForWorker();
  void releaseMain();
  void sendMessage(int msg, int id=-1, 
                   int par1=-1, int par2=-1);
  
  //attributes
  Display* dpy(){return dpy_;}
  win_x_renderer& renderer(){return renderer_;}
  void setWndInput(Window w){wndInput_=w;}
  Window wndInput(){return wndInput_;}
};

win_x_thread_imp::win_x_thread_imp(){
  //init X
  dpy_= win_x_connection::instance().display();
  current_=0;

  // create the worker thread
  sLog("win_x_thread_imp::ctor: creating the worker");
  pthread_cond_init(&cond_,0);
  pthread_mutex_init(&mutex_,0);
  notDone_=true;
  pthread_create(&tid_, 0, WorkerThread, this);
}
win_x_thread_imp::~win_x_thread_imp(){
  int count=0;
  WindowMap::iterator it=map_.begin();
  while (it != map_.end()){
    ++count;
  }
  sLog1("win_x_thread_imp::dtor - "
        "there were %d dangling windows.", count);

  waitForWorker();
  sLog("win_x_thread_imp::dtor - closing the worker");
  sendMessage(ShutDown);
  waitForWorker();

  pthread_mutex_destroy(&mutex_);
  pthread_cond_destroy(&cond_);
}


WindowData& win_x_thread_imp::wnd(int id){
  Guard g(mutex_);
  WindowMap::iterator it=map_.find(id);
  assert(it != map_.end()); 
  return *(it->second);
}
int win_x_thread_imp::index(Window w){
  Guard g(mutex_);
  WindowMap::iterator it=map_.begin();
  int rv=-1;
  while (it != map_.end()){
    if ((it->second)->window()==w){
      rv=it->first;
      break;
    }
    ++it;
  }
  return rv;
}
void win_x_thread_imp::remove(int id){
  Guard g(mutex_);
  WindowMap::iterator it=map_.find(id);
  assert(it != map_.end()); 
  map_.erase(it);
}


void win_x_thread_imp::waitForWorker(){
  Guard g(mutex_);
  while (notDone_){
    sLog("win_x_thread_imp::waiter: waiting...");
    pthread_cond_wait(&cond_, &mutex_);
  }
  sLog("win_x_thread_imp::waiter: success!");
}
void win_x_thread_imp::releaseMain(){
  sLog("win_x_thread_imp: releasing the main thread.");
  Guard g(mutex_);
  notDone_=false;
  pthread_cond_signal(&cond_);
}
void win_x_thread_imp::sendMessage(
  int msg, int id, 
  int par1, int par2)
{
  XEvent ev;
  ev.type=ClientMessage;
  ev.xclient.format=16;
  ev.xclient.data.s[0]=msg;
  ev.xclient.data.s[1]=id;
  ev.xclient.data.s[2]=par1;
  ev.xclient.data.s[3]=par2;

  { Guard g(mutex_); notDone_=true;}
 
  sLog1("win_x_thread_imp::sender: sending %d", msg);

  XSendEvent(dpy_, wndInput_, False,0, &ev);
  XFlush(dpy_);
}




//////////////////////////////
// window creation
int win_x_thread_imp::createWindow(int cx, int cy){
  sLog("Called win_x_thread_imp::createWindow");

  // create new map entry
  int id=current_;

  // store the new entry
  { Guard g(mutex_);
    ++current_;
    map_[id].reset(new WindowData(cx,cy));
  }

  waitForWorker();
  sendMessage(CreateWnd, id,cx,cy);
 
  return id;
}

void win_x_thread_imp::destroyWindow(int id){
  sLog1("Called win_x_thread_imp::destroyWindow(%d)",id);

  // destroy the window
  assert(wnd(id).window()!=0);
  //XDestroyWindow(dpy_, wnd(id).window());
  //XFlush(dpy_);

  // notify the worker thread
  waitForWorker();
  sendMessage(DestroyWnd, id);
  
  waitForWorker();

  remove(id);
}


//////////////////////////////
// window properties
void win_x_thread_imp::position(int id, int x, int y){
  sLog("Called win_x_thread_imp::position");
  waitForWorker();
  sendMessage(SetPosition, id, x, y);
}
void win_x_thread_imp::size(int id, int cx, int cy){
  sLog("Called win_x_thread_imp::size");
  WindowData& wd=wnd(id);
  if ((cx!=wd.width()) || (cy!=wd.height())) {
    wd.setSize(cx,cy);
    waitForWorker();
    sendMessage(SetSize, id, cx, cy);
  }
}

void win_x_thread_imp::caption(int id, char const* title){
  sLog("Called win_x_thread_imp::caption");
  wnd(id).setCaption(title);
  waitForWorker();
  sendMessage(SetCaption, id);
}

void win_x_thread_imp::flash(int){
  sLog("Called win_x_thread_imp::flash");
  XBell(dpy_,0);
}

/////////////////////////////////
// user input
bool win_x_thread_imp::queryEvent(
  int id, int& it, win_event_abstract& e)
{
  return wnd(id).queryEvent(it, e);
}





//////////////////////////////
// output

void win_x_thread_imp::putImg(
  int id, const img_data& img)
{
  sLog("Called win_x_thread_imp::putImg");
  wnd(id).setImg(img);
  waitForWorker();
  sendMessage(Render, id);
}

void win_x_thread_imp::putImgAnn(
  int id, const img_data& img,
  const win_ann_abstract& a)
{
  sLog("Called win_x_thread_imp::putImgAnn");
  WindowData& wd=wnd(id);
  wd.setImgAnn(img,a);
  waitForWorker();
  sendMessage(Render, id);
}

void win_x_thread_imp::putAnn( int id, 
  const win_ann_abstract& a)
{
  sLog("Called win_x_thread_imp::putAnn");
  wnd(id).setAnn(a);
  waitForWorker();
  sendMessage(Update, id);
}



















///////////////////////////////////
// win_x_thread delegates requests 
// to win_x_thread_imp

win_x_thread::win_x_thread():
  pimp(new win_x_thread_imp)
{}
win_x_thread::~win_x_thread(){
  delete pimp;
}
win_x_thread& win_x_thread::instance(){
  static win_x_thread theInstance;
  return theInstance;
}

// window creation
int win_x_thread::createWindow(int cx, int cy){
  return pimp->createWindow(cx, cy);
}
void win_x_thread::destroyWindow(int id){
  pimp->destroyWindow(id);
}

// window operations
void win_x_thread::position(int id, int x, int y){
  pimp->position(id, x, y);
}
void win_x_thread::size(int id, int x, int y){
  pimp->size(id, x, y);
}
void win_x_thread::caption(int id, char const* title){
  pimp->caption(id, title);
}
void win_x_thread::flash(int f){
  pimp->flash(f);
}

// user input
bool win_x_thread::queryEvent(
  int id, int& it, win_event_abstract& e)
{
  return pimp->queryEvent(id,it,e);
}

// output
void win_x_thread::putImg(int id, const img_data& img){
  pimp->putImg(id, img);
}
void win_x_thread::putImgAnn(
  int id, const img_data& img,
  const win_ann_abstract& a)
{
  pimp->putImgAnn(id, img, a);
}
void win_x_thread::putAnn( int id, 
  const win_ann_abstract& a)
{
  pimp->putAnn(id, a);
}












////////////////////////////////
// Worker Thread static functions
namespace{

//simple operations
void wt_caption(
  win_x_thread_imp* self, WindowData& wd)
{
  std::string str;
  wd.getCaption(str);
  XTextProperty wndName;
  char *p=(char*)str.c_str();
  if (!XStringListToTextProperty(&p,1,&wndName)){
    sLog("wt_caption: error in XStringListToTextProperty.");
    assert(0);
  }
  XSetWMName(self->dpy(), wd.window(), &wndName);
  XFlush(self->dpy());
  XFree(wndName.value);//suggested by valgrind
  sLog("wt_caption: on exit.");
}
void wt_size(win_x_thread_imp* self, 
  WindowData& wd, int cx, int cy)
{
  Pixmap pmnew= XCreatePixmap(self->dpy(), 
    XDefaultRootWindow(self->dpy()), 
    wd.width(), wd.height(),
    XDefaultDepth(self->dpy(), XDefaultScreen(self->dpy())));
  XFreePixmap(self->dpy(), wd.pixmap());
  wd.setPixmap(pmnew);

  XResizeWindow(self->dpy(), wd.window(), cx, cy);
  XFlush(self->dpy());
  sLog("wt_size: on exit.");
}
void wt_position(win_x_thread_imp* self, 
  WindowData& wd, int cx, int cy)
{
  XMoveWindow(self->dpy(), wd.window(), cx, cy);
  XFlush(self->dpy());
  sLog("wt_position: on exit.");
}
void wt_fixGeometry(
  win_x_thread_imp* self, WindowData& wd)
{
  XWindowAttributes wa;
  XGetWindowAttributes(self->dpy(),wd.window(),&wa);
  wd.setGeometry(wa.x, wa.y, wa.width, wa.height);
  sLog("wt_fixGeometry: on exit.");
}
 
// painting
void wt_repaint(
  win_x_thread_imp* self, WindowData& wd)
{
  sLog("wt_repaint: drawing xpixmap.");
  XSetWindowBackgroundPixmap(self->dpy(),wd.window(),wd.pixmap());
  XClearWindow(self->dpy(), wd.window());
  //XFlush(self->dpy());

  sLog("wt_repaint: annotating.");
  wd.drawAnn();
}


//input
int wt_getMod(int state){
  int mod=0;
  if (state & ShiftMask)
    mod|=win_event::Shift;
  if (state & ControlMask)
    mod|=win_event::Ctrl;
  if (state & Mod1Mask)
    mod|=win_event::Alt;
  return mod;
}

void wt_fixKey(
  win_x_thread_imp* self, WindowData& wd,
  unsigned int code, unsigned int state)
{
  KeySym ks=XKeycodeToKeysym(self->dpy(), code, 0);

  win_event::Type type=win_event::KeyboardAscii;
  int ch=ks&0xff;
  if (ks>=XK_space && ks<=XK_asciitilde){
    ; //do nothing
  } else if (ks>=XK_BackSpace && ks <=XK_Escape){
    ; //do nothing
  } else if (ks>=XK_F1 && ks <=XK_F12){
    type=win_event::KeyboardFx;
    ch=(ks-XK_F1) & 0xff;
  } else if (ks>=XK_Left && ks <=XK_Down){
    type=win_event::KeyboardArrow;
    ch=(ks-XK_Left) & 0xff;
  } else if (ks==XK_Insert){
    ch=0;
    type=win_event::KeyboardEdit;
  } else if (ks==XK_Delete){
    ch=1;
    type=win_event::KeyboardEdit;
  } else if (ks==XK_Home){
    ch=2;
    type=win_event::KeyboardEdit;
  } else if (ks==XK_End){
    ch=3;
    type=win_event::KeyboardEdit;
  } else if (ks>=XK_Prior && ks <=XK_Next){
    ch=4+ks-XK_Prior;
    type=win_event::KeyboardEdit;
  } else if (ks>=XK_Shift_L && ks <=XK_Hyper_R){
    ch=-1; // ignore
  } else{
    ch=-1; // ignore
    XBell(self->dpy(),-50);
  }

  if (ch>=0){
    int mod=wt_getMod(state);
    wd.pushEvent(win_event(type, mod, ch, 0));
  }
}

void wt_fixButton(
  win_x_thread_imp* self, WindowData& wd,
  int x, int y, unsigned int button, unsigned int state)
{
  int mod=wt_getMod(state);

  static long lastClick1=0;
  static long lastClick2=0;
  const unsigned int doubleClickLimit=500;

  util_time::Data t_ms=util_time::msDayUtc();
  sLog("wt_fixButton: reading buttons.");
  switch(button){
  case Button1:
    if (t_ms-lastClick1<doubleClickLimit){
      lastClick1=0;
      mod|=win_event::DoubleClick;
    } else{
      lastClick1=t_ms;
    }
    wd.pushEvent(win_event(win_event::MouseLeft, 
      mod, x,wd.wndHeight()-y));
    break;
  case Button2:
    if (t_ms-lastClick2<doubleClickLimit){
      lastClick2=0;
      mod|=win_event::DoubleClick;
    } else{
      lastClick2=t_ms;
    }
    wd.pushEvent(win_event(win_event::MouseMiddle, 
      mod, x,wd.wndHeight()-y));
    break;
  case Button3:
    {
      //the file is named after the window caption...
      std::string fileName;
      wd.getCaption(fileName);
      for (unsigned i=0; i<fileName.size(); i++)
        if (!isalpha(fileName[i]) && !isdigit(fileName[i]))
          fileName.erase(i);
      std::string fullPath=util_fs::getLogDir();
      fullPath+=fileName;

      img_data myimg;
      wd.getImg(myimg);
      
      if (state & ControlMask){
        //bmp dump, no annotations
        fullPath+=".bmp";
        std::string unique=util_fs::
          uniqueFilename(fullPath.c_str());
        img_io_save_bmp saver(unique.c_str());
        if (myimg.fmt().szPixel()==2){
          img_data tmp(img_fmt(
            myimg.fmt().width(), 
            myimg.fmt().height(), 1));
          img_access_copy(myimg,tmp);
          saver(tmp);
        } else{
          saver(myimg);
        }
      } else{
        //eps annotation dump
        //if shift is pressed, the image is dumped as well
        fullPath+=".eps";
        std::string unique=util_fs::
          uniqueFilename(fullPath.c_str());

        //dump the annotation to the drawing object
        //dump the image as well if shift was pressed
        if (state & ShiftMask){
          grid::draw_fileEps dc(unique.c_str(),
            wd.width(),wd.height(), myimg);
          wd.drawAnn(dc);
        } else{
          grid::draw_fileEps dc(unique.c_str(),
            wd.width(),wd.height());
          wd.drawAnn(dc);
        }
      }
    }
    break;
  default:
    XBell(self->dpy(),-50);
    break;
  }
}


// creation

void wt_create(
  win_x_thread_imp* self, WindowData& wd, Atom& atomDel)
{
  //Window prev;
  //int mode;
  //XGetInputFocus(self->dpy(), &prev, &mode);
  int screen=XDefaultScreen(self->dpy());
  wd.setPixmap(
    XCreatePixmap(self->dpy(), 
      XDefaultRootWindow(self->dpy()), 
      wd.width(), wd.height(),
      XDefaultDepth(self->dpy(), screen)));
  wd.render(self->renderer());
  
  Window wnd=XCreateSimpleWindow(
    self->dpy(), DefaultRootWindow(self->dpy()),
    0, 0, wd.width(), wd.height(), 0,
    BlackPixel(self->dpy(), screen),
    WhitePixel(self->dpy(), screen));
  wd.setPainter(new grid::draw_x (
    1,0, -1,wd.height()-1, 
    self->dpy(),wnd));
  wd.setWindow(wnd);

  //register wm_delete_window protocol
  XSetWMProtocols(self->dpy(), wnd, &atomDel, 1);

  XSelectInput(
    self->dpy(), wnd,
    StructureNotifyMask | KeyPressMask |
    ExposureMask        | ButtonPressMask);
  XMapWindow(self->dpy(), wnd);

  XFlush(self->dpy());
  //mysleep(1000);
  //XSetInputFocus(self->dpy(), prev, mode, CurrentTime);
}



////////////////////////////////
// Worker Thread entry procedure definition
//
extern "C" void* WorkerThread(void* ptr){
  win_x_thread_imp* self=(win_x_thread_imp*) ptr;

  Atom atomDel = XInternAtom(self->dpy(), "WM_DELETE_WINDOW", False);
  Atom atomProt= XInternAtom(self->dpy(), "WM_PROTOCOLS", False);

  /* mask all allowed signals */
  sigset_t mask;
  sigfillset(&mask); 
  pthread_sigmask(SIG_BLOCK, &mask, NULL);
 
  /* create the window and init self*/
  sLog("WorkerThread: opening window....");
  //XBell(self->dpy(),-50);
  Window wndInput=XCreateWindow(
    self->dpy(), DefaultRootWindow(self->dpy()),
    0,0, 1,1, 0, 0, 
    InputOnly, CopyFromParent, 0,0);
  XSelectInput(self->dpy(), wndInput, 0);
  self->setWndInput(wndInput);
  self->releaseMain();
 
  bool shutdown=false;
  while(!shutdown) {
    /*process event */
    sLog("WorkerThread - Waiting for event");
    XEvent ev;
    XNextEvent(self->dpy(), &ev);
    sLog1("WorkerThread - got event, send_event=%d", ev.xany.send_event);
    if (ev.type == ClientMessage) {
      if (ev.xclient.message_type == atomProt){
        if (static_cast<Atom>(ev.xclient.data.l[0]) == atomDel){
          sLog("WorkerThread - server shutdown or explicit kill");
          //exit(-1);
        } else{
          sLog1("WorkerThread - unknown WM_PROTOCOL %d",
            ev.xclient.data.l[0]);
        }
      } else{
        int c=ev.xclient.data.s[0];
        if (c==win_x_thread_imp::ShutDown){
          sLog("WorkerThread - ShutDown");
          shutdown=true;
        } else{
          int id=ev.xclient.data.s[1];
          WindowData& wd=self->wnd(id);
          int cx=ev.xclient.data.s[2];
          int cy=ev.xclient.data.s[3];
          switch (c){
          case win_x_thread_imp::Render:
            sLog1("WorkerThread - Render(%d)",id);
            wd.render(self->renderer());
            wt_repaint(self, wd);
            break;
          case win_x_thread_imp::Update:
            sLog1("WorkerThread - Update(%d)",id);
            wt_repaint(self, wd);
            break;
          case win_x_thread_imp::SetSize:
            sLog1("WorkerThread - SetSize(%d)",id);
            wd.render(self->renderer());
            wt_size(self, wd, cx,cy);
            break;
          case win_x_thread_imp::SetPosition:
            sLog1("WorkerThread - SetPosition(%d)",id);
            wt_position(self, wd, cx,cy);
            break;
          case win_x_thread_imp::SetCaption:
            sLog1("WorkerThread - SetCaption(%d)",id);
            wt_caption(self, wd);
            break;

          case win_x_thread_imp::CreateWnd:
            sLog1("WorkerThread - CreateWnd(%d)",id);
            wt_create(self,wd, atomDel);
            break;
          case win_x_thread_imp::DestroyWnd:
            XDestroyWindow(self->dpy(), wd.window());
            XFlush(self->dpy());
            XFreePixmap(self->dpy(), wd.pixmap());
            sLog1("WorkerThread - DestroyWnd(%d)",id);
            break;

          default:
            sLog1("WorkerThread - client event %d",c);
            break;
          }
        }
      }

      self->releaseMain();

    } else {
      int id=self->index(ev.xany.window);
      if (id>=0){
        WindowData& wd=self->wnd(id);
        if (ev.type == DestroyNotify) {
          sLog1("WorkerThread: DestroyNotify (%d)", id);
          continue;
        } else if (ev.type == MapNotify) {
          sLog1("WorkerThread: MapNotify (%d)", id);
          continue;
        } else if (ev.type == ConfigureNotify) {
          sLog1("WorkerThread: ConfigureNotify (%d)", id);
          wt_fixGeometry(self, wd);
          continue;
        } else if (ev.type == UnmapNotify) {
          sLog1("WorkerThread: UnmapNotify (%d)", id);
          continue;
        } else if (ev.type == Expose) {
          sLog1("WorkerThread: Expose (%d)", id);
          XEvent e;
          e.type=0;
          while (XCheckTypedWindowEvent(
            self->dpy(), ev.xany.window, Expose, &e));
          if (e.type==Expose){
            XPutBackEvent(self->dpy(), &e);
          } else{
            wt_repaint(self, wd);
          }
          continue;
        } else if (ev.type == ButtonPress) {
          sLog1("WorkerThread: ButtonPress (%d)", id);
          wt_fixButton(self, wd,      
            ev.xbutton.x, ev.xbutton.y,
            ev.xbutton.button, ev.xbutton.state);
          continue;    
        } else if (ev.type == KeyPress) {
          sLog1("WorkerThread: KeyPress (%d)", id);
          wt_fixKey(self, wd,      
            ev.xkey.keycode, ev.xbutton.state);
          continue;    
        } else{
          sLog1("WorkerThread - got event %d",ev.type);
        } 
      } else{
        sLog1("WorkerThread - unrecognized window, event=%d",ev.type);
      }
    }
  }

  static int rv=42;
  return &rv;
}



}// end unnamed namespace








